#include "BmnTrcMainWindow.h"
#include "ui_BmnTrcMainWindow.h"

#include "TrcConfigDispatcher.h"
#include "BmnTrcCountersDialog.h"
#include "BmnTrcDeviceChModel.h"
#include "BmnTrcSignalRouter.h"
#include "BmnTrcTreeColumns.h"
#include "base/DeviceIndexEnabled.h"
#include "base/Globals.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "m-stream-output/MStreamStatisticDialog.h"
#include "mldiscover/LldpDiscovery.h"
#include "widgets/DeviceStatusDialog.h"

#include <QBoxLayout>
#include <QFileDialog>
#include <QHostAddress>
#include <QLabel>
#include <QMessageBox>
#include <QMetaType>
#include <QPlainTextEdit>
#include <QScrollBar>
#include <QSemaphore>
#include <QSettings>
#include <QShortcut>
#include <QSpinBox>
#include <QTimer>

Q_DECLARE_METATYPE(QHostAddress)

BmnTrcMainWindow::BmnTrcMainWindow(QWidget *parent) :
      ReadoutAppMainWindow(parent)
    , ui(new Ui::BmnTrcMainWindow)
    , trcModel(new BmnTrcDeviceChModel(this))
    , countersDialog(new BmnTrcCountersDialog(this))
{
    ui->setupUi(this);
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions, ui->menuTools);
    ui->groupBoxL0Extend->setVisible(Globals::getDebugMode());

    ui->trcTableView->setModel(trcModel);

    init_connections();

    configuration_manager->setDefaultConfig(BaseDeviceAppConfig());

    updateControlState();

    auto signalRouter = &BmnTrcSignalRouter::instance();
    signalRouter->setMainWindow(this);
    signalRouter->setDiscoverDialog(discoverDialog);
    signalRouter->setConfigurationManager(configuration_manager);
    signalRouter->setMStreamStatisticDialog(mStreamStatisticDialog);
    signalRouter->setDeviceChModel(trcModel);
}

BmnTrcMainWindow::~BmnTrcMainWindow()
{
    delete ui;
}

DeviceStatusTableViewFancy* BmnTrcMainWindow::getDeviceStatusTableView()
{
    return ui->deviceStatusTableView;
}

void BmnTrcMainWindow::init_connections()
{
    connect(ui->actionPrint_FSM_inputs, &QAction::triggered,
            this, &BmnTrcMainWindow::printFsmInputs);
    connect(ui->actionSend_soft_clear, &QAction::triggered,
            this, &BmnTrcMainWindow::softClear);
//    connect(ui->actionReset_MStreamDump, &QAction::triggered,
//            this, &TrcMainWindow::mstream_dump_reset_all);

    connect(countersDialog, &BmnTrcCountersDialog::clearStat,
            this, &BmnTrcMainWindow::softClear);

    auto cfgDisp = &TrcConfigDispatcher::instance();
    cfgDisp->subscribeSpillConfig(ui->widgetSpill);
    cfgDisp->subscribeRcTrigConfig(ui->widgetTrig);
    cfgDisp->subscribePhysicalTriggerConfig(trcModel);
}

void BmnTrcMainWindow::saveAppGuiSettings(QSettings &s) const
{
    BaseMainWindow::saveAppGuiSettings(s);
    s.beginGroup("MW");
    s.setValue("showExperts", ui->actionShow_experts_settings->isChecked());
    s.endGroup();
}

void BmnTrcMainWindow::loadAppGuiSettings(const QSettings &s)
{
    BaseMainWindow::loadAppGuiSettings(s);
    bool showVar = s.value("MW/showExperts").toBool();
    ui->actionShow_experts_settings->setChecked(showVar);
    on_actionShow_experts_settings_toggled(showVar);
}

void BmnTrcMainWindow::updateControlState()
{
    bool inStateErrReady = curFsmState == FsmState::Ready || curFsmState == FsmState::Error;

    ui->pushButtonStart->setEnabled(curFsmState == FsmState::Ready && !underRemoteCtrl);
    ui->pushButtonStop->setEnabled(curFsmState == FsmState::Run && !underRemoteCtrl);

    controlStateEnable((inStateErrReady || Globals::getDebugMode()));

    ui->actionReset_MStreamDump->setEnabled(inStateErrReady);
    ui->actionSend_soft_reset->setEnabled(inStateErrReady);
    ui->deviceStatusTableView->setControlEnabled(inStateErrReady);
}

void BmnTrcMainWindow::deviceStatusUpdated(const BmnTrcStatus &status)
{
    ui->widgetSpill->setSpillStatus(status.phyTrig.spill);
}

bool BmnTrcMainWindow::isControlBlocked() const
{
    return blockControl || underRemoteCtrl;
}

void BmnTrcMainWindow::setConfig(const ConfigKey &config_key, const BaseDeviceAppConfig &newConfig)
{
    Q_UNUSED(config_key)
    if(config == newConfig)
        return;

    ReadoutAppMainWindow::setConfig(config_key, newConfig);

    updateGuiVal();
    updateControlState();
}

void BmnTrcMainWindow::updateGuiVal()
{
    ++blockControl;
    const auto &ds = config.defaultSetup;

    const auto &phyTrig = ds.trcCore.phyTrig;
    ui->spinBoxMatchWin->setValue(phyTrig.matchWin);
    ui->spinBoxMatchWin->setMaximum(phyTrig.trigDelay);
    ui->spinBoxTrigDelay->setValue(phyTrig.trigDelay);
    ui->spinBoxL0Extend->setValue(phyTrig.l0Extend);

    ui->groupBoxChWaveform->setChecked(phyTrig.oscEn);
    ui->radioButtonOscDelayed->setChecked(phyTrig.oscDelayed);
    ui->spinBoxLatency->setValue(phyTrig.chWfLatency);
    ui->spinBoxReadPoints->setValue(phyTrig.chWfReadPoints);

    ui->groupBoxCalibTrigger->setConfig(phyTrig.calibTrigger);
    --blockControl;

    updateControlState();
}

void BmnTrcMainWindow::controlStateEnable(bool enable)
{
    qWarning() << "Temporary block:" << __PRETTY_FUNCTION__;
   // trcModel->setEnabled(enable);
}

// ==== Slots from UI

void BmnTrcMainWindow::on_actionShow_experts_settings_toggled(bool showExperts)
{
    ui->trcTableView->setColumnHidden(static_cast<int>(TrcTreeColumns::COL_START_MASK_EN), !showExperts);
    ui->scrollAreaWidgetContents->setEnabled(showExperts);
    trcModel->setExpertsSettings(showExperts);
}

void BmnTrcMainWindow::on_pushButtonStart_clicked()
{
    if(isControlBlocked()) return;
    if(!config.hasEnabledDevices()) {
        qWarning() << "There is no enabled device";
        return;
    }

    emit performLocalStart();
}

void BmnTrcMainWindow::on_pushButtonStop_clicked()
{
    if(isControlBlocked()) return;
    emit performLocalStop();
}

void BmnTrcMainWindow::on_actionReset_MStreamDump_triggered()
{
    emit mstream_dump_reset_all(true);
}

void BmnTrcMainWindow::on_actionSend_soft_reset_triggered()
{
    auto ret = QMessageBox::warning(this, "Confirm device soft reset...",
                                    "Sending soft reset to device entails "
                                    "resetting device to factory settings "
                                    "followed by writting to current configuration",
                                    {QMessageBox::Yes, QMessageBox::No}, QMessageBox::No);
    if(ret == QMessageBox::Yes)
        emit softReset();
}

void BmnTrcMainWindow::on_spinBoxMatchWin_valueChanged(int mw)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.matchWin = static_cast<quint16>(mw);
    config.unify_config();
    emit configChanged(config);
}


void BmnTrcMainWindow::on_spinBoxTrigDelay_valueChanged(int delay)
{
    ui->spinBoxMatchWin->setMaximum(delay);
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.trigDelay = static_cast<quint16>(delay);
    config.unify_config();
    emit configChanged(config);
}

void BmnTrcMainWindow::on_spinBoxL0Extend_valueChanged(int extend)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.l0Extend = static_cast<quint16>(extend);
    config.unify_config();
    emit configChanged(config);
}

void BmnTrcMainWindow::on_radioButtonOscDelayed_toggled(bool checked)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.oscDelayed = checked;
    config.unify_config();
    emit configChanged(config);
}


void BmnTrcMainWindow::on_groupBoxChWaveform_toggled(bool checked)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.oscEn = checked;
    config.unify_config();
    emit configChanged(config);
}

void BmnTrcMainWindow::on_spinBoxLatency_valueChanged(int lat)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.chWfLatency = static_cast<quint16>(lat);
    config.unify_config();
    emit configChanged(config);
}

void BmnTrcMainWindow::on_spinBoxReadPoints_valueChanged(int points)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.chWfReadPoints = static_cast<quint16>(points);
    config.unify_config();
    emit configChanged(config);
}

void BmnTrcMainWindow::on_groupBoxCalibTrigger_cfgUpdated(const CalibrationTriggerConfig &c)
{
    if(blockControl) return;
    config.defaultSetup.trcCore.phyTrig.calibTrigger = c;
    config.defaultSetup.trcCore.phyTrig.unifyCalibTrig();
    config.unify_config();
    emit configChanged(config);
}

void BmnTrcMainWindow::on_actionCounters_dialog_triggered()
{
    countersDialog->show();
}
