#include "CruSignalRouter.h"

#include <QObject>

#include "CruCore.h"
#include "CruDevice.h"
#include "CruDeviceGlobal.h"
#include "CruFsm.h"
#include "base/DeviceId.h"
#include "configuration-manager/ConfigurationManager.h"
#include "mldiscover/DeviceDispatcher.h"
#include "lib-common/ReadoutAppConfigDb.h"
#include "lib-common/ReadoutAppConfigDispatcher.h"
#include "remote-control-server/RemoteControlServer.h"


#ifdef QT_WIDGETS_LIB
#include "device-discover/DiscoverDialog.h"
#include "configuration-manager/ConfigurationManager.h"
#include "FeLinkDialog/FeLinkStatusDialog.h"
#include "m-stream-output/MStreamStatisticDialog.h"
#include "RunLogic/RunLogicWidget.h"
#include "TrigCtrlStat/TrigCtrlStatWidget.h"

#include "CruMainWindow.h"
#endif

static constexpr const auto device_filter_list = {
    DEVICE_ID_CRU16WR
};

CruSignalRouter::CruSignalRouter()
{
    setDeviceIdFilter(device_filter_list);
}

void CruSignalRouter::setCore(CruCore *p)
{
    core = p;
    setReadoutAppCore(p);
}

void CruSignalRouter::setCruFsm(CruFsm *p)
{
    cruFsm = p;
    setDeviceMgrFsm(p);
}

void CruSignalRouter::setReadoutAppConfigDispatcher(ReadoutAppConfigDispatcher *p)
{
    ReadoutAppSignalRouter::setConfigDispatcher(p);
    configDispatcher = p;
}

void CruSignalRouter::connect_ConfigurationDispatcher() const
{
    ReadoutAppSignalRouter::connect_ConfigurationDispatcher();

    Q_ASSERT(configDispatcher);
    if (deviceDispatcher)
        deviceDispatcher->subscribeDevices(configDispatcher);

    if(core) {
        configDispatcher->subscribeConfig(core);
        configDispatcher->subscribeCS(core);
    }
    if(configDb) {
        configDispatcher->subscribeConfig(configDb);
        configDispatcher->subscribeCS(configDb);
    }
    if(configurationManager)
        configDispatcher->subscribeCS(configurationManager);
}

#ifdef QT_WIDGETS_LIB
void CruSignalRouter::setMainWindow(CruMainWindow *p)
{
    // set mainWindow must be before core
    mainWindow = p;
    setReadoutAppMainWindow(p);
}

void CruSignalRouter::connect_MainWindow() const
{
    ReadoutAppSignalRouter::connect_MainWindow();

    if(!core)
        return;

    if(!mainWindow)
        return;

    if (feLinkStatusDialog && feLinkStatusDialog->getModel())
        connect(core, &CruCore::clearStats,
                feLinkStatusDialog->getModel(), &FeLinkTreeModel::clearStats);

    if(configDispatcher)
        configDispatcher->subscribeCS(mainWindow);
     connect(core, &ReadoutAppCore::setStatus,
            mainWindow, &ReadoutAppMainWindow::setStatus);

    connect(mainWindow, &CruMainWindow::performLocalStart,
            core, &CruCore::performLocalStart);
    connect(mainWindow, &CruMainWindow::performLocalStop,
            core, &CruCore::performLocalStop);

    if(auto w = mainWindow->getTrigCtrlStatWidget()) {
        if(configDispatcher)
            configDispatcher->subscribeTrigCtrlStatConfig(w);
        connect(deviceDispatcher, &DeviceDispatcher::deviceListChanged,
                w, &TrigCtrlStatWidget::setDeviceList);
    }
    if (auto w = mainWindow->getRunLogicWidget()) {
        w->setController(core->runLogicController);
    }

//    LldpDiscovery::subscribe(mainWindow, device_filter_list);

//    connect(mainWindow, &CruMainWindow::performLocalStart,
//            core, &CruCore::performLocalStart);

//    connect(mainWindow, &CruMainWindow::performLocalStop,
//            core, &CruCore::performLocalStop);

    LldpDiscovery::subscribe(discoverDialog, device_filter_list);
    deviceDispatcher->subscribeDevices(discoverDialog);

    Q_ASSERT(deviceMgrFsm);
    connect(deviceMgrFsm, &BaseFsm::stateChanged,
            mainWindow, &CruMainWindow::onFsmStateChange);

    connect(mainWindow, &CruMainWindow::printFsmInputs,
            deviceMgrFsm, &BaseFsm::printFsmInput);
}
#endif


void CruSignalRouter::connect_RemoteControlServer() const
{
    ReadoutAppSignalRouter::connect_RemoteControlServer();

    if(!remoteControlServer)
        return;

    if (core) {
        connect(remoteControlServer, &RemoteControlServer::startRequest,
                core, &CruCore::remoteStartRequest);
        connect(remoteControlServer, &RemoteControlServer::stopRequest,
                core, &CruCore::remoteStopRequest);

        connect(core, &CruCore::sendRunError,
                remoteControlServer,
                QOverload<const RcRunId &, QString>::of(&RemoteControlServer::onRunError));

        connect(core, &CruCore::sendRunWarning,
                remoteControlServer, &RemoteControlServer::onRunWarning);
    }
}

void CruSignalRouter::connect_CruDeviceGlobal(const CruDeviceGlobal *devGlob)
{
    if(!devGlob)
        return;

    auto dev = devGlob->getDevice<CruDevice>();
    connect_device(devGlob, dev);

    connect(dev, &ModularMregDevice::deviceDisconnected,
            core, &CruCore::deviceDisconnected);

    connect(dev, &ModularMregDevice::statusUpdated,
            devGlob, &CruDeviceGlobal::deviceStatusUpdated);

    connect(core, &CruCore::preRunReset,
            dev, &CruDevice::preRunReset);

    connect(core, &CruCore::mstreamResetAll,
            devGlob, &CruDeviceGlobal::resetMStream);
    if(cruFsm) {
        connect(dev, &CruDevice::preRunReseted,
                cruFsm, &CruFsm::setDevRunId);
    }

#ifdef QT_WIDGETS_LIB
    if(mainWindow){
        if (auto w = mainWindow->getTrigCtrlStatWidget()) {
            connect(dev, &ModularMregDevice::trigCtrlBpUpdated,
                    w, &TrigCtrlStatWidget::setTrigCtrlBp);
        }
    }
#endif

    deviceDispatcher->subscribeHostAddressUpdates(devGlob->getDeviceIndex(),
                                                  dev);
}
