#include "TcpClient.h"

#include <QtNetwork>
#include <QDebug>
#include <QSemaphore>

#include "base/PrintInfo.h"

#if QT_VERSION < 0x050000
Q_DECLARE_METATYPE(QAbstractSocket::SocketError)
#endif

namespace {
const int PACK_SIZE_LIMIT = 20e6;
}

TcpClient::TcpClient(QSemaphore *dataFreeSizeSem, QObject *parent) :
    QObject(parent),
    queueFreeSizeSem(dataFreeSizeSem),
    enabled(false),
    tcpSocket(new QTcpSocket(this)),
    reconnectTimer(this),
    delayReadTimer(this),
    connectingAttempts(0),
    counter(0)
{
    qRegisterMetaType<QHostAddress>();
    qRegisterMetaType<QAbstractSocket::SocketError>();

    connect(tcpSocket, SIGNAL(readyRead()), this, SLOT(readData()), Qt::QueuedConnection);
    connect(tcpSocket, SIGNAL(connected()), this, SLOT(connected()), Qt::QueuedConnection);
    connect(tcpSocket, SIGNAL(disconnected()), this, SLOT(disconnected()), Qt::QueuedConnection);
    connect(tcpSocket, SIGNAL(error(QAbstractSocket::SocketError)),
            this, SLOT(displayError(QAbstractSocket::SocketError)), Qt::QueuedConnection);
    connect(&reconnectTimer, SIGNAL(timeout()), this, SLOT(requestNewData()), Qt::QueuedConnection);
    reconnectTimer.setSingleShot(true);

    delayReadTimer.setSingleShot(true);
    delayReadTimer.setInterval(1000);
    connect(&delayReadTimer, SIGNAL(timeout()), this, SLOT(readData()), Qt::QueuedConnection);
}


TcpClient::~TcpClient()
{
    stopReadOut();
    delete tcpSocket;
}


void TcpClient::requestNewData()
{
    try{
    if(!enabled) return;

    if(tcpSocket->isValid() &&
            (tcpSocket->state() == QTcpSocket::ConnectedState ||
             tcpSocket->state() == QTcpSocket::ConnectingState) )
        return;
    tcpSocket->abort();
    tcpSocket->connectToHost(hostAddr, hostPort);
    ++connectingAttempts;
    if(tcpSocket->waitForConnected(1000)){
        qInfo() << "TcpClient connected";
    } else {
        if(connectingAttempts<=1){
            qInfo().noquote()
                << tcpSocket->errorString().prepend("TcpClient can't connected.");
            reconnectTimer.start(1000);
        } else
            reconnectTimer.start(5000);
    }
    } catch(...){
        qWarning()<<"Exception in TcpClient::requestNewData()";
    }
}


void TcpClient::readData()
{
    try{
    if(!enabled) return;

    qint64 avail = tcpSocket->bytesAvailable();
    if(!avail)
        return;

    if(!queueFreeSizeSem->tryAcquire()){
        qInfo() << QString("[%1] tcp read delayed (analyzer is slow)")
              .arg(QTime::currentTime().toString("hh:mm:ss:zzz"));
        delayReadTimer.start(200);
        return;
    }

    QVector<QByteArray> dataArray;
    int sumSize=0;
    while(avail){
        QByteArray reciveData = tcpSocket->readAll();
        if(reciveData.size() < avail){
            qWarning() << "We got data lesser that expected.";
        }
        dataArray.push_back(reciveData);
        sumSize += reciveData.size();
        if(sumSize > PACK_SIZE_LIMIT)
            break;

        avail = tcpSocket->bytesAvailable();
//        if(avail){
//            qDebug()<<"double recive";
//        }
    }
//    if(++counter%10==0){
//        qDebug() <<QString("analyzer=%1   writer=%2").arg(analyzerQueueSize->available(),3).arg(writerQueueSize->available(),3);
//    }

    if(dataArray.size()==1){
        emit dataAccuired(dataArray[0]);
    } else {
        QByteArray summedData;
        foreach (QByteArray data, dataArray) {
            summedData.append(data);
        }
        emit dataAccuired(summedData);
    }

    }catch(...){
        qWarning()<<"Exception in TcpClient::readData()";
    }
}


void TcpClient::displayError(QAbstractSocket::SocketError socketError)
{
    switch (socketError) {
    case QAbstractSocket::RemoteHostClosedError:
        qWarning() << "Remote hoste closed connection.";
        break;
    case QAbstractSocket::HostNotFoundError:
        qWarning() << "The host was not found. Please check the "
                    "host name and port settings.";
        break;
    case QAbstractSocket::ConnectionRefusedError:
        if(connectingAttempts<=1){
            qWarning() << tr("The connection was refused by the peer. "
                           "Make sure the fortune server is running, "
                           "and check that the host name and port "
                           "settings are correct.");
        }
        break;
    default:
        qWarning() << tr("The following error occurred: %1.")
                    .arg(tcpSocket->errorString());
    }
}


void TcpClient::setAddress(const QHostAddress &addr, uint port)
{
    if(hostAddr != addr || hostPort != port){
        stopReadOut();
        hostAddr = addr;
        hostPort = port;
        startReadOut();
    }
}


void TcpClient::startReadOut()
{
    qInfo() << QString("[TcpClient] startReadOut from %1:%2")
          .arg(hostAddr.toString()).arg(hostPort);
    enabled = true;
//    requestNewData();
    reconnectTimer.start(0);
}


void TcpClient::stopReadOut()
{
    enabled = false;
    if(tcpSocket->state() != QTcpSocket::UnconnectedState)
        tcpSocket->disconnectFromHost();
    reconnectTimer.stop();
}


void TcpClient::connected()
{
    emit connectedStateChanged(true);
}


void TcpClient::disconnected()
{
    emit connectedStateChanged(false);
    if(enabled){
        qWarning()<<"Client was disconnected, trying to reconnect";
        connectingAttempts=0;
        reconnectTimer.start(1000);
    } else {
        qInfo() << "Client disconnected";
    }
}
