#include "MongoViewerMainWindow.h"
#include "ui_MongoViewerMainWindow.h"
#include "Viewer.h"
#include "lib-common/UIStyles.h"
#include "MongoViewerConfigDB.h"
#include "MongoViewerConfigDispatcher.h"
#include "mongoqt/MongoHost.h"
#include "mongo/MongoConsts.h"
#include "mongo/RootConfigUtil.h"
#include "json-tree-model/JsonTreeModel.h"
#include "base/ConfigSelectorConverter.h"
#include "base/ConfigMetaConverter.h"
#include "base/ConfigMeta.h"
#include "base/ConfigDocument.h"
#include "base/DaqModuleConverter.h"
#include "base/RunIdConverter.h"

#include <QColor>
#include <QDebug>
#include <QDesktopWidget>
#include <QtGui>
#include <QtMath>
#include <limits>
#include <QMetaType>
#include <QMessageBox>
#include <QShortcut>
#include <QStringList>
#include <QSplitter>
#include <QElapsedTimer>
#include <QItemSelectionModel>

const QString SAVED = "SAVED";
const QString HISTORY = "HISTORY";
const QString CLIENTS = "Clients";
const QString CLIENT = "Client";
const QString HEADER = "HEADER";
int const MAX_PAGES_COUNT = 100000;

DaqModule fromString(const QString &s)
{
    auto dm = DaqModule::fromString(s);
    if(dm.isEmpty()) {
        if (s == "TTVXS-RC") return DaqModule(DaqModule::Type::Ttvxs);
    }
    return dm;
}

MongoViewerMainWindow::MongoViewerMainWindow(QWidget *parent) :
    BaseMainWindow(parent),
    ui(new Ui::MongoViewerMainWindow),
    config_dispatcher(new MongoViewerConfigDispatcher()),
    config_db(new MongoViewerConfigDB(this)),
    viewer(new Viewer()),
    split_gorisontal_history(new QSplitter()),
    tree_model(new JsonTreeModel(this)),
    history_navigate_model(new JsonTreeModel(this)),
    main_model(new JsonTreeModel(this))
{
    qRegisterMetaType<MongoCollection>();
    qRegisterMetaType<MongoViewerConfig>();
    AbstractConfigConverter<ConfigDocument>::registerTypes();
    AbstractConfigConverter<DaqModule>::registerTypes();
    AbstractConfigConverter<RcRunId>::registerTypes();
    AbstractConfigConverter<MongoViewerConfig>::registerTypes();
    QTimer::singleShot(0, config_db, &MongoViewerConfigDB::update_config_from_db);

    ui->setupUi(this);
    ui->treeView->setModel(tree_model);
    history_navigate_model->set_key_name("History navigation");
    ui->history_navigate->setModel(history_navigate_model);
    setWindowTitle(QCoreApplication::applicationName());
    setupMenu(ui->menuFile, ui->menuHelp, ui->menuOptions);

    config_dispatcher->subscribe(config_db);
    config_dispatcher->subscribe(viewer);
    config_dispatcher->subscribe(this);

    resize(QDesktopWidget().availableGeometry(this).size() * 0.7);

    new QShortcut(QKeySequence(Qt::CTRL + Qt::Key_R), this,
                  SLOT(on_pushButtonRefresh_clicked()));
    connect(ui->actionQuit, SIGNAL(triggered()), qApp, SLOT(quit()));

    widgets_to_block.append(ui->lineEditPageNumber);
    widgets_to_block.append(ui->lineEditRowsOnPage);
    widgets_to_block.append(ui->groupBoxShowPeriod);
    widgets_to_block.append(ui->dateTimeEditFrom);
    widgets_to_block.append(ui->dateTimeEditTo);
    widgets_to_block.append(ui->groupBoxRunNumber);
    widgets_to_block.append(ui->radioButtonExpand);
    widgets_to_block.append(ui->spinBoxRunNumberFrom);
    widgets_to_block.append(ui->spinBoxRunNumberTo);
    widgets_to_block.append(ui->comboBoxProgramType);
    widgets_to_block.append(ui->comboBoxProgramIndex);
    widgets_to_block.append(ui->comboBoxRunIndex);

    init_gui();
    update_gui();
    refresh_all();

    ui->tableWidgetDB->hide();
    ui->treeViewTest->setModel(main_model);
}

MongoViewerMainWindow::~MongoViewerMainWindow()
{
    emit configChanged(config);
    delete ui;
}

void MongoViewerMainWindow::setConfig(const MongoViewerConfig &newConfig)
{
    if (newConfig == config){
        return;
    }
    config = newConfig;
    update_gui();
}

void MongoViewerMainWindow::init_gui()
{
    set_style_sheet();
    block_widgets_signal(true);

    //table widget db
    ui->tableWidgetDB->setSelectionBehavior(QAbstractItemView::SelectRows);
    ui->tableWidgetDB->setSelectionMode(QAbstractItemView::SingleSelection);

    //--- pushbutton << and >>
    ui->pushButtonLeftPage->setEnabled(false);

    //--- splitters
    QSplitter *split_vertical = new QSplitter;
    QSplitter *split_gorisontal = new QSplitter;

    QVBoxLayout *layout = new QVBoxLayout;

    QWidget *container = new QWidget;
    QVBoxLayout *container_layout = new QVBoxLayout;

    QWidget *container_history = new QWidget;
    QHBoxLayout *container_layout_history = new QHBoxLayout();

    //--- history navigate
    split_gorisontal_history->addWidget(ui->widgetHistoryNavigate);
    split_gorisontal_history->addWidget(ui->widgetTreeView);
    split_gorisontal_history->setOrientation(Qt::Vertical);
    split_gorisontal_history->setCollapsible(0, false);
    split_gorisontal_history->setCollapsible(1, false);
    split_gorisontal_history->setStretchFactor(0,1);
    split_gorisontal_history->setStretchFactor(1,2);

    container_layout_history->addWidget(split_gorisontal_history);
    container_history->setLayout(container_layout_history);

    //----vetical
    split_vertical->addWidget(ui->widgetConfig);
    split_vertical->addWidget(ui->widgetDB);
    split_vertical->addWidget(container_history);

    split_vertical->setCollapsible(0, false);
    split_vertical->setStretchFactor(0,5);
    split_vertical->setStretchFactor(1,10);
    split_vertical->setStretchFactor(2,9);

    container_layout->addWidget(split_vertical);
    container->setLayout(container_layout);

    split_gorisontal->setOrientation(Qt::Vertical);
    split_gorisontal->addWidget(container);
    split_gorisontal->setCollapsible(0, false);
    split_gorisontal->setCollapsible(1, false);
    split_gorisontal->setCollapsible(2, false);
    split_gorisontal->setStretchFactor(0,5);
    split_gorisontal->setStretchFactor(1,2);
    split_gorisontal->setStretchFactor(2,1);

    layout->addWidget(split_gorisontal);
    ui->centralwidget->setLayout(layout);

    //lineedit pages
    ui->lineEditPageNumber->setValidator(new QIntValidator(1, MAX_PAGES_COUNT, this));
    ui->lineEditRowsOnPage->setValidator(new QIntValidator(1, MAX_PAGES_COUNT, this));

    //--- log button
    QPalette Pal(palette());
    Pal.setColor(QPalette::Background, Qt::gray);

    //--- run number groupbox
    ui->spinBoxRunNumberFrom->setMaximum(std::numeric_limits<int>::max());
    ui->spinBoxRunNumberTo->setMaximum(std::numeric_limits<int>::max());

    QSizePolicy size_min;
    size_min.setVerticalPolicy(QSizePolicy::Minimum);
    ui->widgetHistoryNavigate->setSizePolicy(size_min);

    //history navigate
//    hide_history_navigate();

    block_widgets_signal(false);
}

void MongoViewerMainWindow::update_gui()
{
    block_widgets_signal(true);

    //lineedit pages
    ui->lineEditPageNumber->setText(QString::number(config.page_number));
    resize_line_edit_to_content(ui->lineEditPageNumber);
    ui->lineEditRowsOnPage->setText(QString::number(config.rows_on_page));
    resize_line_edit_to_content(ui->lineEditRowsOnPage);

    //--- period groupbox
    ui->groupBoxShowPeriod->setChecked(config.show_period);
    ui->labelPeriodFrom->setDisabled(!config.show_period);
    ui->labelPeriodTo->setDisabled(!config.show_period);
    ui->dateTimeEditFrom->setDisabled(!config.show_period);
    ui->dateTimeEditTo->setDisabled(!config.show_period);
    ui->dateTimeEditFrom->setDateTime(config.period_from);
    ui->dateTimeEditTo->setDateTime(config.period_to);

    //--- run number groupbox
    ui->spinBoxRunNumberFrom->setValue(config.run_number_from);
    ui->spinBoxRunNumberTo->setValue(config.run_number_to);

    //---show run number
    ui->groupBoxRunNumber->setChecked(config.show_run_number);

    //--expand tree
    ui->radioButtonExpand->setChecked(config.expand_tree);
    set_visible_filters();

    block_widgets_signal(false);
}

void MongoViewerMainWindow::hide_history_navigate()
{
    ui->history_navigate->hide();
    ui->checkBoxCloseHN->hide();
    split_gorisontal_history->setSizes({0,1000});
}

void MongoViewerMainWindow::set_style_sheet()
{
    ui->groupBoxRunNumber->setStyleSheet(UIStyles::gbox_fill_blue);
    ui->groupBoxShowPeriod->setStyleSheet(UIStyles::gbox_fill_blue);
    ui->groupBoxPrimaryKeys->setStyleSheet(UIStyles::gbox_fill_blue);
//    ui->groupBoxConfigMeta->setStyleSheet(UIStyles::gbox_fill_blue);
    ui->host_navigate->setStyleSheet(UIStyles::TREE_STYLE_1);
    ui->tableWidgetDB->setStyleSheet(UIStyles::table_style);

    static constexpr const char *widget = (
    "   QTreeView{                          "
    "       border: 1px solid gray;         "
    "       border-radius: 3px;             "
    "       margin-top: 22px;               "
    "   }                                   "
    );
    ui->history_navigate->setStyleSheet(widget);
}

void MongoViewerMainWindow::refresh_all()
{
    config.show_period = false;
    config.show_run_number = false;
    emit configChanged(config);
    viewer->update_hosts();
    update_gui();
    history_navigate_model->clear();
    tree_model->clear();
    clear_qtablewidget();
//    hide_history_navigate();
    populate_host_navigation();
}

void MongoViewerMainWindow::refresh_page()
{
    qInfo() << "=== MongoViewerMainWindow::refresh_page ";
//    history_navigate_model->clear();
//    tree_model->clear();
//    clear_qtablewidget();

//    viewer->update_objects();
//    viewer->update_filters();
//    populate_filters();
//    populate_table_db();
//    set_visible_filters();

    QJsonObject obj;
    QList<ConfigDocument> doc_list = viewer->aggregate();
    qInfo() << "=== " << doc_list.size();
    int local_number = 0;
    for(int i=0; i<doc_list.size(); ++i){
        ConfigDocument doc = doc_list[i];
        QJsonObject tmp = QVariant::fromValue(doc).value<QJsonObject>();
        qInfo() << doc.getConfigOid();
        QString obj_name;
        if(doc.isEmpty()){
            if(config.location == MongoViewerConfig::LOCAL){
                obj_name = QString::number(local_number);
                ++local_number;
            }
            else{
                QJsonObject obj = doc.getConfig()["_id"].toObject();
                ConfigOid oid = QVariant::fromValue(obj).value<ConfigOid>();
                if(oid.isEmpty()){
                    obj_name = QString::number(local_number);
                    ++local_number;
                }
                else
                    obj_name = oid;
            }
        }
        else{
            obj_name = QString("%1 %2")
                    .arg(doc.getConfigSelector().getType().toString(), -20, ' ')
                    .arg(doc.getConfigSelector().getIndex().getIndex());
        }
        obj[obj_name] = tmp;
    }
    main_model->loadJson(obj);

//    qInfo().noquote().nospace() << QJsonDocument(tmp).toJson();
}

void MongoViewerMainWindow::populate_filters()
{
    qInfo() << __PRETTY_FUNCTION__;
    block_widgets_signal(true);
    ui->comboBoxProgramType->clear();
    ui->comboBoxRunIndex->clear();
    ui->comboBoxProgramType->addItem("All daq modules");
    ui->comboBoxRunIndex->addItem("All run indexes");

    QStringList::iterator i;
    QStringList pt = viewer->all_program_types;
    for(i=pt.begin(); i!=pt.end(); ++i){
        ui->comboBoxProgramType->addItem(*i);
    }
    QStringList pi = viewer->all_program_indexes;
    for(i=pi.begin(); i!=pi.end(); ++i){
        ui->comboBoxProgramIndex->addItem(*i);
    }
    QStringList ri = viewer->all_run_indexes;
    for(i=ri.begin(); i!=ri.end(); ++i){
        ui->comboBoxRunIndex->addItem(*i);
    }
    QList<int> rn = viewer->all_run_numbers;
    std::sort(rn.begin(), rn.end());
    if(!rn.isEmpty()){
        ui->spinBoxRunNumberFrom->setValue(rn.first());
        ui->spinBoxRunNumberTo->setValue(rn.last());
    }
}

void MongoViewerMainWindow::populate_host_navigation()
{
    QBrush red(QColor(230,57,0));
    QBrush green(QColor(0,204,68));

    bool oldState = ui->host_navigate->blockSignals(true);
    ui->host_navigate->clear();
    QList<QTreeWidgetItem *> top_level_items;

    QTreeWidgetItem * local_saved = new QTreeWidgetItem();
    local_saved->setText(0, "Local saved");
    local_saved->setData(0, Qt::UserRole, SAVED);
    top_level_items.append(local_saved);

    QTreeWidgetItem * history_saved = new QTreeWidgetItem();
    history_saved->setText(0, "Local history");
    history_saved->setData(0, Qt::UserRole, HISTORY);
    top_level_items.append(history_saved);

    for(int i=0; i<viewer->hosts.size(); ++i){
        QTreeWidgetItem * host_item = new QTreeWidgetItem();
        host_item->setText(0, QString("%1 (%2)")
            .arg(QUrl(viewer->hosts[i].get_uri_string()).host())
            .arg(viewer->hosts[i].environment));
        host_item->setData(0, Qt::UserRole, QString::number(i) +
           "|"+ QString::number(-1) +
           "|" + QString::number(-1));
        viewer->hosts[i].is_connect ? host_item->setBackground(0, green) :
                                      host_item->setBackground(0, red);
        for(int k=0; k<viewer->hosts[i].databeses.size(); ++k){
            QTreeWidgetItem * databeses_item = new QTreeWidgetItem();
            databeses_item->setText(0, QString("%1 (%2 Kb)")
                .arg(viewer->hosts[i].databeses[k].db_name)
                .arg(QString::number(viewer->hosts[i].databeses[k].size/1024)));
            databeses_item->setData(0, Qt::UserRole, QString::number(i) +
                "|"+ QString::number(k) +
                "|" + QString::number(-1));
            QList<MongoCollection> colections = viewer->hosts[i].databeses[k].collections;
            for(int l=0; l<colections.size(); ++l){
                QTreeWidgetItem * collection_item = new QTreeWidgetItem();
                collection_item->setText(0, colections[l].name);
                collection_item->setData(0, Qt::UserRole, QString::number(i) +
                    "|" + QString::number(k) +
                    "|" + QString::number(l));
                databeses_item->addChild(collection_item);
            }
            host_item->addChild(databeses_item);
        }
        top_level_items.append(host_item);
    }

    ui->host_navigate->addTopLevelItems(top_level_items);
    ui->host_navigate->blockSignals(oldState);
    ui->host_navigate->resizeColumnToContents(0);
    ui->host_navigate->expandAll();
}

void MongoViewerMainWindow::set_visible_filters()
{
    bool isSaved = config.type == MongoViewerConfig::SAVED;
    bool isHistory = config.type == MongoViewerConfig::HISTORY;
    //---program type
    ui->comboBoxProgramType->setHidden(!isHistory && !isSaved);
    ui->labelProgramType->setHidden(!isHistory && !isSaved);
    //---program index
    ui->comboBoxProgramIndex->setHidden(!isSaved);
    ui->labelProgramIndex->setHidden(!isSaved);
    //---run index
    ui->comboBoxRunIndex->setHidden(!isHistory);
    ui->labelRunIndex->setHidden(!isHistory);
    //---run number
    ui->groupBoxRunNumber->setHidden(!isHistory);

    //---period
    ui->groupBoxShowPeriod->setHidden(!isHistory);
    ui->groupBoxPrimaryKeys->setHidden(!isHistory && !isSaved);

    //---remove record
    ui->pushButtonDelete->setHidden(!isSaved);
}

void MongoViewerMainWindow::populate_filters_history()
{
//    block_widgets_signal(true);
//    ui->comboBoxProgramType->clear();
//    ui->comboBoxRunIndex->clear();
//    ui->comboBoxProgramType->addItem("All program types");
//    ui->comboBoxRunIndex->addItem("All run indexes");
//    QList<QString> program_types;
//    for(int i=0; i< viewer->filters_history.size(); ++i){
//        if(!program_types.contains(viewer->filters_history[i].first)){
//            program_types.append(viewer->filters_history[i].first);
//        }
//    }

//    for(int i=0; i<program_types.size(); ++i){
//        ui->comboBoxProgramType->addItem(program_types[i]);
//    }

//    if(config.program_type_filter.isEmpty() ||
//       !program_types.contains(config.program_type_filter)){

//        ui->comboBoxProgramType->setCurrentIndex(0);
//        config.program_type_filter.clear();
//        emit configChanged(config);
//    }
//    else{
//        for(int i=0; i<ui->comboBoxProgramType->count(); ++i){
//            if(ui->comboBoxProgramType->itemText(i) == config.program_type_filter){
//                ui->comboBoxProgramType->setCurrentIndex(i);
//            }
//        }
//    }

//    QList<QString> run_indexes;
//    if(config.program_type_filter.isEmpty()){
//        for(int i=0; i<viewer->filters_history.size(); ++i){
//            if(!run_indexes.contains(viewer->filters_history[i].second)){
//                run_indexes.append(viewer->filters_history[i].second);
//            }
//        }
//    }
//    else{
//        for(int i=0; i<viewer->filters_history.size(); ++i){
//            if(!run_indexes.contains(viewer->filters_history[i].second) &&
//              (viewer->filters_history[i].first == config.program_type_filter)){
//                run_indexes.append(viewer->filters_history[i].second);
//            }
//        }
//    }

//    for(int i=0; i<run_indexes.size(); ++i){
//        ui->comboBoxRunIndex->addItem(run_indexes[i]);
//    }

//    if(config.run_index_filter.isEmpty() ||
//            !run_indexes.contains(config.run_index_filter)){

//        ui->comboBoxRunIndex->setCurrentIndex(0);
//        config.run_index_filter.clear();
//        emit configChanged(config);
//    }
//    else{
//        for(int i=0; i<ui->comboBoxRunIndex->count(); ++i){
//            if(ui->comboBoxRunIndex->itemText(i) == config.run_index_filter){
//                ui->comboBoxRunIndex->setCurrentIndex(i);
//            }
//        }
//    }

//    ui->comboBoxProgramType->setSizeAdjustPolicy(QComboBox::AdjustToContents);
//    ui->comboBoxRunIndex->setSizeAdjustPolicy(QComboBox::AdjustToContents);

//    block_widgets_signal(false);
    return;
}

void MongoViewerMainWindow::populate_filters_saved()
{
//    block_widgets_signal(true);

//    ui->comboBoxProgramType->clear();
//    ui->comboBoxProgramIndex->clear();
//    ui->comboBoxProgramType->addItem("All program types");
//    ui->comboBoxProgramIndex->addItem("All program indexes");

//    QList<QString> program_types;
//    for(int i=0; i< viewer->filters_saved.size(); ++i){
//        if(!program_types.contains(viewer->filters_saved[i].first)){
//            program_types.append(viewer->filters_saved[i].first);
//        }
//    }

//    for(int i=0; i<program_types.size(); ++i){
//        ui->comboBoxProgramType->addItem(program_types[i]);
//    }

//    if(config.program_type_filter.isEmpty() ||
//       !program_types.contains(config.program_type_filter)){

//        ui->comboBoxProgramType->setCurrentIndex(0);
//        config.program_type_filter.clear();
//    }
//    else{
//        for(int i=0; i<ui->comboBoxProgramType->count(); ++i){
//            if(ui->comboBoxProgramType->itemText(i) == config.program_type_filter){
//                ui->comboBoxProgramType->setCurrentIndex(i);
//            }
//        }
//    }

//    QList<QString> program_indexes;
//    if(config.program_type_filter.isEmpty()){
//        for(int i=0; i<viewer->filters_saved.size(); ++i){
//            if(!program_indexes.contains(viewer->filters_saved[i].second)){
//                program_indexes.append(viewer->filters_saved[i].second);
//            }
//        }
//    }
//    else{
//        for(int i=0; i<viewer->filters_saved.size(); ++i){
//            if(!program_indexes.contains(viewer->filters_saved[i].second) &&
//              (viewer->filters_saved[i].first == config.program_type_filter)){
//                program_indexes.append(viewer->filters_saved[i].second);
//            }
//        }
//    }

//    for(int i=0; i<program_indexes.size(); ++i){
//        ui->comboBoxProgramIndex->addItem(program_indexes[i]);
//    }

//    if(config.program_index_filter.isEmpty() ||
//            !program_indexes.contains(config.program_index_filter)){

//        ui->comboBoxProgramIndex->setCurrentIndex(0);
//        config.program_index_filter.clear();
//    }
//    else{
//        for(int i=0; i<ui->comboBoxProgramIndex->count(); ++i){
//            if(ui->comboBoxProgramIndex->itemText(i) == config.program_index_filter){
//                ui->comboBoxProgramIndex->setCurrentIndex(i);
//            }
//        }
//    }

//    emit configChanged(config);
//    block_widgets_signal(false);

//    ui->comboBoxProgramType->setSizeAdjustPolicy(QComboBox::AdjustToContents);
//    ui->comboBoxProgramIndex->setSizeAdjustPolicy(QComboBox::AdjustToContents);
    return;
}

void MongoViewerMainWindow::populate_table_db()
{
    QTableWidget * table = ui->tableWidgetDB;
    table->setRowCount(0);

    int record_size = 0;
    if(config.type==MongoViewerConfig::SAVED)
        record_size = viewer->cs_objects.size();
    else
        record_size = viewer->ri_objects.size();
    if(!(record_size==0 && (config.page_number == 1))){
        table->setRowCount(record_size);
        table->setColumnCount(5);
    }
    else{
        table->setRowCount(1);
        QTableWidgetItem * item_id = new QTableWidgetItem("");
        QTableWidgetItem * item_date = new QTableWidgetItem("EMPTY COLLECTION");
        table->setItem(0, 0, item_id);
        table->setItem(0, 1, item_date);
        table->setColumnCount(2);
    }

    for(int k=0; k<record_size; ++k){
//        QJsonObject obj = viewer->objects[k];
        QJsonObject obj;
        RcRunId ri_key;
        ConfigSelector cs_key;
        if(config.type==MongoViewerConfig::SAVED) {
            cs_key = viewer->cs_objects.keys().value(k);
            obj = viewer->cs_objects.value(cs_key);
        } else {
            ri_key = viewer->ri_objects.keys().value(k);
            obj = viewer->ri_objects.value(ri_key);
        }
        int column_number = 0;

//        if(!config.selected_collection.isLocal()){
//            table->setHorizontalHeaderItem(column_number, new QTableWidgetItem("id"));
//            QTableWidgetItem * item_id = new QTableWidgetItem(oid);
//            item_id->setFlags(item_id->flags() ^ Qt::ItemIsEditable);
//            table->setItem(k, column_number, item_id);
//            ++column_number;
//        }
        QString oid = get_oid(obj);
        if(oid.isEmpty())
            oid = QString::number(k);

        if(config.type == MongoViewerConfig::HISTORY){
            table->setHorizontalHeaderItem(
                column_number,
                new QTableWidgetItem("Program type"));
            QTableWidgetItem * item_program_type = new QTableWidgetItem(
                obj[program_type].toString());
            item_program_type->setFlags(item_program_type->flags() ^ Qt::ItemIsEditable);
            table->setItem(k, column_number, item_program_type);
            ++column_number;

            table->setHorizontalHeaderItem(
                column_number,
                new QTableWidgetItem("Date time of modification"));
            QTableWidgetItem * item_date = new QTableWidgetItem(
                RootConfigUtil::getModificationTimestamp(obj).toString(Qt::ISODate));
            item_date->setFlags(item_date->flags() ^ Qt::ItemIsEditable);
            table->setItem(k, column_number, item_date);
            ++column_number;

            table->setHorizontalHeaderItem(
                column_number,
                new QTableWidgetItem("Run index"));
            QTableWidgetItem * item_run_index = new QTableWidgetItem(
                obj[run_index].toString());
            item_run_index->setFlags(item_run_index->flags() ^ Qt::ItemIsEditable);
            table->setItem(k, column_number, item_run_index);
            ++column_number;

            table->setHorizontalHeaderItem(
                column_number,
                new QTableWidgetItem("Run number"));
            QTableWidgetItem * item_run_number = new QTableWidgetItem(
                QString::number(obj[run_number].toInt()));
            item_run_number->setFlags(item_run_number->flags() ^ Qt::ItemIsEditable);
            table->setItem(k, column_number, item_run_number);
            ++column_number;

        }
        if(config.type == MongoViewerConfig::SAVED){
            table->setHorizontalHeaderItem(column_number, new QTableWidgetItem("Program type"));
            QTableWidgetItem * item_program_type = new QTableWidgetItem(cs_key.getType().toString());
            item_program_type->setFlags(item_program_type->flags() ^ Qt::ItemIsEditable);
            table->setItem(k, column_number, item_program_type);
            ++column_number;

            table->setHorizontalHeaderItem(column_number, new QTableWidgetItem("Date time of modification"));
            QTableWidgetItem * item_date = new QTableWidgetItem(obj[k_date_of_modification].toString());
            item_date->setFlags(item_date->flags() ^ Qt::ItemIsEditable);
            table->setItem(k, column_number, item_date);
            ++column_number;

            table->setHorizontalHeaderItem(column_number, new QTableWidgetItem("Program index"));
            QTableWidgetItem * item_program_index = new QTableWidgetItem(cs_key.getIndex().getIndex());
            item_program_index->setFlags(item_program_index->flags() ^ Qt::ItemIsEditable);
            table->setItem(k, column_number, item_program_index);
            ++column_number;

            table->setHorizontalHeaderItem(column_number, new QTableWidgetItem("Configuration name"));
            QTableWidgetItem * item_configuration_name = new QTableWidgetItem(cs_key.getConfigurationName());
            item_configuration_name->setFlags(item_configuration_name->flags() ^ Qt::ItemIsEditable);
            table->setItem(k, column_number, item_configuration_name);
            ++column_number;
        }
        else{
            table->setHorizontalHeaderItem(column_number, new QTableWidgetItem("Id"));
            QTableWidgetItem * item_oid = new QTableWidgetItem(oid);
            table->setItem(k, column_number, item_oid);
            ++column_number;
        }

        table->setColumnCount(column_number);

        if(k==0){
            table->selectRow(k);
            viewer->set_object(0);
            viewer->update_filters();
        }
    }

    QStringList vertical_header_labels;
    for(int i=1; i<=config.rows_on_page; ++i){
        vertical_header_labels << QString::number(config.rows_on_page*(config.page_number-1) + i);
    }
    table->setVerticalHeaderLabels(vertical_header_labels);
    table->horizontalHeader()->setSectionResizeMode(0, QHeaderView::ResizeToContents);
    table->resizeColumnsToContents();
}

void MongoViewerMainWindow::populate_history_navigate()
{
    QTreeView * view   = ui->history_navigate;
    history_navigate_model->loadJson(history_navigation);
    view->show();
    ui->checkBoxCloseHN->show();
    ui->checkBoxCloseHN->setChecked(false);
    view->expandAll();
    view->resizeColumnToContents(0);

    connect(ui->history_navigate->selectionModel(),
            SIGNAL(selectionChanged(QItemSelection, QItemSelection)),
            this,
            SLOT(history_navigate_item_changed(QItemSelection, QItemSelection)));

    if(history_navigate_model->rowCount() !=0)
        view->setCurrentIndex(history_navigate_model->index(0,0));
    history_map_key = history_navigate_model->index(0,0).data(Qt::DisplayRole).toString();
}

void MongoViewerMainWindow::clear_qtablewidget()
{
    QTableWidget * table = ui->tableWidgetDB;
    table->blockSignals(true);
    for(int i=0; i<table->columnCount(); ++i){
        table->removeColumn(i);
    }
    table->setRowCount(0);
    table->blockSignals(false);
}

void MongoViewerMainWindow::populate_tree_viewer(QJsonObject obj)
{
    QTreeView * view   = ui->treeView;
    obj = readble_dates(obj);
    tree_model->loadJson(obj);
    view->show();
    if(config.expand_tree){
        view->expandAll();
    }
    else{
        view->collapseAll();
    }
    view->resizeColumnToContents(0);
}

void MongoViewerMainWindow::split_history()
{
    history_navigation = QJsonObject();
    history_map.clear();

    QJsonObject obj = viewer->selected_obj;
    QJsonObject without_head;
    QString main_key;
    for(const auto &key : obj.keys()){
        if(obj[key].isObject()){
            auto daqModule = fromString(key.split("_").first());
            if(!daqModule.isEmpty()){
                without_head[key] = obj[key].toObject();
                obj.remove(key);
                main_key = key;
                break;
            }
        }
    }
    obj.remove(main_key);
    QJsonObject head = obj;

    if(main_key.isEmpty()){
        history_navigation = obj;
        history_navigation = readble_dates(history_navigation);
        return;
    }

    history_navigation[HEADER] = QJsonObject();
    history_map[HEADER] = head;
    history_navigation[main_key] = history_navigation_recurse(without_head);
    history_navigation = readble_dates(history_navigation);
    history_map_recurse(without_head);
}

QJsonObject MongoViewerMainWindow::history_navigation_recurse(QJsonObject obj)
{
    RunMap rm;
    QJsonObject to_return;
    QString main_key = obj.keys().first();
    QJsonObject obj_ = obj[main_key].toObject();
    for(const auto &key : obj_.keys()){
        if((key == CLIENTS) || (key.split("_").first() == CLIENT)){
            QJsonObject child = obj_[key].toObject();
            for(const auto &child_key : child.keys()){
                auto daqModule = fromString(child_key.split("_").first());
                if(!daqModule.isEmpty()){
                    QJsonObject grandson = child[child_key].toObject();
                    QJsonObject to_recurse;
                    to_recurse[child_key] = grandson;
                    grandson = history_navigation_recurse(to_recurse);
                    to_return[child_key] = grandson;
                }
            }
        }
    }
    return to_return;
}

void MongoViewerMainWindow::history_map_recurse(QJsonObject obj)
{
    QJsonObject to_history_map = obj;
    QString main_key = obj.keys().first();

    for(const auto &key : to_history_map[main_key].toObject().keys()){
        if((key == CLIENTS) || (key.split("_").first() == CLIENT)){
            QJsonObject child = to_history_map[main_key].toObject()[key].toObject();
            for(const auto &child_key : child.keys()){
                auto daqModule = fromString(child_key.split("_").first());
                if(!daqModule.isEmpty()){
                    QJsonObject grandson = child[child_key].toObject();
                    QJsonObject to_recurse;
                    to_recurse[child_key] = grandson;
                    history_map_recurse(to_recurse);
                }
            }
            QJsonObject tt = to_history_map[main_key].toObject();
            tt.remove(key);
            to_history_map[main_key] = tt;
        }
    }
    history_map[main_key] = to_history_map;
    return;
}

void MongoViewerMainWindow::resize_line_edit_to_content(QLineEdit * line_edit)
{
    QString text = line_edit->text();
    QFont font("", 0);
    QFontMetrics fm(font);
    int pixelsWide = fm.width(text) + 10;
    line_edit->setFixedSize(pixelsWide, line_edit->height());
}

void MongoViewerMainWindow::check_left_right_buttons()
{
    if(config.page_number == 1){
         ui->pushButtonLeftPage->setEnabled(false);
    }
    else{
        ui->pushButtonLeftPage->setEnabled(true);
    }
}

void MongoViewerMainWindow::change_refresh_button_color(QColor color)
{
    QPalette pal = ui->pushButtonRefresh->palette();
    pal.setColor(QPalette::Button, color);
    ui->pushButtonRefresh->setAutoFillBackground(true);
    ui->pushButtonRefresh->setPalette(pal);
    ui->pushButtonRefresh->update();
}

void MongoViewerMainWindow::block_widgets_signal(bool block)
{
    for(int i=0; i<widgets_to_block.size(); ++i){
        widgets_to_block[i]->blockSignals(block);
    }
}

QJsonObject MongoViewerMainWindow::readble_dates(QJsonObject obj)
{
    QJsonObject to_return;
    to_return = obj;

    for(const auto &key : obj.keys()){
        if(key == $date){
            QVariant value = to_return[$date].toVariant();
            QDateTime date;
            date.setMSecsSinceEpoch(value.toLongLong());
            QString date_str = date.toString(Qt::ISODate);
            to_return[$date] = date_str;
        }
        if(obj[key].isObject())
            to_return[key] = readble_dates(obj[key].toObject());
    }
    return to_return;
}

QString MongoViewerMainWindow::get_oid(QJsonObject &obj)
{
    if(obj.keys().contains(_id)){
        if(obj[_id].isString())
            return obj[_id].toString();
        if(obj[_id].isObject())
            if(obj[_id].toObject().keys().contains($oid))
                return obj[_id].toObject()[$oid].toString();
    }
    return QString();
}


//--- Slots from UI
void MongoViewerMainWindow::on_host_navigate_itemDoubleClicked(QTreeWidgetItem *item, int column)
{
    QString item_data = item->data(0, Qt::UserRole).toString();
    qInfo() << __PRETTY_FUNCTION__ << " -------------- " << item_data;

    if(item_data == SAVED){
        config.type = MongoViewerConfig::SAVED;
        config.location = MongoViewerConfig::LOCAL;
        emit configChanged(config);
        refresh_page();
        return;
    }
    if(item_data == HISTORY){
        config.type = MongoViewerConfig::HISTORY;
        config.location = MongoViewerConfig::LOCAL;
        emit configChanged(config);
        refresh_page();
        return;
    }

    int host_number = item_data.split("|").at(0).toInt();
    int db_number = item_data.split("|").at(1).toInt();
    int col_number = item_data.split("|").at(2).toInt();

    if(col_number != -1){
        config.selected_host = viewer->hosts[host_number];
        config.selected_db = viewer->hosts[host_number].databeses[db_number];
        config.selected_col = viewer->hosts[host_number].databeses[db_number].collections[col_number];
        config.location = MongoViewerConfig::MONGO;
        if(config.selected_col.isHistory())
            config.type = MongoViewerConfig::HISTORY;
        if(config.selected_col.isSaved())
            config.type = MongoViewerConfig::SAVED;
        emit configChanged(config);
        refresh_page();
    }
}

void MongoViewerMainWindow::on_tableWidgetDB_itemSelectionChanged()
{
    qInfo() << __PRETTY_FUNCTION__ << " --------------";
    int k=0;
    QModelIndexList selection = ui->tableWidgetDB->selectionModel()->selectedRows();
    for(int i=0; i< selection.count(); i++)
    {
        QModelIndex index = selection.at(i);
        k = index.row();
    }

    viewer->set_object(k);
    viewer->update_filters();
    populate_filters();
    if(config.type == MongoViewerConfig::HISTORY){
        split_history();
        populate_history_navigate();
    }
    else{
        populate_tree_viewer(viewer->selected_obj);
    }

    //    if(config.type == Config::HISTORY){
    //        populate_filters_history();
    //        hide_history_navigate();
    //    }

    //    if(config.type == Config::SAVED){
    //        populate_filters_saved();
    //    }
}

void MongoViewerMainWindow::on_pushButtonRefresh_clicked()
{
    refresh_all();
}

void MongoViewerMainWindow::on_groupBoxShowPeriod_toggled(bool checked)
{
    ui->labelPeriodFrom->setDisabled(!checked);
    ui->labelPeriodTo->setDisabled(!checked);
    ui->dateTimeEditFrom->setDisabled(!checked);
    ui->dateTimeEditTo->setDisabled(!checked);
    config.show_period = checked;
    emit configChanged(config);
    refresh_page();
}

void MongoViewerMainWindow::on_pushButtonLeftPage_clicked()
{
    config.page_number = config.page_number - 1;
    bool oldState = ui->lineEditPageNumber->blockSignals(true);
    emit configChanged(config);
    ui->lineEditPageNumber->setText(QString::number(config.page_number));
    ui->lineEditPageNumber->blockSignals(oldState);
    resize_line_edit_to_content(ui->lineEditPageNumber);
    check_left_right_buttons();
    refresh_page();
}

void MongoViewerMainWindow::on_pushButtonRightPage_clicked()
{
    config.page_number = config.page_number + 1;
    bool oldState = ui->lineEditPageNumber->blockSignals(true);
    emit configChanged(config);
    ui->lineEditPageNumber->setText(QString::number(config.page_number));
    ui->lineEditPageNumber->blockSignals(oldState);
    resize_line_edit_to_content(ui->lineEditPageNumber);
    check_left_right_buttons();
    refresh_page();
}


void MongoViewerMainWindow::on_lineEditPageNumber_textChanged(const QString &arg1)
{
    config.page_number = arg1.toInt();
    resize_line_edit_to_content(ui->lineEditPageNumber);
}

void MongoViewerMainWindow::on_lineEditPageNumber_editingFinished()
{
    emit configChanged(config);
    refresh_page();
}

void MongoViewerMainWindow::on_lineEditRowsOnPage_textChanged(const QString &arg1)
{
    config.rows_on_page = arg1.toInt();
    resize_line_edit_to_content(ui->lineEditRowsOnPage);
}

void MongoViewerMainWindow::on_lineEditRowsOnPage_editingFinished()
{
    emit configChanged(config);
    refresh_page();
}

void MongoViewerMainWindow::on_comboBoxProgramType_currentIndexChanged(const QString &arg1)
{
    if(ui->comboBoxProgramType->currentIndex() == 0){
        config.program_type_filter.clear();
    }
    else{
        config.program_type_filter = arg1;
    }
    emit configChanged(config);
    refresh_page();
}

void MongoViewerMainWindow::on_comboBoxProgramIndex_currentIndexChanged(const QString &arg1)
{
    if(ui->comboBoxProgramIndex->currentIndex() == 0){
        config.program_index_filter.clear();
    }
    else{
        config.program_index_filter = arg1;
    }
    emit configChanged(config);
    refresh_page();
}

void MongoViewerMainWindow::history_navigate_item_changed(QItemSelection item, QItemSelection prev)
{
    if(item.indexes().size() != 0){
        QString name = item.indexes().first().data(Qt::DisplayRole).toString();
        populate_tree_viewer(history_map[name]);
    }
}

void MongoViewerMainWindow::on_groupBoxRunNumber_toggled(bool checked)
{
    config.show_run_number = checked;
    emit configChanged(config);
    refresh_page();
}

void MongoViewerMainWindow::on_spinBoxRunNumberFrom_valueChanged(int arg1)
{
    config.run_number_from = arg1;
}

void MongoViewerMainWindow::on_spinBoxRunNumberTo_valueChanged(int arg1)
{
    config.run_number_to = arg1;
}

void MongoViewerMainWindow::on_spinBoxRunNumberFrom_editingFinished()
{
    emit configChanged(config);
    refresh_page();
}

void MongoViewerMainWindow::on_spinBoxRunNumberTo_editingFinished()
{
    emit configChanged(config);
    refresh_page();
}

void MongoViewerMainWindow::on_radioButtonExpand_toggled(bool checked)
{
    config.expand_tree = checked;
    emit configChanged(config);
    if(checked){
        ui->treeView->expandAll();
    }
    else{
        ui->treeView->collapseAll();
    }
}

void MongoViewerMainWindow::on_pushButtonDelete_clicked()
{
//    viewer->delete_config();
//    refresh_page();
    config.rows_on_page++;
    emit configChanged(config);
}

void MongoViewerMainWindow::on_dateTimeEditFrom_dateTimeChanged(const QDateTime &dateTime)
{
    config.period_from = dateTime;
    emit configChanged(config);
    refresh_page();
}

void MongoViewerMainWindow::on_dateTimeEditTo_dateTimeChanged(const QDateTime &dateTime)
{
    config.period_to = dateTime;
    emit configChanged(config);
    refresh_page();
}

void MongoViewerMainWindow::on_checkBoxCloseHN_toggled(bool checked)
{
    if(checked){
        ui->history_navigate->hide();
        split_gorisontal_history->setSizes({0,1000});
        populate_tree_viewer(viewer->selected_obj);
    }
    else{
        ui->history_navigate->show();
        populate_history_navigate();
   }
}

void MongoViewerMainWindow::on_actionShow_all_collections_toggled(bool arg1)
{
    config.show_all_collections = arg1;
    refresh_all();
}

void MongoViewerMainWindow::on_pushButton_3_clicked()
{
//    config_db->write_hist(config);
}

void MongoViewerMainWindow::on_pushButton_clicked()
{
    config_db->update_config_from_db();
}
