#include "EvbGenConfigForm.h"
#include "ui_EvbGenConfigForm.h"
#include "EvbGenConfigDispatcher.h"

#include "base/gui_utils.h"

EvbGenConfigForm::EvbGenConfigForm(QWidget *parent)
    : QWidget(parent),
    ui(new Ui::EvbGenConfigForm)
{
    ui->setupUi(this);
    updateGuiToConfig();
    auto &configDispatcher = EvbGenConfigDispatcher::instance();
    configDispatcher.subscribeEvbGenConfig(this);
}

EvbGenConfigForm::~EvbGenConfigForm()
{
    delete ui;
}

void EvbGenConfigForm::setEvbGenConfig(const EvbGenConfig &newConfig)
{
    if(config == newConfig)
        return;
    config = newConfig;
    updateGuiToConfig();
}

void EvbGenConfigForm::on_spinBoxClients_valueChanged(int val)
{
    if(config.numberOfClients == val)
        return;
    config.numberOfClients = val;
    emit evbGenConfigChanged(config);
}

void EvbGenConfigForm::on_spinBoxPreRunWait_valueChanged(int val)
{
    if(config.cfgBase.preRunWait == val)
        return;
    config.cfgBase.preRunWait = val;
    emit evbGenConfigChanged(config);
    updateLabels();
}

void EvbGenConfigForm::on_spinBoxSpillCount_valueChanged(int val)
{
    if(config.cfgBase.spill_count == val)
        return;
    config.cfgBase.spill_count = val;
    emit evbGenConfigChanged(config);
    updateLabels();
}

void EvbGenConfigForm::on_spinBoxSpillTime_valueChanged(int val)
{
    if(config.cfgBase.spillMs == val)
        return;
    config.cfgBase.spillMs = val;
    emit evbGenConfigChanged(config);
    updateLabels();
}

void EvbGenConfigForm::on_spinBoxInterspillTime_valueChanged(int val)
{
    if(config.cfgBase.interspillMs == val)
        return;
    config.cfgBase.interspillMs = val;
    emit evbGenConfigChanged(config);
    updateLabels();
}

void EvbGenConfigForm::on_doubleSpinBoxEvRate_valueChanged(double val)
{
    if(config.cfgBase.evRate == val)
        return;
    config.cfgBase.evRate = val*1e3;

    emit evbGenConfigChanged(config);
    updateLabels();
}

void EvbGenConfigForm::on_spinBoxEvPayload_valueChanged(int val)
{
    if(config.cfgBase.event_length == val)
        return;
    config.cfgBase.event_length = val;
    emit evbGenConfigChanged(config);
    updateLabels();
    updatePayloadLimits();
}

void EvbGenConfigForm::on_checkBoxPayloadMax_toggled(bool useMax)
{
    if(config.cfgBase.useMax == useMax)
        return;
    config.cfgBase.useMax = useMax;
    if(useMax && config.cfgBase.eventLengthMax < config.cfgBase.event_length) {
        config.cfgBase.eventLengthMax = config.cfgBase.event_length;
        ui->spinBoxEvPayloadMax->setValue(config.cfgBase.eventLengthMax);
    }
    updatePayloadLimits();
    ui->spinBoxEvPayloadMax->setEnabled(useMax);
    emit evbGenConfigChanged(config);
    updateLabels();
}

void EvbGenConfigForm::on_spinBoxEvPayloadMax_valueChanged(int val)
{
    if(config.cfgBase.eventLengthMax == val)
        return;
    config.cfgBase.eventLengthMax = val;
    emit evbGenConfigChanged(config);
    updateLabels();
    updatePayloadLimits();
}

void EvbGenConfigForm::updateGuiToConfig()
{
    RecursiveWidgetSignalBlocker b(this);
    Q_UNUSED(b)
    ui->spinBoxClients->setValue(config.numberOfClients);

    ui->doubleSpinBoxEvRate->setValue(config.cfgBase.evRate/1e3);

    ui->spinBoxPreRunWait->setValue(config.cfgBase.preRunWait);
    ui->spinBoxSpillCount->setValue(config.cfgBase.spill_count);
    ui->spinBoxSpillTime->setValue(config.cfgBase.spillMs);
    ui->spinBoxInterspillTime->setValue(config.cfgBase.interspillMs);

    updatePayloadLimits();
    ui->spinBoxEvPayload->setValue(config.cfgBase.event_length);
    ui->checkBoxPayloadMax->setChecked(config.cfgBase.useMax);
    ui->spinBoxEvPayloadMax->setEnabled(config.cfgBase.useMax);
    ui->spinBoxEvPayloadMax->setValue(config.cfgBase.eventLengthMax);

    updateLabels();
}

void EvbGenConfigForm::updateLabels()
{
    quint64 runMs = config.cfgBase.spillMs + config.cfgBase.interspillMs;
    runMs *= config.cfgBase.spill_count;
    runMs += config.cfgBase.preRunWait;

    ui->labelRunTimeVal->setText(QString("%L1:%2:%3.%4")
                                     .arg(runMs/3600000)
                                     .arg((runMs/60000)%60, 2, 10, QChar('0'))
                                     .arg((runMs/1000)%60, 2, 10, QChar('0'))
                                     .arg(runMs%1000, 3, 10, QChar('0'))
                                 );
    double evLenKB = config.cfgBase.event_length;
    evLenKB += 4*(2+1+2); // MpdHdr+evNum+MpdDeviceHdr
    evLenKB /= 1e3;
    double spillKB = evLenKB * config.cfgBase.getEvCntPerSpill();
    auto maxDataRate = spillKB / config.cfgBase.spillMs;
    auto avrDataRate = spillKB / (config.cfgBase.spillMs+config.cfgBase.interspillMs);
    ui->labelDataRateVal->setText(QString("%L1/%L2 MB/s")
                                      .arg(avrDataRate, 0, 'f', 3)
                                      .arg(maxDataRate, 0, 'f', 3));
}

void EvbGenConfigForm::updatePayloadLimits() const
{
    ui->spinBoxEvPayload->setMaximum(config.cfgBase.useMax ? config.cfgBase.eventLengthMax : 1000000);
    ui->spinBoxEvPayloadMax->setMinimum(config.cfgBase.useMax ? config.cfgBase.event_length : 0);
}
