#include "AdjustBufferModel.h"

#include <QMap>
#include <optional>
#include <QColor>

#include "app/AppStyle.h"
#include "base/ClientIndex.h"
#include "base/ClientIndexEnabled.h"
#include "EvBConfig.h"
#include "EventSizePeriod.h"
#include "SizeHelper.h"

//namespace {
struct ClientBufStatus
{
    std::optional<quint64> curSize; // from TcpClient after resize
    std::optional<quint64> prefSize; // from Alanyzer after next EvSizePeriod
};

class AdjustBufferMap
{
public:
    int rowOf(const ClientIndex &deviceIndex) const;
    ClientIndex indexOf(int row) const;
    ClientBufStatus getCurSize(const ClientIndex i) const;

    EvBConfig cfg;
    QMap<ClientIndex, ClientBufStatus> status;
    ClientBufStatus combinerStatus;
};

int AdjustBufferMap::rowOf(const ClientIndex &i) const
{
    auto v = std::distance(cfg.clients.begin(), cfg.clients.find(i));
    ++v; // Combiner on first row
    return v;
}

ClientIndex AdjustBufferMap::indexOf(int row) const
{
    --row; // Combiner on first row
    int r = 0;
    for (auto it = cfg.clients.begin(); it != cfg.clients.end(); ++it)
        if (r++ == row)
            return it.key();
    return {};
}

ClientBufStatus AdjustBufferMap::getCurSize(const ClientIndex i) const
{
    auto it = status.constFind(i);
    if(it == status.cend())
        return {};
    return *it;
}

QVariant getSizeData(std::optional<quint64> v)
{
    if(v.has_value())
        return humanReadableSize(*v);
    else
        return "-";
}

//}

AdjustBufferModel::AdjustBufferModel(QObject *parent)
    : ClientIndexTableModel(parent)
    , clients(new AdjustBufferMap())
{
}

AdjustBufferModel::~AdjustBufferModel()
{
    delete clients;
}

ClientIndex AdjustBufferModel::clientIndexOf(int row) const
{
    return clients->indexOf(row);
}

int AdjustBufferModel::rowOf(const ClientIndex &sourceIndex) const
{
    return clients->rowOf(sourceIndex);
}

int AdjustBufferModel::rowCount(const QModelIndex &) const
{
    return clients->cfg.clients.size()+1;
}

int AdjustBufferModel::columnCount(const QModelIndex &) const
{
    return COL_TOTAL;
}

QVariant AdjustBufferModel::data(const QModelIndex &index, int role) const
{
    auto column = index.column();
    if(index.row() == 0) {
        auto cfg = clients->cfg.combinerCfg;
        auto st = clients->combinerStatus;
        switch (role) {
        case Qt::EditRole: return cfg.outputBufSize;
        case Qt::DisplayRole: return data_DisplayRole(cfg, st, column);
        case Qt::ToolTipRole: return data_ToolTip(cfg, st, column);
        case Qt::BackgroundRole: return data_BackgroundRole(cfg.outputBufSize, st, column);
        }
        return {};
    }

    auto clientIndex = clientIndexOf(index.row());
    if (clientIndex.isEmpty())
        return {};

    const auto cfg = clients->cfg.clients[clientIndex];
    const auto st = clients->status[clientIndex];
    switch (role) {
    case Qt::EditRole: return data_EditRole(cfg, column);
    case Qt::DisplayRole: return data_DisplayRole(cfg, st, column);
    case Qt::ToolTipRole: return data_ToolTip(cfg, st, column);
    case Qt::BackgroundRole: return data_BackgroundRole(cfg, st, column);
    }
    return {};
}

bool AdjustBufferModel::setData(const QModelIndex &index, const QVariant &value, int role)
{
    if(role != Qt::EditRole)
        return false;
    if(index.column() != COL_CFG_SIZE)
        return false;

    bool ok;
    auto size = value.toInt(&ok);
    if(!ok)
        return false;
    if(size<=0)
        return false;

    if(index.row()==0) {
        auto &cfg = clients->cfg.combinerCfg;
        if(cfg.outputBufSize == size)
            return false;
        cfg.outputBufSize = size;
        cfg.checkAndRepair();

        invalidateCombinerRow();
        emit configChanged(clients->cfg);
        return true;
    }

    auto clientIndex = clientIndexOf(index.row());
    if(clientIndex.isEmpty())
        return false;

    auto &clCfg = clients->cfg.clients[clientIndex];
    if(clCfg.bufSize == size)
        return false;
    clCfg.bufSize = size;
    clCfg.checkAndRepair();

    invalidateRow(clientIndex);
    emit configChanged(clients->cfg);
    return true;
}

QVariant AdjustBufferModel::headerData(int section, Qt::Orientation orientation, int role) const
{
    if (orientation == Qt::Horizontal && role == Qt::DisplayRole) {
        switch (section) {
        case COL_TYPE: return "Type";
        case COL_INDEX: return "Index";
        case COL_CFG_SIZE: return "Size";
        case COL_PREF_SIZE: return "Prefered";
        case COL_CUR_SIZE: return "Cur size";
        }
    }
    return ClientIndexTableModel::headerData(section, orientation, role);
}

Qt::ItemFlags AdjustBufferModel::flags(const QModelIndex &i) const
{
    auto f = ClientIndexTableModel::flags(i);
    switch (i.column()) {
    case COL_CFG_SIZE:
        f |= Qt::ItemIsEditable;
        break;
    }
    return f;
}

void AdjustBufferModel::setConfig(const EvBConfig &newConfig)
{
    auto oldClientsMap = clients->cfg.getClientIndexEnabledMap();
    auto newClientsMap = newConfig.getClientIndexEnabledMap();
    bool doReset = makeClientIndexSet(oldClientsMap.keys())
            != makeClientIndexSet(newClientsMap.keys());
    if(doReset)
        beginResetModel();
    clients->cfg = newConfig;
    bool changed = false;
    if(clients->cfg.allowBufferAutoResize) {
        changed = autoResizeAll();
    }
    if(!doReset)
        emit dataChanged(index(0, COL_CFG_SIZE), index(rowCount(), COL_CFG_SIZE));
    if(doReset)
        endResetModel();
    if(changed)
        emit configChanged(clients->cfg);
}

void AdjustBufferModel::clientEvSizePeriodUpdated(const ClientIndex &i, const EventSizePeriod &period)
{
    auto clCfgIt = clients->cfg.clients.find(i);
    if(clCfgIt == clients->cfg.clients.end())
        return;

    auto prefSize = period.getNewBufSize();
    if(!prefSize)
        return;
    clients->status[i].prefSize = prefSize;

    bool changed = autoResizeClient(i);
    invalidateRow(i);
    if(changed)
        emit configChanged(clients->cfg);
}

void AdjustBufferModel::clientBuffSizeUpdated(const ClientIndex &i, quint64 size)
{
    auto clCfgIt = clients->cfg.clients.find(i);
    if(clCfgIt == clients->cfg.clients.end())
        return;

    clients->status[i].curSize = size;
    invalidateRow(i);
}

void AdjustBufferModel::combinerEvSizePeriodUpdated(const EventSizePeriod &period)
{
    auto prefSize = period.getNewBufSize();
    if(!prefSize)
        return;
    clients->combinerStatus.prefSize = prefSize;

    auto changed = autoResizeCombiner();
    invalidateCombinerRow();
    if(changed)
        emit configChanged(clients->cfg);
}

void AdjustBufferModel::combinerBuffSizeUpdated(int id, quint64 size)
{
    //todo: process size from writer and each tcp
    clients->combinerStatus.curSize = size;
    invalidateCombinerRow();
}

QVariant AdjustBufferModel::data_DisplayRole(const EvbCombinerConfig &cfg, const ClientBufStatus &st, int column) const
{
    switch (column) {
    case COL_TYPE: return "Combiner";
//    case COL_INDEX: return cfg.index.getProgramIndex().toString();
    case COL_CFG_SIZE: return humanReadableSize(cfg.outputBufSize);
    case COL_PREF_SIZE: return getSizeData(st.prefSize);
    case COL_CUR_SIZE: return getSizeData(st.curSize);
    }
    return {};
}

QVariant AdjustBufferModel::data_ToolTip(const EvbCombinerConfig &cfg, const ClientBufStatus &st, int column) const
{
    switch (column) {
    case COL_CFG_SIZE: return QString("%L1").arg(cfg.outputBufSize);
    case COL_PREF_SIZE: return QString("%L1").arg(st.prefSize.value_or(0));
    case COL_CUR_SIZE: return QString("%L1").arg(st.curSize.value_or(0));
    default:;
    }
    return {};
}

QVariant AdjustBufferModel::data_EditRole(const EvBClientConfig &cfg, int column) const
{
    switch(column) {
    case COL_CFG_SIZE: return cfg.bufSize;
    }
    return {};
}

QVariant AdjustBufferModel::data_DisplayRole(const EvBClientConfig &cfg, const ClientBufStatus &st, int column) const
{
    switch (column) {
    case COL_TYPE: return cfg.index.getProgramType().toString();
    case COL_INDEX: return cfg.index.getProgramIndex().toString();
    case COL_CFG_SIZE: return humanReadableSize(cfg.bufSize);
    case COL_PREF_SIZE: return getSizeData(st.prefSize);
    case COL_CUR_SIZE: return getSizeData(st.curSize);
    }
    return {};
}

QVariant AdjustBufferModel::data_ToolTip(const EvBClientConfig &cfg, const ClientBufStatus &st, int column) const
{
    switch (column) {
    case COL_CFG_SIZE: return QString("%L1").arg(cfg.bufSize);
    case COL_PREF_SIZE: return QString("%L1").arg(st.prefSize.value_or(0));
    case COL_CUR_SIZE: return QString("%L1").arg(st.curSize.value_or(0));
    default:;
    }
    return {};
}

QVariant AdjustBufferModel::data_BackgroundRole(const EvBClientConfig &cfg, const ClientBufStatus &st, int column) const
{
    if(!cfg.enabled)
        return AppStyle::tableBg(Severity::Unknown);

    return data_BackgroundRole(cfg.bufSize, st, column);
}

QVariant AdjustBufferModel::data_BackgroundRole(quint64 bufSize, const ClientBufStatus &st, int column) const
{
    switch (column) {
    case COL_CUR_SIZE:
        return bufSize == st.curSize.value_or(0)
                ? AppStyle::tableBg(Severity::Normal)
                : AppStyle::tableBg(Severity::Error);
    case COL_PREF_SIZE:
        if(st.curSize.has_value() || st.prefSize.has_value()) {
            return st.prefSize == *st.curSize
                    ? AppStyle::tableBg(Severity::Normal)
                    : AppStyle::tableBg(Severity::Error);
        } else {
            return {};
        }
    default:;
    }
    return {};
}

void AdjustBufferModel::invalidateRow(const ClientIndex &i, const QVector<int> roles)
{
    auto row = clients->rowOf(i);
    emit dataChanged(index(row, 0), index(row, COL_TOTAL-1), roles);
}

void AdjustBufferModel::invalidateCombinerRow(const QVector<int> roles)
{
    emit dataChanged(index(0, 0), index(0, COL_TOTAL-1), roles);
}

bool AdjustBufferModel::autoResizeAll()
{
    bool changed = false;
    changed |= autoResizeCombiner();
    if(changed)
        invalidateCombinerRow();

    for(auto cfg : clients->cfg.clients) {
        auto v = autoResizeClient(cfg.index);
        if(v)
            invalidateRow(cfg.index);
        changed |= v;
    }
    return changed;
}

bool AdjustBufferModel::autoResizeCombiner()
{
    if(!clients->cfg.allowBufferAutoResize)
        return false;

    auto &cfgSize = clients->cfg.combinerCfg.outputBufSize;
    auto prefSize = clients->combinerStatus.prefSize;
    if(!prefSize)
        return false;

    double factor = 1.*prefSize.value()/cfgSize;
    if(factor<0.5 || factor>1.1) {
        qInfo() << QString("Combiner autoresized from %1 to %2")
                   .arg(humanReadableSize(cfgSize))
                   .arg(humanReadableSize(*prefSize));
        cfgSize = *prefSize;
        return true;
    }

    return false;
}

bool AdjustBufferModel::autoResizeClient(const ClientIndex &i)
{
    auto clCfgIt = clients->cfg.clients.find(i);
    if(clCfgIt == clients->cfg.clients.end())
        return false;

    if(!clients->cfg.allowBufferAutoResize)
        return false;

    auto prefSize = clients->status.value(i).prefSize;
    if(!prefSize)
        return false;

    double factor = 1.*prefSize.value()/clCfgIt->bufSize;
    if(factor<0.5 || factor>1.1) {
        qInfo() << QString("Client '%1' autoresized from %2 to %3")
                   .arg(i.toString())
                   .arg(humanReadableSize(clCfgIt->bufSize))
                   .arg(humanReadableSize(*prefSize));
        clCfgIt->bufSize = *prefSize;
        return true;
    }

    return false;
}
