#include "StatisticDialog.h"
#include "ui_StatisticDialog.h"
#include "SizeHelper.h"

#include <QTimer>

namespace {
constexpr static const char* prop_type = "type";

StatisticDialog::Table getTableType(const QTableWidget *t)
{
    return static_cast<StatisticDialog::Table>(t->property(prop_type).toInt());
}

enum COL_T {
    COL_T_INDEX,
    COL_T_PEER,
    COL_T_RING_BUFFER_SIZE,
    COL_T_FREE_PART,
    COL_T_DATA_PART,
    COL_T_TRASH_PART,

    COL_T_RCV_BUFF_SIZE, // soReceiveBufferSize
    COL_T_READ_BUFFER_SIZE, // readBufferSize
    COL_T_AVAIL_SIZE, // bytesAvailable

    COL_T_TOTAL_SIZE,
    COL_T_SPEED,
    COL_T_TOTAL
};
enum COL_A {
    COL_A_INDEX,
    COL_A_TOT_SIZE,
    COL_A_CORR_SIZE,
    COL_A_BAD_SYNC,
    COL_A_LAST_EV_NUM,
    COL_A_LAST_SYNC,
    COL_A_EV_SIZE,
    COL_A_EV_CNT,
    COL_A_PERIOD_EV_SIZE,
    COL_A_TOTAL
};
enum COL_CC {
    COL_CC_INDEX,
    COL_CC_LAST_EV_NUM,
    COL_CC_LAST_MISSED_EV_NUM,
    COL_CC_PRESENT_IN_LAST_EVENT,
    COL_CC_FLUSHED_SIZE,
    COL_CC_EVENT_BLOCKS,
    COL_CC_STAT_BLOCKS,
    COL_CC_JSON_BLOCKS,
    COL_CC_TRASHED_EVENTS,
    COL_CC_META_RB_STATE,
    COL_CC_TOTAL
};
enum COL_FW {
    COL_FW_ID,
    COL_FW_DATAPART,
    COL_FW_WRITER_RB_SIZE,
    COL_FW_WRITER_RB_PREF_SIZE,
    COL_FW_SPEED,
    COL_FW_TOTAL
};

enum COL_OUT_TCP {
    COL_OUT_TCP_ID,
    COL_OUT_TCP_STATUS,
    COL_OUT_TCP_PEER,

    COL_OUT_TCP_RB_SIZE,
    COL_OUT_TCP_DATAPART,

    COL_OUT_TCP_SOCKET_SIZE_TO_WRITE, // bytesToWrite
    COL_OUT_TCP_SOCKET_SND_BUFF_SIZE, // soSendBufferSize

    COL_OUT_TCP_TOTAL_SIZE,
    COL_OUT_TCP_SPEED,
    COL_OUT_TCP_TOTAL
};

const QMap<StatisticDialog::Table, QMap<int, QString>> TableHeaders = {
    {StatisticDialog::Table::Tcp,
     {
         {COL_T_INDEX, "Client"}, // prev was program_index
         {COL_T_PEER, "Peer"},
         {COL_T_RING_BUFFER_SIZE, "Ring buffer size"},
         {COL_T_DATA_PART, "Data %"},
         {COL_T_FREE_PART, "Free %"},
         {COL_T_TRASH_PART, "Trash %"},

         {COL_T_RCV_BUFF_SIZE, "Rcv buff size"},
         {COL_T_READ_BUFFER_SIZE, "Read buffer size"},
         {COL_T_AVAIL_SIZE, "Avail"},

         {COL_T_TOTAL_SIZE, "Tot. size"},
         {COL_T_SPEED, "Speed"},
     }
    },
    {StatisticDialog::Table::Anal,
     {
         {COL_A_INDEX, "Client"}, // prev was program_index
         {COL_A_TOT_SIZE, "Tot. size"},
         {COL_A_CORR_SIZE, "Corrupted"},
         {COL_A_BAD_SYNC, "Bad sync"},
         {COL_A_LAST_EV_NUM, "Last EvNum"},
         {COL_A_LAST_SYNC, "Last sync"},
         {COL_A_EV_SIZE, "Ev size"},
         {COL_A_EV_CNT, "Ev cnt"},
         {COL_A_PERIOD_EV_SIZE, "Period ev. size"},
     }
    },
    {StatisticDialog::Table::CombClient,
     {
         {COL_CC_INDEX, "id"},
         {COL_CC_LAST_EV_NUM, "Last evNum"},
         {COL_CC_LAST_MISSED_EV_NUM, "Last missed evNum"},
         {COL_CC_FLUSHED_SIZE, "Flushed size"},
         {COL_CC_EVENT_BLOCKS, "Events blocks"},
         {COL_CC_STAT_BLOCKS, "Stat blocks"},
         {COL_CC_JSON_BLOCKS, "JSON blocks"},
         {COL_CC_TRASHED_EVENTS, "Trashed ev"},
         {COL_CC_META_RB_STATE, "Mete RB"},
         {COL_CC_PRESENT_IN_LAST_EVENT, "In last ev"},
     }
    },
    {StatisticDialog::Table::Writer,
     {
         {COL_FW_ID, "Id"},
         {COL_FW_DATAPART, "FW datapart"},
         {COL_FW_WRITER_RB_SIZE, "FW RB size"},
         {COL_FW_WRITER_RB_PREF_SIZE, "FW pref size"},
         {COL_FW_SPEED, "FW speed, MB/s"},
     }
    },
    {StatisticDialog::Table::OutTcp,
     {
         {COL_OUT_TCP_ID, "TcpOut #"},
         {COL_OUT_TCP_STATUS, "Status"},
         {COL_OUT_TCP_PEER, "Peer"},
         {COL_OUT_TCP_RB_SIZE, "RB size"},
         {COL_OUT_TCP_DATAPART, "datapart"},
         {COL_OUT_TCP_SOCKET_SIZE_TO_WRITE, "Socket bytes to write"},
         {COL_OUT_TCP_SOCKET_SND_BUFF_SIZE, "Send buff size"},
         {COL_OUT_TCP_TOTAL_SIZE, "Tot. size"},
         {COL_OUT_TCP_SPEED, "Speed"}
     }
    }
};

const QMap<StatisticDialog::Table, QMap<int, QString>> TableToolTip = {
    {StatisticDialog::Table::Tcp,
     {
         {COL_T_RCV_BUFF_SIZE, "socketOption: ReceiveBufferSizeSocketOption"},
         {COL_T_READ_BUFFER_SIZE, "QAbstractSocket's internal read buffer"},
         {COL_T_AVAIL_SIZE, "QIODevice: Number of incoming bytes"},

         {COL_T_TOTAL_SIZE, "Total received size"},
         {COL_T_SPEED, "Average speed for last update"}
     }
    },
    {StatisticDialog::Table::CombClient,
     {
         {COL_CC_META_RB_STATE, "free/prepared/flushed"}
     }
    },
    {StatisticDialog::Table::OutTcp,
     {
         {COL_OUT_TCP_SOCKET_SIZE_TO_WRITE, "QIODevice: number of bytes to write"},
         {COL_OUT_TCP_SOCKET_SND_BUFF_SIZE, "socketOption: SendBufferSizeSocketOption"},
         {COL_OUT_TCP_TOTAL_SIZE, "Total written size"},
         {COL_OUT_TCP_SPEED, "Average speed for last update"},
     }
    },
};

QString getTableToolTip(const QTableWidget *t, int col) {
    StatisticDialog::Table type = getTableType(t);
    auto ttPrefix = TableToolTip[type][col];
    return ttPrefix;
}

template<typename T>
void setTableItemSize(QTableWidgetItem *i, T s) {
    i->setText(humanReadableSize(s));
    auto tt = QString("%L1").arg(s);
    auto ttPrefix = getTableToolTip(i->tableWidget(), i->column());
    if(!ttPrefix.isEmpty())
        tt.prepend(ttPrefix+'\n');
    i->setToolTip(tt);
}
}

StatisticDialog::StatisticDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::StatisticDialog),
    refreshTimer(new QTimer(this))
{
    ui->setupUi(this);
    init();
    refreshTimer->setInterval(100);
    refreshTimer->setSingleShot(true);
    connect(refreshTimer, &QTimer::timeout, this, &StatisticDialog::updateStat);
}

StatisticDialog::~StatisticDialog()
{
    delete ui;
}

void StatisticDialog::setVisible(bool visible)
{
    QDialog::setVisible(visible);
    if(visible) {
        updateStat();
        if(pendingCombStat.has_value()) {
            updateCombinerStats(pendingCombStat.value());
        }
    }
}

void StatisticDialog::analizerListUpdated(const ClientIndexSet &clientIds)
{
    const auto &toAdd = clientIds - curClientIds;
    const auto &toDel = curClientIds - clientIds;
    curClientIds = clientIds;

    for(auto i : toDel) {
        for(auto t : {Table::Tcp, Table::Anal, Table::CombClient}) {
            auto table = getTableWidget(t);
            int r = getRowByClientIndex(t, i);
            if(r == -1) continue;
            table->removeRow(r);
        }
        tcpStats.remove(i);
        analyzerStats.remove(i);
    }

    for(auto t : {Table::Tcp, Table::Anal, Table::CombClient}) {
        auto table = getTableWidget(t);
        for(auto i : toAdd) {
            int r = table->rowCount();
            table->setRowCount(r+1);
            for(int c=0; c<table->columnCount(); ++c) {
                table->setItem(r, c, new QTableWidgetItem());
            }
            auto item = table->item(r, getTableClientIndexCol(t));
            item->setData(Qt::UserRole, QVariant::fromValue(i));
            item->setText(i.toString());
        }
    }
}

void StatisticDialog::setCombinerConfig(const EvbCombinerConfig &cfg)
{
    ui->labelWriter->setVisible(cfg.combinerOut==COMBINER_FILE);
    ui->tableWidgetFileWriter->setVisible(cfg.combinerOut==COMBINER_FILE);
    ui->labelOutTcp->setVisible(cfg.combinerOut==COMBINER_TCP);
    auto t = ui->tableWidgetOutTcp;
    // t->setVisible(cfg.combinerOut==COMBINER_TCP);
    int totTcpNum = cfg.combinerOutTcpNum + 1; // with monitor
    while(t->rowCount()>totTcpNum)
        t->removeRow(t->rowCount()-1);
    while(t->rowCount()<totTcpNum) {
        int r = t->rowCount();
        t->setRowCount(r+1);
        for(int c=0; c<t->columnCount(); ++c)
            t->setItem(r, c, new QTableWidgetItem());
        t->item(r, COL_OUT_TCP_ID)->setText(r ? QString::number(r) : "Monitor");
        t->item(r, COL_OUT_TCP_ID)->setData(Qt::UserRole, r);
    }
}

void StatisticDialog::analyzerStatUpdated(const AnalyzerStat &s)
{
    analyzerStats[s.index] = s;
    if(!refreshTimer->isActive())
        refreshTimer->start();
}

void StatisticDialog::updateTcpStatus(const TcpClientStatus &status)
{
    tcpStats[status.index] = status.statistic;
    if(!refreshTimer->isActive())
        refreshTimer->start();
}

void StatisticDialog::updateCombinerStats(const CombinerStats &combStat)
{
    if(isHidden()) {
        pendingCombStat = combStat;
        return;
    }

    updateFWStats(combStat);
    updateOutTcpStats(combStat);
    updateCombClientsStats(combStat);

    prevCombStat = combStat;
    pendingCombStat.reset();
}

void StatisticDialog::statusUpdated(const FileWriterStats &s)
{
    const qint64 dtMs = (prevFwStat.timeStamp.isValid() && s.timeStamp.isValid()) ? prevFwStat.timeStamp.msecsTo(s.timeStamp) : 0;
    qint64 dSize = s.writtenTotalSize - prevFwStat.writtenTotalSize;
    const double speed = dtMs ? (1. * dSize / dtMs *1e3) : 0;

    auto t = getTableWidget(Table::Writer);
    t->item(0, COL_FW_DATAPART)->setText(QString::number(s.dataPart*100,'f',1)+'%');
    setTableItemSize(t->item(0, COL_FW_WRITER_RB_SIZE), s.rbSize);
//    setTableItemSize(t->item(0, COL_FW_WRITER_RB_PREF_SIZE), prevCombStat.fwPrefSize);
    t->item(0, COL_FW_SPEED)->setText(dtMs ? QString::number(1e-6*speed,'f',3) : QString("-"));

    prevFwStat = s;
}

void StatisticDialog::updateStat()
{
    if(isHidden())
        return;

    updateAnalStat();
    updateTcpStat();
}

void StatisticDialog::init()
{
    QMap<Table, int> tablesColCount = {
        {Table::Tcp, COL_T_TOTAL},
        {Table::Anal, COL_A_TOTAL},
        {Table::CombClient, COL_CC_TOTAL},
        {Table::Writer, COL_FW_TOTAL},
        {Table::OutTcp, COL_OUT_TCP_TOTAL}
    };
    for(auto t : TableHeaders.keys()) {
        auto table = getTableWidget(t);
        table->setProperty(prop_type, static_cast<int>(t));
        const QMap<int, QString> headers = TableHeaders[t];
        const int colCnt = tablesColCount[t];
        table->setColumnCount(colCnt);
        for(int c=0; c<colCnt; ++c) {
            auto i = new QTableWidgetItem(headers.value(c));

            auto tt = getTableToolTip(table, c);
            if(!tt.isEmpty())
                i->setToolTip(tt);

            table->setHorizontalHeaderItem(c, i);
        }

#if QT_VERSION >= 0x050000
        table->horizontalHeader()->setSectionResizeMode(QHeaderView::ResizeToContents);
#else
        table->horizontalHeader()->setResizeMode(QHeaderView::ResizeToContents);
#endif
        if(t==Table::Writer) {
            table->setRowCount(1);
            for(int col=0;col<colCnt;++col)
                table->setItem(0, col, new QTableWidgetItem());
            table->item(0, COL_FW_ID)->setText("File writer");
            updateTableSize(table);
        }
    }
}

void StatisticDialog::updateTableSize(QTableWidget *table)
{
    if(table == ui->tableWidgetFileWriter) {
        int height=table->horizontalHeader()->height();
        for(int i=0;i<table->rowCount();++i){
            height += table->rowHeight(i);
        }
        height += 5;
        table->setFixedHeight(height);
    }

    int width=table->verticalHeader()->width();
    for(int i=0;i<table->columnCount();++i){
        width += table->columnWidth(i);
    }
    width += 5;
    table->setMinimumWidth(width);
}

QTableWidget *StatisticDialog::getTableWidget(StatisticDialog::Table t) const
{
    switch (t) {
    case Table::Tcp: return ui->tableWidgetClient;
    case Table::Anal: return ui->tableWidgetAnalyzer;
    case Table::CombClient: return ui->tableWidgetCombClient;
    case Table::Writer: return ui->tableWidgetFileWriter;
    case Table::OutTcp: return ui->tableWidgetOutTcp;
    }
    qFatal(__PRETTY_FUNCTION__);
}

int StatisticDialog::getTableClientIndexCol(StatisticDialog::Table t) const
{
    switch (t) {
    case Table::Tcp:  return COL_T_INDEX;
    case Table::Anal: return COL_A_INDEX;
    case Table::CombClient: return COL_CC_INDEX;
    default: qFatal(__PRETTY_FUNCTION__);
    }
}

int StatisticDialog::getRowByClientIndex(Table t, const ClientIndex &i) const
{
    auto table = getTableWidget(t);

    for(int r=0; r<table->rowCount(); ++r)
        if(getClientIndexByRow(t, r) == i)
            return r;
    return -1;
}

ClientIndex StatisticDialog::getClientIndexByRow(Table t, int r) const
{
    auto table = getTableWidget(t);
    int col = getTableClientIndexCol(t);

    auto item = table->item(r, col);
    return item ? item->data(Qt::UserRole).value<ClientIndex>() : ClientIndex();
}

int StatisticDialog::getRowByIndex(Table t, int id) const
{
    auto table = getTableWidget(t);

    for(int r=0; r<table->rowCount(); ++r)
        if(getIndexByRow(t, r) == id)
            return r;
    return -1;
}

int StatisticDialog::getIndexByRow(Table t, int r) const
{
    Q_ASSERT(t==Table::OutTcp);
    if(t!=Table::OutTcp)
        return -1;

    auto table = getTableWidget(t);
    int col = COL_OUT_TCP_ID;

    auto item = table->item(r, col);
    return item ? item->data(Qt::UserRole).value<int>() : -1;
}

void StatisticDialog::updateAnalStat()
{
    auto t = getTableWidget(Table::Anal);
    t->setUpdatesEnabled(false);
    for(const auto &s : analyzerStats) {
        int row = getRowByClientIndex(Table::Anal, s.index);
        if(row==-1)
            continue;

        t->item(row, COL_A_TOT_SIZE)->setText(humanReadableSize(s.totalSize));
        t->item(row, COL_A_CORR_SIZE)->setText(humanReadableSize(s.corruptedSize));
        t->item(row, COL_A_BAD_SYNC)->setText(QString::number(s.badSync));
        t->item(row, COL_A_LAST_EV_NUM)->setText(QString::number(s.lastEvNum));
        t->item(row, COL_A_LAST_SYNC)->setText(MpdHdr::getSyncName(s.lastSyncWord));
        t->item(row, COL_A_EV_SIZE)->setText(humanReadableSize(s.evSize));
        t->item(row, COL_A_EV_CNT)->setText(QString("%L1").arg(s.evCnt));
        auto evSize = s.period.getEvSize();
        t->item(row, COL_A_PERIOD_EV_SIZE)->
            setText(evSize ? QString("%L1").arg(*evSize) : "-");
    }
    analyzerStats.clear();

    updateTableSize(t);
    t->setUpdatesEnabled(true);
}

void StatisticDialog::updateTcpStat()
{
    auto t = getTableWidget(Table::Tcp);
    t->setUpdatesEnabled(false);
    for (auto s : tcpStats) {
        int row = getRowByClientIndex(Table::Tcp, s.index);
        if(row==-1)
            continue;

        t->item(row, COL_T_PEER)->setText(s.peer.host.isNull() ? "-" : s.peer.toString());
        setTableItemSize(t->item(row, COL_T_RING_BUFFER_SIZE), s.ringBufSize);
        t->item(row, COL_T_FREE_PART)->setText(QString::number(s.freePart, 'f', 1));
        t->item(row, COL_T_DATA_PART)->setText(QString::number(s.dataPart, 'f', 1));
        t->item(row, COL_T_TRASH_PART)->setText(QString::number(s.trashPart, 'f', 1));

        setTableItemSize(t->item(row, COL_T_RCV_BUFF_SIZE), s.soReceiveBufferSize.value_or(0));
        setTableItemSize(t->item(row, COL_T_READ_BUFFER_SIZE), s.readBufferSize.value_or(0));
        t->item(row, COL_T_AVAIL_SIZE)->setText(QString("%L1 B").arg(s.bytesAvailable.value_or(0)));

        t->item(row, COL_T_TOTAL_SIZE)->setText(humanReadableSize(s.bytes));
        t->item(row, COL_T_SPEED)->setText(QString("%1 MB/s").arg(s.speed, 0, 'f', 1));
    }

    updateTableSize(t);
    t->setUpdatesEnabled(true);
}

void StatisticDialog::updateCombClientsStats(const CombinerStats &combStat)
{
    const qint64 dtMs = (prevCombStat.timeStamp.isValid() && combStat.timeStamp.isValid())
        ? prevCombStat.timeStamp.msecsTo(combStat.timeStamp) : 0;

    auto t = getTableWidget(Table::CombClient);
    t->setUpdatesEnabled(false);
    for(auto it=combStat.clientsStat.cbegin(); it!=combStat.clientsStat.cend(); ++it) {
        auto cs = it.value();
        int row = getRowByClientIndex(Table::CombClient, it.key());
        std::optional<CombinerClientStats> prevCs;
        auto prevIt = prevCombStat.clientsStat.constFind(it.key());
        if(prevIt != prevCombStat.clientsStat.cend())
            prevCs = *prevIt;
        if(row==-1)
            continue;

        t->item(row, COL_CC_LAST_EV_NUM)->setText(cs.lastEventNum.has_value()
                                                      ? QString("%L1").arg(cs.lastEventNum.value())
                                                      : "-");
        t->item(row, COL_CC_LAST_MISSED_EV_NUM)->setText(cs.lastMissingEvNum.has_value()
                                                      ? QString("%L1").arg(cs.lastMissingEvNum.value())
                                                      : "-");
        bool isInLastEv = combStat.lastEventClients.contains(it.key());
        t->item(row, COL_CC_PRESENT_IN_LAST_EVENT)->setText(isInLastEv ? QChar(0x2714) : QChar(0x2716));
        auto fg = t->item(row, COL_CC_PRESENT_IN_LAST_EVENT)->foreground();
        fg.setColor(isInLastEv ? Qt::darkGreen : Qt::red);
        t->item(row, COL_CC_PRESENT_IN_LAST_EVENT)->setForeground(fg);

        auto upCell = [&, this](int col, quint64 CombinerClientStats::* field) {
            auto cell = t->item(row, col);
            if(!cell)
                return;
            cell->setText(QString("%L1").arg(cs.*field));
            if(prevCs && dtMs && dtMs<1e6)
                cell->setToolTip(QString("%L1 Hz").arg(1e3*(cs.*field-prevCs.value().*field)/dtMs));
        };
        t->item(row, COL_CC_FLUSHED_SIZE)->setText(humanReadableSize(cs.flushedSize));
        upCell(COL_CC_EVENT_BLOCKS, &CombinerClientStats::eventBlocks);
        upCell(COL_CC_STAT_BLOCKS, &CombinerClientStats::statBlocks);
        upCell(COL_CC_JSON_BLOCKS, &CombinerClientStats::jsonBlocks);
        upCell(COL_CC_TRASHED_EVENTS, &CombinerClientStats::trashedEvents);

        t->item(row, COL_CC_META_RB_STATE)->setText(QString("%1/%2/%3")
                                                 .arg(cs.freeEvPart)
                                                 .arg(cs.preparedEvPart)
                                                 .arg(cs.flushedEvPart));
    }

    updateTableSize(t);
    t->setUpdatesEnabled(true);
}

void StatisticDialog::updateFWStats(const CombinerStats &combStat)
{
    auto t = getTableWidget(Table::Writer);
    setTableItemSize(t->item(0, COL_FW_WRITER_RB_PREF_SIZE), combStat.fwPrefSize);
}

void StatisticDialog::updateOutTcpStats(const CombinerStats &combStat)
{
    auto t = getTableWidget(Table::OutTcp);
    t->setUpdatesEnabled(false);
    for(const auto &s : combStat.tcpStatus) {
        int row = getRowByIndex(Table::OutTcp, s.index);
        if(row==-1)
            continue;

        QString status;
        if(s.listening) {
            if(s.connected) {
                status = s.isEmpty ? "empty" : "non-empty";
            } else {
                status = "disconnected";
            }
        } else {
            status = "disabled";
        }
        t->item(row, COL_OUT_TCP_STATUS)->setText(status);

        QString peerStr = s.peer.hostName;
        if(peerStr.isEmpty()) {
            if(s.peer.host.isNull()) {
                peerStr = "-";
            } else {
                peerStr = s.peer.host.toString();
            }
        }
        if(s.peer.port)
            peerStr += ":"+QString::number(s.peer.port);

        auto peerItem = t->item(row, COL_OUT_TCP_PEER);
        peerItem->setText(peerStr);
        peerItem->setToolTip(s.peer.host.toString());

        t->item(row, COL_OUT_TCP_DATAPART)->setText(QString("%1%").arg(s.dataPart,0,'f',1));
        t->item(row, COL_OUT_TCP_RB_SIZE)->setText(humanReadableSize(s.ringBufSize));

        t->item(row, COL_OUT_TCP_SOCKET_SIZE_TO_WRITE)->setText(s.bytesToWrite ? humanReadableSize(*s.bytesToWrite) : "-");
        t->item(row, COL_OUT_TCP_SOCKET_SND_BUFF_SIZE)->setText(s.soSendBufferSize ? humanReadableSize(*s.soSendBufferSize) : "-");

        t->item(row, COL_OUT_TCP_TOTAL_SIZE)->setText(humanReadableSize(s.bytes));
        t->item(row, COL_OUT_TCP_SPEED)->setText(QString("%1 MB/s").arg(s.speed, 0, 'f', 1));
    }

    updateTableSize(t);
    t->setUpdatesEnabled(true);
}
