#include "TcpClient.h"

#include "params.h"
#include "Analizer.h"
#include "EvbSignalRouter.h"
#include "EvBConfig.h"
#include "RingBuffer.h"

#include "base/DnsLookup.h"
#include "pnp-server/PNPMessage.h"
#include "metric-sender/MetricSender.h"
#include "metric-sender/MetricTypes.h"

#include <QDebug>
#include <QElapsedTimer>
#include <QHostInfo>
#include <QTimer>
#include <QTime>

#include <arpa/inet.h>
#include <errno.h>

namespace {
//const int PACK_SIZE_LIMIT = 20*1024*1024;
//const int QUEUE_SIZE_LIMIT = 100;
//const int RING_SIZE = 100*1024*1024;
const int RING_SIZE = 50*1000*1000;
const int RECONNECT_INTERVAL = 200; //ms
const double RB_HIGH_WATERMARK = 0.9;
const double RB_LOW_WATERMARK = 0.5;
const int LOW_SPACE_INTERVAL = 1000; //ms
const qint64 SOCKET_READ_BUFFER_SIZE=1e8; // 100 MB


QString removeLocalDomain(const QString &host)
{
    const QString &tail = "." + QHostInfo::localDomainName();
    int i = host.lastIndexOf(tail);
    if (i > 0)
        return host.left(i);
    return host;
}

struct sockaddr_in getPeerAddr(const QHostAddress &hostAddr, quint32 hostPort)
{
    // set up addres
    struct sockaddr_in peAddr;
    memset(&peAddr, 0, sizeof(peAddr));
    peAddr.sin_family = AF_INET;
    peAddr.sin_addr.s_addr = htonl(hostAddr.toIPv4Address());
    peAddr.sin_port = htons(hostPort);
    return peAddr;
}
}

TcpClient::TcpClient(ClientIndex _index, int _evNumIndex, RingBuffer<char> *_byteRB, quint64 inputBufSize)
    : TcpClientFsm(nullptr)
    , index(_index)
    , evNumIndex(_evNumIndex)
    , dnsLookup(new DnsLookup(this))
    , reconnectTimer(new QTimer(this))
    , readTimer(new QTimer(this))
    , statusTimer(new QTimer(this))
    , sendMetricsTimer(new QTimer(this))
    , resizeTimer(new QTimer(this))
    , low_space_timer(new QTimer(this))
    , byteRB(_byteRB)
    , status{index, int(inputBufSize)}
{
    setObjectName(__func__+(":"+index.toString()));

    qRegisterMetaType<QAbstractSocket::SocketError>();
    qRegisterMetaType<TcpClientStatus>();

    EvbSignalRouter::instance().connect_TcpClient(this);

    connect(dnsLookup, &DnsLookup::reverseEntryUpdated,
            this, &TcpClient::reverseEntryUpdated);
    connect(reconnectTimer, &QTimer::timeout, this, &TcpClient::connectToHost, Qt::QueuedConnection);
    reconnectTimer->setSingleShot(true);

    connect(this, &BaseFsm::stateChanged,
            this, &TcpClient::readData, Qt::QueuedConnection);
    connect(readTimer, &QTimer::timeout,
            this, &TcpClient::readData, Qt::QueuedConnection);
    readTimer->setSingleShot(true);
    readTimer->start(0);

    connect(statusTimer, &QTimer::timeout, this, &TcpClient::updateStatus);
    statusTimer->start(500);

    connect(sendMetricsTimer, &QTimer::timeout, this, &TcpClient::doSendMetrics);
    sendMetricsTimer->start(1000);

    connect(resizeTimer, &QTimer::timeout, this, &TcpClient::resizeBuffer);
    resizeTimer->setInterval(100);
    resizeTimer->setSingleShot(100);

    connect(low_space_timer, &QTimer::timeout, this, &TcpClient::low_space_timer_slot);
    low_space_timer->setSingleShot(true);
    low_space_timer->setInterval(LOW_SPACE_INTERVAL);

    QTimer::singleShot(0, this, &TcpClient::resizeBuffer);
    QTimer::singleShot(0, this, &TcpClient::connectToHost);

    idleBuf.resize(1e6);
}

TcpClient::~TcpClient()
{
    readoutEnabled = false;
    disconnectFromHost();
}

void TcpClient::connectToHost()
{
    try{
        if(!readoutEnabled || isConnectionAlive()){
            return;
        }

        if(!hasHostAddrPort()){
            qWarning() << objectName() << "can't connect; client address is unknown.";
            updateStatus();
        } else {
            ++connectingAttempts;

            socketFd = socket(AF_INET, SOCK_STREAM, 0);
            if (*socketFd < 0) {
              qFatal("%s can't create socket", __PRETTY_FUNCTION__);
            }
            static const constexpr int keepAliveVal = 1;
            setsockopt(*socketFd, SOL_SOCKET, SO_KEEPALIVE, &keepAliveVal,
                       sizeof(keepAliveVal));
            static const constexpr int keepTimeoutVal = 10;
            setsockopt(*socketFd, SOL_SOCKET, SO_KEEPALIVE, &keepTimeoutVal,
                       sizeof(keepTimeoutVal));

            auto addr = getPeerAddr(hostAddr, hostPort);
            if(::connect(*socketFd, (struct sockaddr *)&addr, sizeof(addr)) == 0) {
                suppresConnectionWarn.clear();
                qInfo() << objectName() << "connected.";
                readTimer->start(0);
                connectingAttempts=0;
            } else {
              auto mess = objectName().append(" can't connect.Err:%1")
                      .arg(strerror(errno));
              if(suppresConnectionWarn != mess) {
                  qWarning() << mess;
                  suppresConnectionWarn = mess;
              }
              if(connectingAttempts<=10) {
                  reconnectTimer->start(RECONNECT_INTERVAL);
              }
            }
            status.statistic.peer.updateAddr(*socketFd);
            status.statistic.updateLocalPort(*socketFd);
            status.statistic.connected = true; // todo unify to isConnectionAlive()?
            updateStatus();
        }
    } catch(...) {
        qDebug() << objectName() << "Exception in TcpClient::requestNewData()";
    }
    setConnected(isConnectionAlive());
}

void TcpClient::setReadOutEn(bool en)
{
    setClientEn(en);
    if(readoutEnabled == en) {
        qWarning("Trying to st%sed tcp-client", en ? "art start" : "op stopp");
        return;
    }
    readoutEnabled = en;
    suppresConnectionWarn.clear();

    if(readoutEnabled)
        connectToHost();
    else
        disconnectFromHost();
    updateStatus();
}

void TcpClient::disconnectFromHost()
{
    if(socketFd) {
        status.statistic.peer = {};
        status.statistic.localPort.reset();
        close(*socketFd);
        socketFd.reset();
        qInfo() << objectName() << "disconnected.";
    }
    reconnectTimer->stop();
    status.statistic.connected = false;
    setConnected(isConnectionAlive());
}

void TcpClient::readData()
{
    if(status.waitingResize)
        resizeBuffer();

    if(getState() != FsmState::Ready && getState() != FsmState::Run) {
        readTimer->start(100);
        return;
    }
    try {
        struct timeval timeout;
        timeout.tv_sec = 0;
        timeout.tv_usec = 100e3;
        fd_set read_fds;

        while (isConnectionAlive() && (timeout.tv_usec || timeout.tv_sec)) {
            checkFreeSizeWaterMark();

            // Select() updates fd_set's, so we need to build fd_set's before each select()call.
            FD_ZERO(&read_fds);
            FD_SET(*socketFd, &read_fds);

            int activity = select(*socketFd+1, &read_fds, nullptr, nullptr, &timeout);

            switch (activity) {
              case -1:
                qWarning("[%s] select failed: %s",
                       objectName().toStdString().c_str(), strerror(errno));
                disconnectFromHost();
                break;
              case 0:
                break;
              default:
                if (FD_ISSET(*socketFd, &read_fds)) {
                  if (!receiveFromPeer()) {
                      disconnectFromHost();
                      break;
                  }
                }
            }
        }
    } catch(...) {
        qDebug() << objectName() << "Exception in TcpClient::readData()";
    }
    readTimer->start(isConnectionAlive() ? 0 : 100);
}

void TcpClient::low_space_timer_slot()
{
    qInfo() << objectName() << ": free space available";
}

void TcpClient::doReqStateActions(const FsmState &st)
{
    if(performedStateAction == st)
        return;
    performedStateAction = st;

    qDebug() << objectName() << __PRETTY_FUNCTION__ << st;
    switch (st) {
    case FsmState::Configure:
        resetReadOut();
        break;
    default:;
    }
}

void TcpClient::setAddress(const ProgramDescription &descr, const ProgramInterface &interface)
{
    status.lastProgDescription = descr;
    //    qDebug() << QString("[ClientMgr #%1] connect to interface '%2' (id:%3)").arg(index)
    //                .arg(newInterface.type).arg(newInterface.id);

    if(hostAddr != interface.host || hostPort != interface.port) {
        connectingAttempts=0;
        disconnectFromHost();

        hostAddr = interface.host;
        hostPort = interface.port;
        if(descr.hostName.isEmpty()) {
            hostName.clear();
            dnsLookup->reverseLookup(interface.host.toString());
        } else {
            hostName = removeLocalDomain(descr.hostName);
        }

        //    qDebug()<<getName().append("; address=%2:%3; bufferSize=%4")
        //              .arg(hostAddr.toString()).arg(hostPort).arg(RING_SIZE);
        if(readoutEnabled)
            connectToHost();
    }
}


void TcpClient::resetReadOut()
{
    qDebug() << objectName() << "start resetting";
    Q_ASSERT(readoutEnabled);
//    Q_ASSERT(tcpSocket->state() == QTcpSocket::UnconnectedState);
//    Q_ASSERT(reconnectTimer->isActive()==false);
//    readTimer;
//    statusTimer;
    connectingAttempts=0;
    packageCounter=0;
    noFreeSpace=false;
    status.statistic.clear();

#ifdef RESET_ROUGH
    byteRB->clear();
#else
    int curFreeSpace = byteRB->getFreeSize();
    int needToFreeSize = byteRB->getAllSize()-curFreeSpace;
    qDebug().noquote() << QString("curFreeSpace=%1; needToFreeSize=%2")
            .arg(curFreeSpace).arg(needToFreeSize);
    QElapsedTimer timer;
    timer.start();
    while(byteRB->getFreeSize() != byteRB->getAllSize() && !timer.hasExpired(2000)){
        if(byteRB->waitTrashSpace(needToFreeSize, 50))
            byteRB->freeData(needToFreeSize);
    }

//    newDataWC;
    if(timer.hasExpired(2000)){
        qDebug().noquote() << QString("free=%1 data=%2 trash=%3")
                .arg(byteRB->getFreeSize())
                .arg(byteRB->getDataSize())
                .arg(byteRB->getTrashSize());
        qWarning().noquote() << QString("TcpClient index #%1 reset failed").arg(index);
    } else {
        qDebug().noquote() << QString("TcpClient index #%1 reset complete").arg(index);
        emit resetted(false);
    }
#endif
}

bool TcpClient::receiveFromPeer()
{
    if(!socketFd)
        return false;

    const bool runMod = getState() == FsmState::Run;
    ssize_t received_count;
    updateFreeSize();
    updateContFreeSize();

    while (1) {
        if(runMod) {
            Q_ASSERT(contFreeSizeCache>=0);
            if(contFreeSizeCache==0) {
//                if(updateContFreeSize()==0) // todo unblock this (somehow cause Q_ASSERT(byteRB->getWriterPos()==0); err)
                    break;
            }
            received_count = recv(*socketFd, byteRB->getWritePointer(),
                                  contFreeSizeCache, MSG_DONTWAIT);
        } else {
            received_count = recv(*socketFd, idleBuf.data(),
                                  idleBuf.size(), MSG_DONTWAIT);
        }
        if (received_count < 0) {
            if (errno == EAGAIN || errno == EWOULDBLOCK) {
                // peer is not ready right now, try again later
                return true;
            } else {
                qCritical("%s from peer error: %s",
                          __PRETTY_FUNCTION__, strerror(errno));
                return false;
            }
        } else if (received_count == 0) {
            return false;
        } else if (received_count > 0) {
            ++packageCounter;
            status.statistic.bytes += received_count;
            if(runMod) {
                bool doWrapCheck = false;
#ifdef QT_DEBUG
                QPair<int,int> spacesBeforeUp, spacesAfterUp;
                if(contFreeSizeCache==received_count && freeSizeCache!=received_count) {
                    spacesBeforeUp = {contFreeSizeCache, freeSizeCache};
                    if(updateContFreeSize()==received_count && updateFreeSize()!=received_count) {
                        doWrapCheck = true;
                        spacesAfterUp = {updateContFreeSize(), updateFreeSize()};
                    }
                }
#endif
                contFreeSizeCache -= received_count;
                freeSizeCache -= received_count;
                byteRB->writeData(received_count);
#ifdef QT_DEBUG
                if(doWrapCheck) {
                    // Transition from end to start of RB
                    if(byteRB->getWriterPos()!=0) {
                        qInfo() << objectName() << __PRETTY_FUNCTION__ << "wrap check:";
                        qInfo() << "before update:      " << spacesBeforeUp.first << spacesBeforeUp.second;
                        qInfo() << "after update:       " << spacesAfterUp.first << spacesAfterUp.second;
                        qInfo() << "after  substruction:" << contFreeSizeCache << freeSizeCache;
                        qInfo() << "RB WriterPos:       " << byteRB->getWriterPos();
                        qInfo() << "received_count:     " << received_count;
                        Q_ASSERT(byteRB->getWriterPos()==0);
                    }
                    qInfo() << __PRETTY_FUNCTION__ << "Wrap - ok";
                }
#endif
            }
        }
    }

    return true;
}

int TcpClient::updateContFreeSize() {
//    return contFreeSizeCache = byteRB->getContinuosFreeSize();
    contFreeSizeCache = byteRB->getContinuosFreeSize();
    if(contFreeSizeCache<0)
        qWarning() << objectName() << "ContinuosFreeSize:" << contFreeSizeCache;
    return contFreeSizeCache;
}

int TcpClient::updateFreeSize() {
//    return freeSizeCache = byteRB->getFreeSize();
    const auto &s = freeSizeCache = byteRB->getFreeSize();
    const auto &f = byteRB->getAllSize();
    auto &st = status.freeSpaceState;
    if(s==f)
        st = FreeSpaceState::Full;
    else if(s==0)
        st = FreeSpaceState::Empty;
    else if(s>(f*RB_HIGH_WATERMARK))
        st = FreeSpaceState::AlmostFull;
    else if(s<(f*RB_LOW_WATERMARK))
        st = FreeSpaceState::AlmostEmpty;
    else
        st = FreeSpaceState::Medium;
    if(freeSizeCache<0)
        qWarning() << objectName() << "freeSizeCache:" << freeSizeCache;
    return freeSizeCache;
}

void TcpClient::checkFreeSizeWaterMark()
{
    //TODO
//    const auto prevSt = status.freeSpaceState;
////    if(freeSizeCache<0)
//        updateFreeSize();

//    if(noFreeSpace != status.freeSpaceState<=FreeSpaceState::AlmostEmpty) {
//        lastNoFreeSpace.isValid();
//        if(low_space_timer->isActive()) {
//            low_space_timer->stop();
//        } else {
//            qWarning() << objectName() << "free space low";
//        }
//        noFreeSpace=true;
//    }
}

bool TcpClient::hasHostAddrPort() const
{
    return !hostAddr.isNull() && hostPort!=0;
}

void TcpClient::doSendMetrics()
{
    const QMap<QString, QString> tags({{"client", index.getProgramIndex().toString()}});
    MetricFields fields{{"free", byteRB->getFreePart()},
                        {"data", byteRB->getDataPart()},
                        {"trash", byteRB->getTrashPart()}};
    MetricSender::send("evb.in.buffer",  fields, tags);
}

void TcpClient::updateStatus()
{
    auto &statistic = status.statistic;
    statistic.updateRingBuffet(byteRB);
    statistic.updateSocketOptions(socketFd);
    statistic.updateSpeed();

    if(isConnectionAlive()) {
        if(statistic.peer.hostName != hostName)
            statistic.peer.hostName = hostName;

        const auto &loopSpeed = statistic.loopSpeed();
        status.state = TcpClientState::Connected;
        if(getState() == FsmState::Run) {
            int writtenSize = byteRB->getDataSize();
            auto buffLoad = writtenSize*100.0/byteRB->getAllSize();
            status.stateStr = QString("%1 packages recived; load=%2%%3")
                    .arg(packageCounter).arg(buffLoad,0,'f',1)
                    .arg((buffLoad>80)?"; lack of space":"");
        } else {
            status.stateStr = QString("Idle readout");
            if(loopSpeed.has_value())
                status.stateStr += QString(" (%1kB)")
                        .arg(loopSpeed.value()/1e3, 0, 'f', 2);
        }
    } else if(!readoutEnabled) {
        status.state = TcpClientState::Disabled;
        status.stateStr = "Disabled";
    } else if(srcBusy) {
        status.state = TcpClientState::SourceBusy;
        status.stateStr = "Source is busy";
    } else {
        if(hasHostAddrPort()) {
            status.stateStr = suppresConnectionWarn + QString(" (%1 attempt)")
                                                          .arg(connectingAttempts);
        } else {
            status.state = TcpClientState::WaitSource;
            status.stateStr = "Address is unknown";
        }
    }
    emit sendStatus(status);
}

void TcpClient::reverseEntryUpdated(const QPair<quint32, QString> &ent)
{
    const auto ip = hostAddr.toIPv4Address();
    if(ent.first == ip && hostName != ent.second) {
        hostName = ent.second;
        if(isConnectionAlive())
            updateStatus();
    }
}

void TcpClient::pnpCacheUpdated(const PnpInfoCache &cache)
{
    auto it = cache.constFind(status.lastProgDescription.uuid);
    if(it == cache.cend()) {
        status.lastProgDescription = {};
        status.usedInterface = {};
        hostAddr.clear();
        hostPort = 0;
        disconnectFromHost();
        updateStatus();
    } else {
        if(isConnectionAlive()) {
            status.lastProgDescription = *it;
            // May be update smth?
            return;
        }
    }

    srcBusy = false;
    for(auto pd : cache) {
        if(pd.getClientIndex() != index)
            continue;

        bool ok;
        auto interface = pd.getProgramInterface(ProgramInterfaceDataFlow,
                                                &ok, evNumIndex);
        if(!ok)
            continue;

        srcBusy = !interface.isFree;
        if(interface.isFree) {
            if(!interface.host.isNull() && interface.port)
                setAddress(pd, interface);
            srcBusyWarn = true;
            if(isConnectionAlive())
                break; // Skip other PD (which can be incorrect)
        } else {
            if(srcBusyWarn)
                qWarning() << objectName()
                           << QString("Can't connect to busy interface '%2' (id:%3)")
                                  .arg(interface.type).arg(interface.id);
            srcBusyWarn = false;
            disconnectFromHost();
        }
    }
}

void TcpClient::setClientConfig(const EvBClientConfig &cfg)
{
    if(cfg.index != index)
        return;
    if(status.preferedBufSize == cfg.bufSize)
        return;
    status.preferedBufSize = cfg.bufSize;
    status.waitingResize = true;
    failToResize = false;
    if(!resizeTimer->isActive())
        resizeTimer->start();
}

void TcpClient::resizeBuffer()
{
    Q_ASSERT(status.preferedBufSize>0);
    if(byteRB->getAllSize() == status.preferedBufSize) {
        status.waitingResize = false;
        return;
    }

    if(byteRB->getFreeSize()==byteRB->getAllSize()) {
        auto sufix = QString(" from %1MB to %2MB")
                .arg(1e-6*byteRB->getAllSize(),0,'f',3)
                .arg(1e-6*status.preferedBufSize,0,'f',3);
        bool ok = byteRB->resize(status.preferedBufSize);
        if(ok) {
            qInfo() << objectName() << "buffer resized" << sufix;
            failToResize = false;
            emit inputBufferSizeUpdated(index, status.preferedBufSize);
            status.waitingResize = false;
            setRbCreated();
            return;
        } else {
            if(!failToResize) {
                qWarning() << objectName() << "fail to resize" << sufix;
                failToResize = true;
            }
        }
    }
    resizeTimer->start();
}

bool TcpClient::isConnectionAlive() const
{
    return socketFd.has_value();
}
