/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "TrigDeviceController_p.h"
#include "TrigDeviceController.h"

#include <QJsonDocument>

#include "RcConfigConverter.h"
#include "run-control-fsm/TrigDeviceMgrFsm.h"

TrigDeviceControllerPrivate::TrigDeviceControllerPrivate(TrigDeviceController *p, QObject *parent)
    : QObject(parent), q_ptr(p), fsm(new TrigDeviceMgrFsm(this))
{
    connect(fsm, &DeviceMgrFsm::stateChanged,
            q_ptr, &TrigDeviceController::stateChanged);
    connect(fsm, &DeviceMgrFsm::stateChanged,
            this, &TrigDeviceControllerPrivate::onFsmStateChange);
    connect(this, &TrigDeviceControllerPrivate::printFsmInput,
            fsm, &DeviceMgrFsm::printFsmInput);
}

void TrigDeviceControllerPrivate::onFsmStateChange(const FsmState &state)
{
    switch (state) {
    case FsmState::Error:
        setTrigState(false);
        break;
    case FsmState::Run:
        fsm->clearStopEvNumber();
        break;
    case FsmState::Starting:
    {
        if(megamaster_flg) {
            setTrigState(false);
        } else {
            emit preRunReset(startEvNum);
            setTrigState(true);
            sendJsonOnRun();
        }
        fsm->setRequestState(megamaster_flg ? FsmState::CloseGlobTrig : FsmState::Run);
        break;
    }
    case FsmState::CloseGlobTrig:
        Q_ASSERT(megamaster_flg);
        break;
    case FsmState::OpenGlobTrig:
        Q_ASSERT(megamaster_flg);
        sendJsonOnRun();
        fsm->setRequestState(FsmState::Run);
        break;
    case FsmState::Stopping:
        setTrigState(false);
        fsm->setRequestState(FsmState::Ready);
        break;
    default:;
    }
}

void TrigDeviceControllerPrivate::setTrigStatus(const TrigStatus &status)
{
    auto prevSpill = trigStatus[status.getDeviceIndex()].spillStatus;
    if (!prevSpill || (status.spillStatus && (prevSpill->isReadoutPhase() && !status.spillStatus->isReadoutPhase()))) {
        sendJsonRuntime(status);
    }
    trigStatus[status.getDeviceIndex()] = status;

    if(lookingForStopEvNum) {
        stopEvNum = status.coreStatus.globEvNum;
        qDebug() << "StopEvNum detected as:" << stopEvNum;
        fsm->markStopEvNumber();
        lookingForStopEvNum = false;
    }
}

void TrigDeviceControllerPrivate::setupWritten(const DeviceIndex &index, const ConfigKey &setupSeqNum)
{
    fsm->setDevSetupKey(index, setupSeqNum);
    if(fsm->getState() == FsmState::Stopping)
        lookingForStopEvNum = true;
}

void TrigDeviceControllerPrivate::sendJsonRuntime(const TrigStatus &devSt)
{
    QVariantMap devRunStatus;

    {
        QVariantMap spillStatus;
        const auto spill = devSt.spillStatus;
        if (spill) {
            spillStatus["count"] = spill->count;
            spillStatus["phase"] = spill->status.reg.bit.phase;
            spillStatus["curSpillTimeMs"] = spill->curSpillTimeMs;
            devRunStatus["spill"] = spillStatus;
        }
    }

    {
        devRunStatus["temp"] = ConfigConverterUtil::toJsonValue(devSt.getTemps());
    }

    {
        QVariantMap cntStatus;
        const auto &cnt = devSt.statCounters.rc_counters;
        cntStatus["trigAll"] = cnt.trigAll();
        cntStatus["trigOnRun"] = cnt.trigOnRun();
        cntStatus["trigOnSpill"] = cnt.trigOnSpill();
        cntStatus["trigOnRunOnSpill"] = cnt.trigOnRunOnSpill();
        cntStatus["trigAccepted"] = cnt.trigAccepted();
        cntStatus["trigMissed"] = cnt.trigMissed();
        cntStatus["trigReadoutMissed"] = cnt.trigReadoutMissed();
        cntStatus["trigDeadtimeMissed"] = cnt.trigDeadtimeMissed();
        devRunStatus["counters"] = cntStatus;
    }

    {
        QVariantMap timeStatus;
        timeStatus["app"] = QDateTime::currentDateTime().toTime_t();
        devRunStatus["time"] = timeStatus;
    }

    QVariantMap devStatus{{"status",QVariantMap{{"runTime", devRunStatus}}}};
    auto jsonObj = AbstractConfigConverter<int>::fromVariantMap(devStatus);
    sendJson(devSt.getDeviceIndex(), jsonObj);
}

void TrigDeviceControllerPrivate::sendJsonFirmware(const TrigStatus &devSt)
{
    QVariantMap devFwStatus;

    {
        const auto &ver = devSt.getFwVer();
        QVariantMap v{{"min", ver.v_min},
                      {"maj", ver.v_maj},
                      {"rev", ver.revision}};
        devFwStatus["fw_version"] = v;
    }

    devFwStatus["sdb"] = QVariant::fromValue(devSt.getSdb());

    QVariantMap devStatus{{"status",QVariantMap{{"firmware", devFwStatus}}}};
    auto jsonObj = AbstractConfigConverter<int>::fromVariantMap(devStatus);
    sendJson(devSt.getDeviceIndex(), jsonObj);
}

void TrigDeviceControllerPrivate::sendJsonOnRun()
{
    for(auto st : trigStatus)
        sendJsonFirmware(st);
}

void TrigDeviceControllerPrivate::sendJson(const DeviceIndex &deviceIndex,
                                           const QJsonObject &jsonObj)
{
//    qInfo().noquote() << QJsonDocument(jsonObj).toJson(QJsonDocument::Indented);
    emit sendJsonBlock(deviceIndex, QJsonDocument(jsonObj).toJson(QJsonDocument::Compact));
}

void TrigDeviceControllerPrivate::incrementSetupKey()
{
    ++setupKey;
    fsm->setMainSetupKey(setupKey);
}

void TrigDeviceControllerPrivate::onTrigDevConfigUpdate()
{
    incrementSetupKey();
    emit writeSetup(config.rcTrigConfig, setupKey);
}

void TrigDeviceControllerPrivate::updateDeviceSet()
{
//    DeviceIndexSet set = makeDeviceIndexSet(trigThreads.keys());
//    DeviceIndexEnabledMap map;
    //    fsm->setDeviceList(map);
}

void TrigDeviceControllerPrivate::onOpenGlobTrig()
{
    Q_ASSERT(megamaster_flg);
    emit preRunReset(startEvNum);
    setTrigState(true);
    fsm->setRequestState(fsm->getState()==FsmState::CloseGlobTrig ? FsmState::OpenGlobTrig : FsmState::Error);
}

void TrigDeviceControllerPrivate::setTrigState(bool open)
{
    bool hasEnabledModules = !config.rcTrigConfig.index.isNull();
    if(hasEnabledModules) {
        if(config.rcTrigConfig.trig.trigOpen != open) {
            config.rcTrigConfig.trig.trigOpen = open;
            onTrigDevConfigUpdate();
            Q_Q(TrigDeviceController);
            emit q->configChanged(config);
        }
    }
    if(fsm->getState()==FsmState::Stopping) {
        if(hasEnabledModules && megamaster_flg) {
            lookingForStopEvNum = true;
        } else {
            qDebug() << "Reset StopEvNum.";
            stopEvNum.clear();
            fsm->markStopEvNumber();
        }
    }
}
