/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "VmeTrigsController_p.h"

#include <QDebug>

#include "CoreTypes.h"
#include "RcConfig.h"
#include "VmeTrigDevice.h"
#include "VmeTrigsController.h"
#include "base/ThreadController.h"
#include "mldiscover/DeviceDispatcher.h"
#include "VmeTrigFsm.h"

class VmeTrigDeviceProxy
{
public:
    VmeTrigDeviceProxy(VmeTrigDevice *dev)
        : tc(new ThreadController(dev, nullptr)) {}
private:
    QSharedPointer<ThreadController> tc;
};

VmeTrigsControllerPrivate::VmeTrigsControllerPrivate(VmeTrigsController *p, QObject *parent)
    : QObject(parent)
    , q_ptr(p)
    , config(new RcConfig)
    , fsm(new VmeTrigFsm(this))
{
    qRegisterMetaType<VmeTrigStatus>();
    connect(fsm, &DeviceMgrFsm::stateChanged,
            q_ptr, &VmeTrigsController::stateChanged);
    connect(this, &VmeTrigsControllerPrivate::printFsmInput,
            fsm, &DeviceMgrFsm::printFsmInput);
    QTimer::singleShot(0, this, &VmeTrigsControllerPrivate::incrementSetupKey);
}

VmeTrigsControllerPrivate::~VmeTrigsControllerPrivate()
{
    // intentionally left blank
}

void VmeTrigsControllerPrivate::addTrigDevice(const DeviceIndexEnabled &die)
{
    Q_Q(VmeTrigsController);

    const auto &index = die.index;

    // create config entry if not exists
    bool existed = config->vmeTrigs.contains(index);
    auto &vmeTrigConfig = config->vmeTrigs[index];
    vmeTrigConfig.enabled = die.enabled;
    if (!existed)
        emit q->trigConfigChanged(index, vmeTrigConfig);

    if (trigThreads.contains(index))
        return;

    if(die.enabled)
        addTrigThread(index, vmeTrigConfig);
    emit q->deviceAdded(index); // to model
}

void VmeTrigsControllerPrivate::removeTrigDevice(const DeviceIndex &index)
{
    Q_Q(VmeTrigsController);

    removeTrigThread(index);
    config->vmeTrigs.remove(index);
    emit q->configChanged(*config);
    emit q->deviceRemoved(index); // to model
}

void VmeTrigsControllerPrivate::addTrigThread(const DeviceIndex &index, const VmeTrigConfig &vc)
{
    if(trigThreads.contains(index))
        return;

    Q_Q(VmeTrigsController);

    auto dev = new VmeTrigDevice(index, nullptr);
    dev->setLiveMagicCheck(false);
    dev->setObjectName(QString("VmeTrig_%1").arg(index.getSerialStr()));
    trigThreads.insert(index, VmeTrigDeviceProxy(dev));

    if (q->deviceDispatcher)
        q->deviceDispatcher->subscribeHostAddressUpdates(index, dev);

    QObject::connect(q, &VmeTrigsController::vmeReadoutUpdated,
                     dev, &VmeTrigDevice::vmeReadoutUpdated);
    QObject::connect(dev, &VmeTrigDevice::updateStatus,
                     this, &VmeTrigsControllerPrivate::trigStatusUpdated);
    QObject::connect(dev, &VmeTrigDevice::updateStatus,
                     q, &VmeTrigsController::deviceStatusUpdated);
    QObject::connect(this, &VmeTrigsControllerPrivate::deviceConfigUpdated,
                     dev, &VmeTrigDevice::setDeviceConfig);
    QObject::connect(dev, &VmeTrigDevice::deviceConfigWritten,
                     fsm, &DeviceMgrFsm::setDevSetupKey);
    QObject::connect(q, &VmeTrigsController::evNumRequest,
                     dev, &VmeTrigDevice::evNumRequest);
    QObject::connect(dev, &VmeTrigDevice::evNumResponse,
                     q, &VmeTrigsController::evNumResponse);

    ModularDeviceConfig devCfg;
    devCfg.index = index;
    emit deviceConfigUpdated(setupKey, devCfg);
}

void VmeTrigsControllerPrivate::removeTrigThread(const DeviceIndex &index)
{
    //    emit q->deleteTrig(index);
    trigThreads.remove(index);
    trigStatus.remove(index);
    qInfo() << "vme trig device removed" << index;
}

DeviceIndexEnabledMap VmeTrigsControllerPrivate::myDevices() const
{
    return deviceListFromConfig(*config);
}

DeviceIndexEnabledMap VmeTrigsControllerPrivate::deviceListFromConfig(const RcConfig &newConfig)
{
    DeviceIndexEnabledMap map;
    for (auto it = newConfig.vmeTrigs.begin(); it != newConfig.vmeTrigs.end(); ++it)
        map[it.key()] = DeviceIndexEnabled(it.key(), it.value().enabled);
    return map;
}

void VmeTrigsControllerPrivate::trigStatusUpdated(const DeviceIndex &index, const VmeTrigStatus &status)
{
    Q_Q(VmeTrigsController);

    bool stopRunCondition = false;
    QString errStr;

    // Detect changes in status
    const VmeTrigStatus &oldStatus = trigStatus[index];
    const QString &oldPrefix = oldStatus.ident + " ";
    const QString &prefix = trigStatus[index].ident + " ";

    if (status.wrStatus.has_value()) {
        const auto &wrStatus = status.wrStatus.value();
        qint32 wrLinkErrorCount = status.wrStatus->linkError - wrStatus.linkError;
        if (wrLinkErrorCount > 0) {
            qWarning() << prefix + QString("WR link errors: %1").arg(wrLinkErrorCount);
        }
        if (oldStatus.wrStatus.has_value()) {
            qint32 wrSyncLostCount = wrStatus.syncFail - oldStatus.wrStatus->syncFail;
            if (wrSyncLostCount > 0) {
                qWarning() << prefix + QString("WR sync lost errors: %1").arg(wrSyncLostCount);
            }
        }
    }
    bool isInRun = fsm->getState() == FsmState::Run;
    if(config->vmeTrigs[index].checkWr) {
        fsm->setWrState(index, status.isWrTimeValid());
        if (!oldStatus.isWrTimeValid() && status.isWrTimeValid()) {
            qInfo() << prefix + "WR time valid";
        }
        if (oldStatus.isWrTimeValid() && !status.isWrTimeValid()) {
            errStr = prefix + "WR time invalid";
            (isInRun?qWarning():qInfo()) << errStr;
            stopRunCondition = true;
        }
    }
    if (!oldStatus.connected && status.connected) {
        qInfo() << prefix + "connected";
    }
    if (oldStatus.connected && !status.connected) {
        errStr = oldPrefix + "disconnected";
        (isInRun?qWarning():qInfo()) << errStr;
        stopRunCondition = true;
    }
    if(isInRun && config->vmeTrigs[index].enabled && stopRunCondition) {

        const MessageTreeItem m(SlaveType::VmeTrig,
                               last_got_run_id,
                               SeverityMessage(SlaveMessageSeverity::Error, errStr));
        emit q->sendSlaveMessage(m);
    }

    trigStatus[index] = status;
}

void VmeTrigsControllerPrivate::incrementSetupKey()
{
    ++setupKey;
    fsm->setMainSetupKey(setupKey);

    sendConfigToDevices();
}

void VmeTrigsControllerPrivate::sendVmeMap() const
{
    Q_Q(const VmeTrigsController);

    QMap<DeviceIndex, QString> evNumCheckMap;
    DeviceIndexEnabledMap dieMap;
    DeviceIndexSet wrCheckSet;
    for (auto it = config->vmeTrigs.begin(); it != config->vmeTrigs.end(); ++it) {
        const auto &index = it.key();
        const auto &trigConfig = it.value();
        dieMap[index] = DeviceIndexEnabled(index, trigConfig.enabled);
        if (trigConfig.enabled) {
            if(it->checkEvNum)
                evNumCheckMap[index] = index.getSerialStr();
            if(it->checkWr)
                wrCheckSet.insert(it.key());
        }
    }
    emit q->vmeMapUpdated(evNumCheckMap);
    fsm->setDeviceList(dieMap);
    fsm->setWrCheckSet(wrCheckSet);
}

void VmeTrigsControllerPrivate::sendConfigToDevices() const
{
    for(auto it=config->vmeTrigs.cbegin(); it!=config->vmeTrigs.cend(); ++it) {
        ModularDeviceConfig devCfg;
        devCfg.index = it.key();
        emit deviceConfigUpdated(setupKey, devCfg);
    }
}
