#ifndef VIEWERMAINWINDOW_H
#define VIEWERMAINWINDOW_H

#include <QMainWindow>
#include <QAbstractButton>
#include <QElapsedTimer>
#include <QHostAddress>
#include <QProcess>
#include <QMessageBox>
#include <QMetaType>
#include <QTime>
#include <QMap>

#include "base/app/BaseMainWindow.h"
#include "pnp-server/PNPMessage.h"
#include "configuration-manager/ConfigurationManager.h"
#include "histogramm/AnalyzerStat.h"

#include "AfiViewerConfig.h"
#include "AfiViewerStartupOptions.h"
#include "CalConfig.h"
#include "TcpClientStatus.h"

namespace Ui {
class ViewerMainWindow;
}

class QTableWidget;
class QSemaphore;
class QTreeWidgetItem;

class AfiViewerConfigDb;
class AfiViewerSignalRouter;
class AfiViewerConfigDispatcher;
class CalibrationDialog;
class FileReaderDialog;
class LogDialog;
class InlCheckDialog;
class TQDCAnalyzer;
class TcpClient;
class HistWorker;
class StatusDialog;
class ViewerZmqPublisher;
class WaveformViewer;
class ZmqDialog;

Q_DECLARE_METATYPE(QHostAddress)

class ViewerMainWindow : public BaseMainWindow
{
    Q_OBJECT

public:
    explicit ViewerMainWindow(AfiViewerStartupOptions startup_options, QWidget *parent = nullptr);
    ~ViewerMainWindow();

    FileReaderDialog* getFileReaderDialog() const;
    WaveformViewer* getWaveformViewer() const;

signals:
    void deleteDevice(const DeviceIndex &);
    void closeAll();
    void resetAll();
    void requestRawData(int events);
    void tcpAddressChanged(QHostAddress addr, quint16 port);
    void rawDataEveryEvUpdated(bool dataFromFile);
    void closeFile();
    void invertUpdated(QMap<DeviceIndex, QVector<bool> > invertMap) const;
    void tdcTtvxMapUpdated(TdcTtvxsMapType ttMap) const;
    void inlConfigUpdated(InlResults) const;
    void requestInlResults(bool allRes) const;
    void set_can_skip(bool skip) const;
    void configChanged(const AfiViewerConfig &);
    void switch_to_config(const ConfigIndexName &conf) const;

public slots:
    void pnpCacheUpdated(const PnpInfoCache &cache);
    void setConfig(const AfiViewerConfig &newConfig);
    void configSelectorChanged(const ConfigIndexName& conf);
    void inl_corrections_updated(InlResults res, bool allRes);

private slots:
    void on_actionConfiguration_manager_triggered();
    void on_actionSet_source_file_triggered();
    void on_actionConnect_to_source_program_triggered();
    void on_actionDisconnect_from_program_triggered();

    void on_pushButtonResetAll_clicked();

    void tcpStatusUpdated(const TcpClientStatus &status);
    void updateHitStats(const AnalyzerStat &stats);
    void rawDataUpdated(const QStringList &data);

    void on_actionCalibrationData_triggered();
    void on_actionReset_all_triggered();
    void on_actionSave_INL_corrections_triggered();
    void on_actionShow_INL_check_triggered();
    void on_actionSet_INL_separate_location_toggled(bool checked);
    void on_actionHPTDC_vol_check_triggered();
    void on_actionSet_TDC_TTVXS_map_triggered();
    void on_actionSatistic_dialog_triggered();
    void on_action0MQ_publisher_triggered();
    void on_actionHPTDC_errors_triggered();

    void on_treeWidgetChCfg_itemChanged(QTreeWidgetItem *item, int column);
    void on_radioButtonRawEvery_toggled(bool checked);
    void on_pushButtonRequestRaw_clicked();
    void on_spinBoxRawEvents_valueChanged(int limit);
    void on_checkBoxRawPeriodical_toggled(bool checked);
    void on_spinBoxRawTimer_valueChanged(int val);

    void rawDataPeriodicRequest();
    void updateStatus();

    void on_pushButtonSaveToFile_clicked();

private:
    void updateGuiToConfig();
    void write_gui_config();
    void read_gui_config();
    void initTree();
    void refreshModChNum(const AnalyzerStat &analStat);
    void updateModelList();
    QTreeWidgetItem *addChTreeItem(const DeviceIndex &index, QTreeWidgetItem *modItem, int ch);
    void updateTreeWithStats(const AnalyzerStat &analStat);
    void updateTreeItemStates();
    void updateInlMarks();
    void updateTtvxsMarks();
    void setTableRowsCount(int nch);
    void updateControlState();
    void updateSourceLabel();
    void changeSourceState(AfiViewerConfig::SOURCE_DATA st);
    void onSourceStateChange(AfiViewerConfig::SOURCE_DATA st1, AfiViewerConfig::SOURCE_DATA st2);
    QVector<ProgramDescription> getFreeProgs() const;
    bool hasFreeProgs() const;
    void updateTcpInterface(const ProgramDescription &descr);
    void updateRawDataControl();
    void disconnectFromProg();
    void setSourceColor(QString col);
    void reloadCalCalibration();

private:
    Ui::ViewerMainWindow *ui;
    QSemaphore * const analyzerSem;// free space in analyzer queue
    QSemaphore * const histSem;// free space in histogram queue
    QSemaphore * const writerSem;
    QSemaphore * const zmqSem;

    AfiViewerConfig config;
    ConfigurationManager * const configuration_manager;
    AfiViewerConfigDb * const afiViewerConfigDb;
    AfiViewerSignalRouter * const afiViewerSignalRouter;
    AfiViewerConfigDispatcher * const configDispatcher;
    CalConfig calConfig;
    FileReaderDialog * const fileReaderDialog;
    ClientIndex selectedProgram;
    ProgramInterface selectedInterface;

    QMap<DeviceIndex, QTreeWidgetItem *> modTreeItem;
    QHash<DeviceIndex, int> modChNum;
    QLabel * const dataSourceLabel;
    QLabel * const dataRateLabel;
    QLabel * const analyzerStatusLabel;
    QLabel * const semStatusLabel;
    QLabel * const vxsClockLabel;
    QLabel * const formatErrorLabel;
    LogDialog *const logDialog;
    CalibrationDialog * const calibrationDialog;
    InlCheckDialog * const inlCheckDialog;

    TcpClient * const tcpClient;
    QThread * const tcpClientThread;
    TQDCAnalyzer * const analyzer;
    QThread * const analyzerThread;
    HistWorker * const worker;
    QThread * const worker_thread;

    QTreeWidgetItem * const rootTreeItem;
    PnpInfoCache pnpCache;
    AnalyzerStat lastAnalStat;
    QStringList plaint_text;
    QString storageDir;
    QTimer * const rawDataPeriodicTimer;
    QTimer * const statusTimer;
    bool rawDataReqDone = true;
    uint rawEventCnt = 0;
    QElapsedTimer rawEventCntTimer;
    TcpClientStatus tcpClientStatus;
    StatusDialog * const statusDialog;
    ViewerZmqPublisher * const zmqPublisher;
    QThread * zmq_thread;
    ZmqDialog * const zmqDialog;
};

#endif // VIEWERMAINWINDOW_H
