/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DaqGuiApp.h"

#include <iostream>

#include <QApplication>
#include <QCommandLineParser>
#include <QSplashScreen>
#include <QStyleFactory>
#include <QStyle>
#include <QThread>

//#include "mongo/BaseConfig.h"
#include "base/Globals.h"
#include "DaqGuiApp_p.h"

DaqGuiApp::DaqGuiApp(int &_argc, char *_argv[])
    : DaqBaseApp(_argc, _argv),
    d_ptr(new DaqGuiAppPrivate(argc, _argv, this)),
    noguiOption(QCommandLineOption("nogui"))
{}

DaqGuiApp::~DaqGuiApp()
{
    delete d_ptr;
}

void DaqGuiApp::init()
{
    DaqBaseApp::init();

    QThread::currentThread()->setObjectName("GUI thread");
    if (!option_no_gui) {
        initStyle();
        showSplashScreen();
    }
    Globals::setEnvironment("production");
}

int DaqGuiApp::exec()
{
    if (!option_no_gui)
        hideSplashScreen();
    return DaqBaseApp::exec();
}

void DaqGuiApp::setupOptions()
{
    DaqBaseApp::setupOptions();
    options << noguiOption;
}

void DaqGuiApp::processParseResults(QCommandLineParser &parser)
{
    DaqBaseApp::processParseResults(parser);
    option_no_gui = parser.isSet(noguiOption) || option_config_list;
}

QCoreApplication *DaqGuiApp::createApplication() const
{
    if (option_no_gui)
        return DaqBaseApp::createApplication();
    return new QApplication(argc, argv);
}

QPixmap DaqGuiApp::getSplashPixmap() const
{
    return QPixmap(":/afi-splash.png");
}

void DaqGuiApp::initStyle()
{
    if (option_no_gui)
        return;
    auto styles = QStringList{{"Fusion", "Plastique", "Windows"}};
    for (const auto &s : styles) {
        if (QStyleFactory::keys().contains(s)) {
            QApplication::setStyle(QStyleFactory::create(s));
            break;
        }
    }
    // Workaround for inactive-window-shading
    QPalette pal = qApp->palette();
    auto roles = {QPalette::Text, QPalette::Window, QPalette::WindowText, QPalette::Button, QPalette::ButtonText};
    for (auto r : roles)
        pal.setColor(QPalette::Inactive, r,
                     pal.color(QPalette::Active, r));
    qApp->setPalette(pal);
}

void DaqGuiApp::showSplashScreen()
{
    if (option_no_gui)
        return;
    Q_D(DaqGuiApp);
    Q_ASSERT(qGuiApp);
    d->splash.reset(new QSplashScreen(getSplashPixmap()));
    d->splash->show();
    d->splash->showMessage("Starting " + QCoreApplication::applicationName(),
                           Qt::AlignBottom, QColor(0x001880));
    QApplication::processEvents();
}

void DaqGuiApp::hideSplashScreen()
{
    if (option_no_gui)
        return;
    Q_D(DaqGuiApp);
    d->splash->hide();
}
