/*
**    Copyright 2020-2022 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef GUI_UTILS_H
#define GUI_UTILS_H

#include <QComboBox>
#include <QList>
#include <QSharedPointer>
#include <QSignalBlocker>
#include <QVariant>
#include <QWidget>

Qt::CheckState toCheckState(const QVariant &v);

struct ComboBoxItemData
{
    ComboBoxItemData(const QString &text, const QVariant &data)
        : text(text), data(data) {}
    QString text;
    QVariant data;
};
using ComboBoxData = QList<ComboBoxItemData>;

void updateComboBox(QComboBox *cb, const ComboBoxData &contents);

template <typename T>
static void setComboBoxCurrentItem(QComboBox *cb, T userValue)
{
    for (int i=0; i<cb->count(); i++) {
        if (cb->itemData(i).value<T>() == userValue) {
            cb->setCurrentIndex(i);
            return;
        }
    }
}

template<class T>
class SignalBlocker
{
public:
    SignalBlocker(T *blocked)
        : blocked(blocked),
        previous(blocked->blockSignals(true)) {}
    ~SignalBlocker() { blocked->blockSignals(previous); }
    T *operator->() { return blocked; }
private:
    T *blocked;
    bool previous;
};

template<class T>
inline SignalBlocker<T> whileBlocking(T *blocked)
{
    return SignalBlocker<T>(blocked);
}

QList<QWidget *> listChildWidgetsRecursive(QWidget *p);

class MultiWidgetSignalBlocker
{
public:
    MultiWidgetSignalBlocker(const QList<QWidget *> &widgets) {
        for (auto w : widgets) {
            list.push_back(QSharedPointer<QSignalBlocker>(new QSignalBlocker(w)));
        }
    }
private:
    QList<QSharedPointer<QSignalBlocker>> list;
};

class RecursiveWidgetSignalBlocker
{
public:
    RecursiveWidgetSignalBlocker(QWidget *w)
        : blocker(listChildWidgetsRecursive(w)) {}
private:
    MultiWidgetSignalBlocker blocker;
};

using  WidgetSignalBlocker [[deprecated]] = QSignalBlocker;

class MultiSignalBlocker
{
public:
    MultiSignalBlocker(const QList<QObject *> &objects) {
        for (auto obj : objects) {
            list.push_back(QSharedPointer<QSignalBlocker>(new QSignalBlocker(obj)));
        }
    }
private:
    QList<QSharedPointer<QSignalBlocker>> list;
};

#endif // GUI_UTILS_H
