/*
**    Copyright 2020 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "LldpInfo.h"

#include <QHostAddress>
#include <QStringList>

#include "base/DeviceIndex.h"

void LldpInfo::startTimer()
{
    timer.setRemainingTime(common.ttl * 1000);
}

DeviceIndex LldpInfo::getDeviceIndex() const
{
    return DeviceIndex(static_cast<DeviceId>(ext.device_id), ext.serial_id);
}

QString LldpInfo::formatVersion(const QString &input)
{
    auto list = input.split(".");
    QStringList sl;
    for(QString item : list) {
        bool ok = false;
        int n = item.toInt(&ok);
        if (ok)
            sl += QString::number(n);
        else
            sl += item;
    }
    if (sl.size() > 2 && sl.last() == "0")
        sl.removeLast();
    return sl.join(".");
}

QString LldpInfo::getIdent() const
{
    return getDeviceIndex().getIdent();
}

QString LldpInfo::getSerialStr() const
{
    return getDeviceIndex().getSerialStr();
}

QString LldpInfo::getModelStr() const
{
    QString s = med.modelname_str.trimmed();
    if(s.isEmpty())
        s = QString("0x%1").arg(ext.device_id, 2, 16, QChar('0'));
    return s;
}

QString LldpInfo::getModelVerStr() const
{
    QString s = getModelStr();
    if (s.isEmpty())
        return s;
    if (!med.hw_str.isEmpty())
        s += "-" + med.hw_str;
    return s;
}

QString LldpInfo::getFirmWareStr() const
{
    return med.fw_str.trimmed();
}

QPair<int, int> LldpInfo::getHwVer() const
{
    return getHwVer(med.hw_str);
}

QPair<int,int> LldpInfo::getHwVer(const QString &str)
{
    QPair<int,int> ver(0,0);
    QStringList list = str.split('.');
    if(list.size()==2) {
        int a,b;
        bool ok1,ok2;
        a = list.first().toInt(&ok1);
        b = list.last().toInt(&ok2);
        if(ok1 && ok2) {
            ver.first = a;
            ver.second = b;
        }
    }
    return ver;
}

QDebug operator<<(QDebug dbg, const LldpInfo &r)
{
    dbg << QString("LLDP: {'%1', ip: %2}")
           .arg(r.getIdent())
           .arg(QHostAddress(r.common.ip_addr).toString());
    return dbg;
}
