/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "RunStatistic.h"
#include "sdb/AfiSdbId.h"

static QString formatRunTime(double time)
{
    quint64 sec = time;
    quint64 min = sec/60;
    quint64 h = min/60;
    QString str = QString("%1:%2:%3")
                      .arg(h)
                      .arg(min % 60, 2, 10, QChar('0'))
                      .arg(sec % 60, 2, 10, QChar('0'));
    return str;
}

static QString formatBusyTime(double time)
{
    return QString("%1 us").arg(time * 1e6, 0, 'f', 2);
}

QString RunStatistic::getText(ReportMode full) const
{
    if (getCounter(Accepted) == 0) return {};
    if (full == Brief)
        return getTimerStr(TimerBusyTotal);
    QStringList lines;
    lines << "Total " + getTimerStr(TimerBusyTotal);
    if (getTimer(TimerBusyAdc) > 0)
        lines << "ADC " + getTimerStr(TimerBusyAdc);
    if (getTimer(TimerBusyTdc) > 0)
        lines << "TDC " + getTimerStr(TimerBusyTdc);
    if (getTimer(TimerBusyReadout) > 0)
        lines << "Readout " + getTimerStr(TimerBusyReadout);
    if (getTimer(TimerExtVeto) > 0)
        lines << "Veto " + getTimerStr(TimerExtVeto);
    return lines.join("\n");
}

QString RunStatistic::getTimerStr(TimerIndex index) const
{
    auto time = getTimer(index);
    switch (index) {
    case TimerTotal:
    case TimerRun:
        return formatRunTime(time);
    default:
        return formatBusyTime(time);
    }
}

double RunStatistic::getTimer(TimerIndex index) const
{
    double time = timers.at(index) / clockFreqHz();
    switch (index) {
    case TimerTotal:
    case TimerRun:
        return time;
    default:
        if (!getCounter(Accepted))
            return 0;
        return time / getCounter(Accepted);
    }
}

quint64 RunStatistic::getCounter(CounterIndex index) const
{
    return counters.at(index);
}

quint64 RunStatistic::getRate(CounterIndex index) const
{
    auto t = getTimer(TimerTotal);
    return t ? (getCounter(index) / t) : 0;
}

bool RunStatistic::isValid() const
{
    return id == AFI_SDB_TDC_RUN_STAT_CNTS;
}
