/*
**    Copyright 2014-2016 Ivan Filippov
**    Copyright 2014-2020 Ilja Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef PHYSICALTRIGGERCONFIG_H
#define PHYSICALTRIGGERCONFIG_H

#include <QMetaType>
#include <QMap>

#include "base/DeviceIndex.h"
#include "device-modules/SpillControl/SpillConfig.h"
#include "device-modules/RcTrig/RcTrigConfig.h"

class QDataStream;
typedef struct __attribute__((packed)) {
    quint8 before_protect_en: 1;
    quint8 after_protect_en: 1;
    quint8 reduction_en: 1;
    quint8 : 1;
    quint16 delay:8;
    quint8 : 4;
    quint16 reduction_factor;
    quint16 before_protect_time;
    quint16 after_protect_time;
} reg_phy_rtig_ch_bit;

typedef union {
    quint64 all;
    reg_phy_rtig_ch_bit bit;
} reg_phy_rtig_ch;
Q_DECLARE_METATYPE(reg_phy_rtig_ch)

struct PhysicalTriggerChConfig
{
    PhysicalTriggerChConfig &operator=(PhysicalTriggerChConfig const &r)
    {
        name = r.name;
        reg = r.reg;
        return *this;
    }
    QString name;
    reg_phy_rtig_ch reg = {0};
};
Q_DECLARE_METATYPE(PhysicalTriggerChConfig)
bool operator==(const PhysicalTriggerChConfig &l, const PhysicalTriggerChConfig &r);
QDebug operator<<(QDebug dbg, const PhysicalTriggerChConfig &s);
QDataStream &operator<<(QDataStream &out, const PhysicalTriggerChConfig &r);
QDataStream &operator>>(QDataStream &in, PhysicalTriggerChConfig &r);

struct CalibrationTriggerConfig
{
    CalibrationTriggerConfig() = default;
    CalibrationTriggerConfig(const CalibrationTriggerConfig&) = default;
    CalibrationTriggerConfig &operator=(const CalibrationTriggerConfig&) = default;
    friend bool operator==(const CalibrationTriggerConfig &, const CalibrationTriggerConfig &) = default;
    bool en = false;
    quint32 period = 0;
    quint16 count = 0;
    quint32 offset = 0;
};
Q_DECLARE_METATYPE(CalibrationTriggerConfig)
QDebug operator<<(QDebug dbg, const CalibrationTriggerConfig &s);
QDataStream &operator<<(QDataStream &out, const CalibrationTriggerConfig &r);
QDataStream &operator>>(QDataStream &in, CalibrationTriggerConfig &r);

struct PhysicalTriggerConfig
{
    friend bool operator==(const PhysicalTriggerConfig &, const PhysicalTriggerConfig &) = default;
    enum ChMask {
        ChMaskCalib = 1 << 16,
        ChMaskRandom = 1 << 17,
        ChMaskTimer = 1 << 18,
        ChMaskExt = 1 << 19,
    };

    PhysicalTriggerConfig() {
        chCfg.resize(NCH);
    }
    inline int getNCh() const { return chCfg.size(); }

    void unifyCalibTrig()
    {
        if(calibTrigger.en) {
            startChMask |= ChMaskCalib;
            readChMask |= ChMaskCalib;
        } else {
            startChMask &= ~quint32(ChMaskCalib);
            readChMask &= ~quint32(ChMaskCalib);
        }
    }

    const static int NCH=16;
    quint16 matchWin = 2;
    quint16 trigDelay = 0;
    quint16 l0Extend = 50;

    CalibrationTriggerConfig calibTrigger;

    quint32 startChMask = 0;
    quint32 readChMask = 0;
    quint16 chWfLatency = 2;
    quint16 chWfReadPoints = 4;
    bool oscEn = true;
    bool oscDelayed = false;

    QVector<PhysicalTriggerChConfig> chCfg;

    RcTrigConfig trig;
    SpillConfig spill;
};
Q_DECLARE_METATYPE(PhysicalTriggerConfig)
QDebug operator<<(QDebug dbg, const PhysicalTriggerConfig &s);
QDataStream &operator<<(QDataStream &out, const PhysicalTriggerConfig &r);
QDataStream &operator>>(QDataStream &in, PhysicalTriggerConfig &r);


#endif // PHYSICALTRIGGERCONFIG_H
