#include "CrosspointSwitchDilog.h"
#include "ui_CrosspointSwitchDilog.h"

#include <QComboBox>
#include <QDebug>
#include <QLabel>
#include <optional>

CrosspointSwitchDilog::CrosspointSwitchDilog(QWidget *parent)
    : QDialog(parent)
    , ui(new Ui::CrosspointSwitchDilog)
{
    ui->setupUi(this);
}

CrosspointSwitchDilog::~CrosspointSwitchDilog()
{
    delete ui;
}

void CrosspointSwitchDilog::crosspointSwitchBpUpdated(const DeviceIndex &, const CrosspointSwitchBuildParams &bp)
{
    if(buildParams) {
        if(buildParams->outSize != bp.outSize || buildParams->inSize != bp.inSize)
            qWarning() << QString("Missmatch of CrosspointSwitchBuildParams: "
                                  "out:%1 in:%2 <> out:%3 in:%4")
                          .arg(buildParams->outSize).arg(buildParams->inSize)
                          .arg(bp.outSize).arg(bp.inSize);
        return;
    }
    buildParams = bp;
    updateGui();
}

void CrosspointSwitchDilog::setCrosspointSwitchConfig(const CrosspointSwitchConfig &cfg)
{
    if(config && *config == cfg)
        return;
    config = cfg;
    updateGui();
}

void CrosspointSwitchDilog::cbCurrentIndexChanged()
{
    auto cb = qobject_cast<QComboBox*>(sender());
    if(!cb)
        return;
    auto par = qobject_cast<QGroupBox*>(cb->parent());
    if(!par)
        return;
    auto formLayout = qobject_cast<QFormLayout*>(par->layout());
    bool typeWf = par == ui->groupBoxWfRec;
    std::optional<quint16> row;
    for(auto i=0; i<getOutSize(); ++i) {
        auto item = formLayout->itemAt(i, QFormLayout::FieldRole);
        if(!item)
            continue;
        if(item->widget() == cb) {
            row = i;
            break;
        }
    }
    if(!row)
        return;
    quint16 id = *row;
    if(!typeWf)
        id += getOutSize();

    quint16 val = cb->currentData().toUInt();
    if(config->out.value(id, -1) != val) {
        config->out[id] = val;
        emit crosspointSwitchConfigChanged(*config);
    }
}

quint16 CrosspointSwitchDilog::getOutSize() const
{
    if(!buildParams)
        Q_UNREACHABLE();
    return buildParams->outSize/2;
}

quint16 CrosspointSwitchDilog::getInputSize() const
{
    if(!buildParams)
        Q_UNREACHABLE();
    return buildParams->inSize;
}

void CrosspointSwitchDilog::updateGui()
{
    if(!buildParams || !config)
        return;
    if(config->out.size() != buildParams->outSize)
        adoptCfgToBp();
    for(auto i=0; i<getOutSize(); ++i) {
        updateCb(true, i);
        updateCb(false, i);
    }
}

void CrosspointSwitchDilog::updateCb(bool typeWf, quint16 index) const
{
    auto p = typeWf ? ui->groupBoxWfRec : ui->groupBoxDigTrig;
    auto cb = getCb(p->layout(), index);
    if(!cb)
        cb = createCb(typeWf, index);
    updateCbItems(cb);

//    QSignalBlocker block(cb);
    cb->setCurrentIndex(config->get(index, buildParams->outSize));
}

QComboBox *CrosspointSwitchDilog::createCb(bool typeWf, quint16 index) const
{
    auto cbPar = typeWf ? ui->groupBoxWfRec : ui->groupBoxDigTrig;
    auto layout = qobject_cast<QFormLayout*>(cbPar->layout());

    auto lab = new QLabel(QString("Out #%1").arg(index), cbPar);
    layout->setWidget(index, QFormLayout::LabelRole, lab);

    auto cb = new QComboBox(cbPar);
    layout->setWidget(index, QFormLayout::FieldRole, cb);
    connect(cb, QOverload<int>::of(&QComboBox::currentIndexChanged),
            this, &CrosspointSwitchDilog::cbCurrentIndexChanged);
    return cb;
}

void CrosspointSwitchDilog::updateCbItems(QComboBox *cb) const
{
    QSignalBlocker block(cb);
    auto getName = [](int i, int size) { return QString("Ch #%1 %2")
                .arg(i%(size/2))
                .arg(i>=(size/2) ? "after DSP" : "Raw"); };

    auto size = getInputSize();
    auto upSize = std::min(quint16(cb->count()), size);
    // up
    for(auto i=0; i<upSize; ++i) {
        cb->setItemText(i, getName(i, size));
        cb->setItemData(i, i);
    }
    // add
    for(auto i=upSize; i<size; ++i) {
        cb->addItem(getName(i, size), i);
    }
    while(cb->count()>size) {
        cb->removeItem(size);
    }
}

void CrosspointSwitchDilog::adoptCfgToBp()
{
    config = getConfigAfterDsp();
    emit crosspointSwitchConfigChanged(*config);
}

CrosspointSwitchConfig CrosspointSwitchDilog::getConfigAfterDsp() const
{
    auto size = getOutSize();
    CrosspointSwitchConfig cfg;
    cfg.out.resize(size*2);
    for(auto i=0; i<size; ++i)
        cfg.out[i] = i+size;
    for(auto i=0; i<size; ++i)
        cfg.out[i+size] = i+size;
    return cfg;
}

CrosspointSwitchConfig CrosspointSwitchDilog::getConfigBeforeDsp() const
{
    auto size = getOutSize();
    CrosspointSwitchConfig cfg;
    cfg.out.resize(size*2);
    for(auto i=0; i<size; ++i)
        cfg.out[i] = i;
    for(auto i=0; i<size; ++i)
        cfg.out[i+size] = i;
    return cfg;
}

QComboBox *CrosspointSwitchDilog::getCb(QLayout *l, int row) const
{
    auto fl = qobject_cast<QFormLayout*>(l);
    if(fl) {
        auto i = fl->itemAt(row, QFormLayout::FieldRole);
        if(i) {
            auto cb = qobject_cast<QComboBox*>(i->widget());
            if(cb)
                return cb;
        }
    }
    return nullptr;
}

void CrosspointSwitchDilog::on_pushButtonResetAfterDsp_clicked()
{
    emit crosspointSwitchConfigChanged(getConfigAfterDsp());
}

void CrosspointSwitchDilog::on_pushButtonResetRaw_clicked()
{
    emit crosspointSwitchConfigChanged(getConfigBeforeDsp());
}
