/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "WaveRecWidget.h"
#include "base/gui_utils.h"
#include "ui_WaveRecWidget.h"

#include <QSettings>

namespace {
void setWidgetWarn(QWidget *w, const QStringList &warnList)
{
    bool en = w->isEnabled();
    auto c = en ? (warnList.isEmpty() ? QColor()         : QColor(0xFF8800)) :
        (warnList.isEmpty() ? QColor(0xbfbfbf) : QColor(0xFFCC44));
    auto p = w->palette();

    auto b = p.brush(w->foregroundRole());
    if(p.color(w->foregroundRole()) != c) {
        p.setColor(w->foregroundRole(), c);
        w->setPalette(p);
    }

    QString tt = warnList.join('\n');
    if(w->toolTip() != tt)
        w->setToolTip(tt);
}
}

WaveRecWidget::WaveRecWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::WaveRecWidget)
{
    ui->setupUi(this);

    auto sb = ui->spinBoxThreshold;
    {
        QSignalBlocker blocker(sb);
        sb->setSuffix(" mV");
        sb->setMinimum(-3000);
        sb->setMaximum(3000);
    }
    {
        auto cb = ui->comboBoxZsMode;
        QSignalBlocker blocker(cb);
        cb->addItem("Disabled", static_cast<int>(WaveRecConfig::ZsDisabled));
        cb->addItem("Analog", static_cast<int>(WaveRecConfig::ZsAnlg));
        cb->addItem("Digital", static_cast<int>(WaveRecConfig::ZsDig));
    }

    ui->labelAdcPreWin->hide();
    ui->labelAdcPreWinNs->hide();
    ui->spinBoxAdcPreWin->hide();
//    ui->spinBoxThreshold->hide();
//    ui->labelThreshold->hide();
}

WaveRecWidget::~WaveRecWidget()
{
    delete ui;
}

void WaveRecWidget::setWaveRecConfig(const WaveRecConfig &newConfig)
{
    if (config == newConfig)
        return;
    config = newConfig;

    restoreGuiFromConfig();
}

void WaveRecWidget::restoreGuiFromConfig()
{
    RecursiveWidgetSignalBlocker blocker(this);

    ui->comboBoxZsMode->setCurrentIndex(static_cast<int>(config.zsMode));
    ui->spinBoxAdcWin->setEnabled(config.zsMode!=WaveRecConfig::ZsDisabled);
    ui->spinBoxAdcPreWin->setEnabled(config.zsMode!=WaveRecConfig::ZsDisabled);


    ui->spinBoxAdcPreWin->setValue(config.adcPreWin);
    ui->spinBoxAdcWin->setValue(config.readCellNumber);
    ui->spinBoxLatency->setValue(config.latency);
    updateMatchWinMaximum();
    ui->spinBoxMatchWin->setValue(config.matchWin);
    updateWidgetWarnings();
}

void WaveRecWidget::updateWidgetWarnings()
{
//    const auto &ds= config.defaultSetup.tqdcCore;
//    QStringList warnList;
//    if(config.defaultSetup.waveRec.zsEn && ((ds.matchWin*3 + ds.readCellNumber) >= (670*3)))
//        warnList << "MatchWin+AdcWin > 16us";

//    setWidgetWarn(ui->checkBoxAdcZs, warnList);
//    setWidgetWarn(ui->labelAdcWinNs, warnList);

//    if(ds.matchWin > ds.latency)
//        warnList << "MatchWin > Latency";

//    setWidgetWarn(ui->labelMatchWinNs, warnList);
}

void WaveRecWidget::updateMatchWinMaximum() const
{
    const QSettings settings;
    auto matchWinLimitEn = settings.value("matchWinLimitEn").toBool();

    int v = matchWinLimitEn ? config.latency : 670;
    ui->spinBoxMatchWin->setMaximum(std::min(v, 670));
}

void WaveRecWidget::on_comboBoxZsMode_currentIndexChanged(int index)
{
    bool zsEn = index != WaveRecConfig::ZsDisabled;
    ui->spinBoxAdcWin->setEnabled(zsEn);
    ui->spinBoxAdcPreWin->setEnabled(zsEn);

    config.zsMode = static_cast<WaveRecConfig::ZsMode>(index);
    emit waveRecConfigChanged(config);
}

void WaveRecWidget::on_spinBoxThreshold_valueChanged(int thr)
{
//    config.defaultSetup.tqdcCore.setThreshold(thr);
//    for(auto &ds : config.knownConfigs) {
//        ds.tqdcCore.setThreshold(thr);
//    }
//    emit waveRecConfigChanged(config);
}

void WaveRecWidget::on_spinBoxLatency_valueChanged(int lat)
{
    ui->labelLatencyNs->setText(QString("%1 ns").arg(lat*24));
    config.latency = lat;

    const QSettings settings;
    auto matchWinLimitEn = settings.value("matchWinLimitEn").toBool();
    if (matchWinLimitEn)
        updateMatchWinMaximum();
    updateWidgetWarnings();
    emit waveRecConfigChanged(config);
}

void WaveRecWidget::on_spinBoxMatchWin_valueChanged(int win)
{
    ui->labelMatchWinNs->setText(QString("%1 ns").arg(win*24));
    config.matchWin = win;
    updateWidgetWarnings();
    emit waveRecConfigChanged(config);
}

void WaveRecWidget::on_spinBoxAdcWin_valueChanged(int val)
{
    ui->labelAdcWinNs->setText(QString("%1 ns").arg(val*8));

    config.readCellNumber = val;
    emit waveRecConfigChanged(config);
}

void WaveRecWidget::on_spinBoxAdcPreWin_valueChanged(int preWin)
{
    ui->labelAdcPreWinNs->setText(QString("%1 ns").arg(preWin*8));
    config.adcPreWin = preWin;
    emit waveRecConfigChanged(config);
}
