#include "TelnetRemoteControl.h"
#include "TelnetRemoteControl/TelnetBlockedEvents.h"

#include <QHostInfo>
#include <QTimer>

namespace {
const char * RC_CMD_START = "start";
const char * RC_CMD_STOP  = "stop";
const char * RC_CMD_BLOCKED_EVENTS  = "blockedEvents";
const char * RC_CMD_STATE = "state";
const char * RC_CMD_SET   = "set";
const char * RC_CMD_HELP   = "help";
const char * RC_CMD_END   = "\r\n";
}

TelnetRemoteControl::TelnetRemoteControl(quint16 port, QObject *parent)
    : QObject(parent)
    , tcpPort(port)
    , qTcpServer(new QTcpServer(this))
{
    setObjectName("TelnetRemoteControl");
    Q_ASSERT(tcpPort);
    QTimer::singleShot(0, this, &TelnetRemoteControl::init);
}

TelnetRemoteControl::~TelnetRemoteControl()
{
    qTcpServer->close();
    if(tcpSocket){
        tcpSocket->abort();
        delete tcpSocket;
        tcpSocket = nullptr;
        portInUse=false;
    }
    qDebug() << QString("TcpServer closed on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
}

void TelnetRemoteControl::timedRunFinished(const TelnetBlockedEvents &evBloked)
{
    QStringList strList;
    for(auto it=evBloked.cbegin(); it!=evBloked.cend(); ++it)
        strList << it.key().getSerialStr() << QString::number(*it);
    sendString(QString("%1 %2").arg(RC_CMD_BLOCKED_EVENTS).arg(strList.join("; ")));
}

void TelnetRemoteControl::onFsmStateChange(const FsmState &state)
{
    fsmSt = state;
    sendState();
}

void TelnetRemoteControl::updateConnections()
{
    while(QTcpSocket *pendingConnection = qTcpServer->nextPendingConnection()) {
        if(tcpSocket && (!tcpSocket->isValid() || tcpSocket->state() != QTcpSocket::ConnectedState)){
            tcpSocket->abort();
            delete tcpSocket;
            tcpSocket = nullptr;
            portInUse=false;
        }
        if(tcpSocket==nullptr){
            tcpSocket = pendingConnection;
            portInUse=true;
            connect(tcpSocket, &QTcpSocket::readyRead, this, &TelnetRemoteControl::readCommandRequest);
            connect(tcpSocket, &QTcpSocket::disconnected, this, &TelnetRemoteControl::onDisconnected);
            QHostInfo info = QHostInfo::fromName(tcpSocket->peerAddress().toString());
            QString msg = QString("TelnetRemoteControl accepted connection from %1:%2")
                    .arg(info.hostName())
                    .arg(tcpSocket->peerPort());
            qInfo() << msg;
            sendState();
        } else {
            pendingConnection->abort();
            delete pendingConnection;
        }
    }
    if(tcpSocket==nullptr){
        qDebug()<<"There are no connections.";
    }
}

void TelnetRemoteControl::readCommandRequest()
{
    while(tcpSocket->bytesAvailable()){
        QByteArray data = tcpSocket->readAll();
        cachedData.append(data);

        QString mess = QString::fromUtf8(data.data(), data.size());
        for(QString command : mess.split(RC_CMD_END, Qt::SkipEmptyParts)) {
            command = command.trimmed();
            qInfo() << QString("Remote command:").append(command);
        }
    }
    if(!cachedData.isEmpty()) {
        decode();
    }
}

void TelnetRemoteControl::onDisconnected()
{
    portInUse=false;
    qInfo() << "TelnetRemoteControl disconnected";
}

void TelnetRemoteControl::init()
{
    bool ok = qTcpServer->listen(QHostAddress::Any, tcpPort);
    if(!ok) {
        qFatal("Failed to start %s on port %d: %s",
               objectName().toStdString().c_str(), tcpPort,
               qTcpServer->errorString().toStdString().c_str());
    }
//    tcpPort = qTcpServer->serverPort();

    connect(qTcpServer, &QTcpServer::newConnection,
            this, &TelnetRemoteControl::updateConnections);
    qInfo() << QString("TelnetRemoteControl listening on %1:%2").arg(qTcpServer->serverAddress().toString()).arg(tcpPort);
}

void TelnetRemoteControl::setParam(const QString &key, const QVariant &val)
{
    sendWarn("Setting params does not implemented");
}

void TelnetRemoteControl::decode()
{
    int pos=-1;
    while((pos=cachedData.indexOf(RC_CMD_END)) != -1) {
        auto data = cachedData.left(pos);
        cachedData.remove(0, pos+static_cast<int>(sizeof (RC_CMD_END)));

        auto str = QString::fromLatin1(data);
        auto strList = str.split(' ', Qt::SkipEmptyParts);
        if(strList.isEmpty())
            continue;
        const auto &cmd = strList.first();
        if(cmd == RC_CMD_START) {
            qInfo("Got request for startTelnetRun");
            emit startRun();
        } else if(cmd == RC_CMD_STOP) {
            qInfo() << "emit stopRun();";
            emit stopRun();
        } else if(cmd == RC_CMD_SET) {
            if(strList.size() < 3) {
                qCritical() << "Got incorrect command from TelnetRemoteControl:" << str;
                continue;
            }
            const auto &paramKey = strList[1];
            QVariant paramVal(strList[2]);
            qInfo() << QString("Got request for setting param '%1' to '%2'")
                       .arg(paramKey, paramVal.toString());
            setParam(paramKey, paramVal);
        } else if(cmd == RC_CMD_STATE) {
            sendState();
        } else if(cmd == RC_CMD_HELP) {
            printHelp();
        } else {
            sendWarn("Unknown command:"+str);
            printHelp();
        }
    }
}

bool TelnetRemoteControl::isRunControlConnected() const
{
    return tcpSocket != nullptr && tcpSocket->isValid() && (tcpSocket->state() == QTcpSocket::ConnectedState);
}

void TelnetRemoteControl::sendWarn(const QString &s) const
{
    QString warn = "warn:" + s;
    qWarning() << s;
    sendString(s);
}

void TelnetRemoteControl::sendString(const QString &mess) const
{
    if (!isRunControlConnected())
        return;
    QString line = mess;
    if(!line.endsWith(RC_CMD_END))
        line += RC_CMD_END;

    QByteArray messData = line.toUtf8();
    const auto written = tcpSocket->write(messData);
    if(written != static_cast<qint64>(messData.size()))
        qWarning()<<"Failed to send info string to run control client";
    else
        qDebug()<<"sent to run control client:" << mess.trimmed();
}

void TelnetRemoteControl::printHelp() const
{
    QStringList list;
    const char * SEP = "-------------";
    list << "";
    list << "Commands:";
    list << QString("%1 - start run execution").arg(RC_CMD_START);
    list << QString("%1 - stop run execution").arg(RC_CMD_STOP);
    list << QString("%1 <key> <value> - setting value in program configuration").arg(RC_CMD_SET);
    list << QString("%1 - print help").arg(RC_CMD_HELP);
    list << SEP;
    list << "Status:";
    list << QString("%1 <stateStr> - current program state").arg(RC_CMD_STATE);
    list << QString("%1 <dev serial #1> <blockedEv #1>; <dev serial #2> <blockedEv #2>; ... - number of blocked events in devices").arg(RC_CMD_BLOCKED_EVENTS);
    sendString(list.join(RC_CMD_END));
}

void TelnetRemoteControl::sendState() const
{
    sendString(QString("%1 %2").arg(RC_CMD_STATE).arg(fsmStateToString(fsmSt)));
}
