#ifndef HISTMNGRCFG_H
#define HISTMNGRCFG_H

#include <QHash>
#include <QMap>
#include <QMetaType>
#include <QVector>

#include "base/DaqModule.h"
#include "base/DeviceIndex.h"

const int HIST_MATCHWIN_MAX = 10000;
const int HIST_MATCHWIN_MIN = 0;

enum class HistId {
    unknown = 0,
    tdcTime = 1,
    tdcDoublePulseTime = 2,
    tdcCh = 3,
    tdcHitEv = 4,
    errors = 5,
    tdcTaiTime  = 6,
    tdcWidth = 7,
    adcTime = 8,
    tdcAdcDt = 9,
    adcCh = 10,
    adcHitEv = 11,
    chSamples = 12,
    chAmpl = 13,
    inl = 14,
    inlCheck = 15,
    mscPlot  = 16,
};
Q_DECLARE_METATYPE(HistId)

uint qHash(const HistId &key, uint seed = 0);
QString getStrKeyForHistId(HistId id);
HistId getHistIdForStrKey(QString key);

struct HistKey {
    HistKey():
        HistKey(HistId::unknown, DeviceIndex(), -1)
    {}
    HistKey(HistId id, DeviceIndex index = DeviceIndex() , int ch = -1):
        hist_id(id),
        index(index),
        ch(ch)
    {}
    void print_hist_key() const;
    bool operator <(const HistKey &key) const{
        if (hist_id < key.hist_id)
            return true;
        if (hist_id > key.hist_id)
            return false;
        if(index < key.index)
            return true;
        if(key.index < index)
            return false;
        if(ch < key.ch)
            return true;
        if(ch < key.ch)
            return false;
        return false;
    }
    friend bool operator==(const HistKey &, const HistKey &) = default;

    friend uint qHash(const HistKey &key, uint seed = 0) {
        return qHash(key.hist_id, seed) ^ qHash(key.index, seed)
                ^ qHash(key.ch, seed);
    }
    bool isEmpty() const { return hist_id == HistId::unknown; }

    HistId hist_id;
    DeviceIndex index;
    int ch;
};
Q_DECLARE_METATYPE(HistKey)

struct HistGroupInfo
{
    HistGroupInfo() = default;

    HistGroupInfo(HistId group_id, QString key, QString title, int n_bins, double min, double max,
                  QString toolTip)
        :
          strKey(key),
          title(title),
          tool_tip(toolTip),
          histId(group_id),
          n_bins(n_bins),
          min(min),
          max(max)
    {
    }

    QString strKey = "unknown";
    QString title;
    QString tool_tip;
    HistId histId = HistId::unknown;
    int n_bins = 0;
    double min = 0;
    double max = 0;
};

struct HistGroupCfg
{
    HistGroupCfg() = default;

    HistGroupCfg(HistId group_id) : hist_id(group_id)
    {
    }
    friend bool operator==(const HistGroupCfg &, const HistGroupCfg &) = default;
    static bool isBaseHist(HistId histGroupId);
    bool isBaseHist() const { return isBaseHist(hist_id); }
    HistGroupInfo getHistGroupInfo() const { return getHistGroupInfo(hist_id); }
    static HistGroupInfo getHistGroupInfo(HistId id);

    HistId hist_id = HistId::unknown;
    bool hist_group_en = true;
    bool show_fit = true;
};
Q_DECLARE_METATYPE(HistGroupCfg)

struct HistMscCfg
{
    HistMscCfg() = default;
    friend bool operator==(const HistMscCfg &, const HistMscCfg &) = default;

    double width = 10; // in sec
    uint nbins = 1000;
};
Q_DECLARE_METATYPE(HistMscCfg)

struct HistMngrCfg
{
    HistMngrCfg() = default;
    friend bool operator==(const HistMngrCfg &, const HistMngrCfg &) = default;

    void init(const DaqModule &daq_module_ = DaqModule(),
              bool is_inl_enabled_ = false);
    HistKey get_cur_hist_key() const;
    HistGroupInfo getHistGroupInfo(HistId id) const;
    void makeConfigValid();
private:
    void addHistGroup(HistId id);

public:
    DaqModule daq_module;
    bool is_inl_enabled = false;
    HistId cur_hist_id = HistId::unknown;

    DeviceIndex cur_dev_index;
    int cur_ch = 0;
    bool curEdgeLead = true;
    bool curEdgeTrail = false;
    bool use_stop_ch = false; // worker
    DeviceIndex stop_dev_index; // worker
    quint16 stop_ch = 0; // worker
    bool stopEdgeLead = true;
    bool stopEdgeTrail = false;

    bool inl_hist_code_dens = true; // hist
    bool global_enabled = true; // worker
    int tai_ts_window = 3600; // in sec, worker
    int mw_min = 0; // match window min in ns
    int mw_max = 1000; // match window max in ns
    bool hitTimeEdgeLead = true;
    bool hitTimeEdgeTrail = false;
    bool log_y_axis = false;
    bool hide_zero_bin = false;
    bool ampl_hist_with_max = true; // fill adcAmpl with max or min value
    QMap<HistId, HistGroupCfg> hist_groups; // by hist group id
    QMap<HistKey, HistMscCfg> histMscCfg; // by hist group id
    QString histExportPath;
};
Q_DECLARE_METATYPE(HistMngrCfg)

QDebug operator<<(QDebug dbg, const HistId &r);
QDebug operator<<(QDebug dbg, const HistKey &r);
QDebug operator<<(QDebug dbg, const HistGroupCfg &c);
QDebug operator<<(QDebug dbg, const HistMngrCfg &c);
QDataStream &operator<<(QDataStream &out, const HistGroupCfg &r);
QDataStream &operator>>(QDataStream &in, HistGroupCfg &r);
QDataStream &operator<<(QDataStream &out, const HistMngrCfg &r);
QDataStream &operator>>(QDataStream &in, HistMngrCfg &r);

#endif // HISTMNGRCFG_H

