/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "ReadoutAppMainWindow.h"

#include <QLabel>
#include <QMenu>
#include <QMenuBar>
#include <QMessageBox>
#include <QShortcut>
#include <QSignalBlocker>
#include <QStatusBar>

#include "AdcMpdTrigDialog/AdcMpdTrigDialog.h"
#include "AdcSerDesCalDialog/AdcSerDesCalDialog.h"
#include "ClockControl/ClockControlDialog.h"
#include "CrosspointSwitchDilog/CrosspointSwitchDilog.h"
#include "FeLinkDialog/FeLinkStatusDialog.h"
#include "HptdcStatusDialog/HptdcStatusDialog.h"
#include "NetworkPortWidget/NetworkPortConfigDialog.h"
#include "base/DaqModuleTraits.h"
#include "base/Globals.h"
#include "configuration-manager/ConfigurationManager.h"
#include "device-discover/DiscoverDialog.h"
#include "device-utils/AdcDeserCheckStatus.h"
#include "m-stream-output/MStreamStatisticDialog.h"
#include "remote-control-server/RemoteControlServerStatus.h"
#include "sdb/AfiSdbId.h"
#include "SparseDialog/SparseDialog.h"
#include "widgets/DeviceStatusDialog.h"
#include "zmq-lib/ZmqDialog.h"

namespace {
constexpr static const char* k_show_run_logic = "showRunLogic";
constexpr auto k_des_check = "des_check";
}

ReadoutAppMainWindow::ReadoutAppMainWindow(QWidget *parent)
    : PnpBaseMainWindow(parent)
    , adcMpdTrigDialog(new AdcMpdTrigDialog(this))
    , adcSerDesCalDialog(new AdcSerDesCalDialog(this))
    , clockControlDialog(new ClockControlDialog(this))
    , crosspointSwitchDilog(new CrosspointSwitchDilog(this))
    , deviceStatusDialog(new DeviceStatusDialog(this))
    , discoverDialog(new DiscoverDialog(this))
    , feLinkStatusDialog(new FeLinkStatusDialog(this))
    , hptdcStatusDialog(new HptdcStatusDialog(this))
    , mStreamStatisticDialog(new MStreamStatisticDialog(this))
    , networkPortConfigDialog(new NetworkPortConfigDialog(this))
    , sparseDialog(new SparseDialog(this))
    , zmqDialog(new ZmqDialog(this))
{
    qRegisterMetaType<ConfigIndexName>();
    qRegisterMetaType<ModularDeviceStatus>();
    qRegisterMetaType<ZmqConfig>();

    connect(this, &ReadoutAppMainWindow::setStatus, deviceStatusDialog, &DeviceStatusDialog::setStatus);
    configuration_manager->setDefaultConfig(BaseDeviceAppConfig());
}

QAction *ReadoutAppMainWindow::addToolAction(const QString &name,
                                         void (ReadoutAppMainWindow::*pt2ConstMember)() const,
                                         const QVariantList sdbList)
{
    auto action = menuTools->addAction(name, this, pt2ConstMember);
    if(!sdbList.isEmpty()) {
        action->setProperty(k_sdb_list, sdbList);
        action->setVisible(false);
    }
    return action;
}

void ReadoutAppMainWindow::updateAdcDesCheckLabel(const QVariantMap &map)
{
    auto globSt = AdcDeserCheckStatus::Status::Ok;
    QStringList tt;
    for(auto it = map.begin(); it != map.end(); ++it) {
        auto devSt = it->value<AdcDeserCheckStatus>();
        auto locSt = devSt.getStatus();
        if(locSt == AdcDeserCheckStatus::Status::Ok)
            continue;
        QStringList devStStr;
        for(auto desIt = devSt.failedDes.cbegin(); desIt != devSt.failedDes.cend(); ++desIt) {
            QStringList pointStr;
            for(auto i : *desIt)
                pointStr << QString::number(i);
            devStStr <<  QString("[%1,%2] - %3").arg(desIt.key().first).arg(desIt.key().first).arg(pointStr.join(','));
        }
        tt << QString("%1: %2").arg(it.key(), devStStr.join("; "));
        globSt = locSt;
        if(globSt == AdcDeserCheckStatus::Status::Error)
            break;
    }
    adcDesCheckLabel->setVisible(globSt != AdcDeserCheckStatus::Status::Ok);

    QString labelText = "DesCheck - ";
    QColor c;
    switch (globSt) {
    case AdcDeserCheckStatus::Status::Ok:
        labelText += "ok";
        break;
    case AdcDeserCheckStatus::Status::Warning:
        labelText += "warn";
        c = Qt::darkYellow;
        break;
    case AdcDeserCheckStatus::Status::Error:
        labelText += "fail";
        c = Qt::red;
        break;
    default: Q_UNREACHABLE();
    }

    adcDesCheckLabel->setText(labelText);

    if(!tt.isEmpty())
        tt.prepend("<serial>: [clock, data] - <failed des.>");
    adcDesCheckLabel->setToolTip(tt.join('\n'));

    auto p = adcDesCheckLabel->palette();
    p.setColor(adcDesCheckLabel->foregroundRole(), c);
    adcDesCheckLabel->setPalette(p);
}

void ReadoutAppMainWindow::sendShowRunLogicState() const
{
    emit showRunLogicChanged(menuItemShowRunLogic->isChecked() && menuItemShowRunLogic->isVisible());
}

void ReadoutAppMainWindow::sendErrOnFeLink(bool) const
{
    emit errOnFeLinkErrChanged(menuItemErrorOnFeLinkErr->isChecked(), menuItemWarnOnFeLinkErr->isChecked());
}

void ReadoutAppMainWindow::setPeriodicalDeserializerCheck(bool en)
{
    config.defaultSetup.periodicalAdcDelayCheck = en;
    config.unify_config();
    emit configChanged(config);
}

void ReadoutAppMainWindow::confirmAdcClockDividerSync() const
{
    QMessageBox msgBox;
    msgBox.setWindowTitle("Sync ADC Clock divider...");
    msgBox.setText("Are you sure to perform clock divider syncronization?");
    msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    msgBox.setDefaultButton(QMessageBox::No);
    int ret = msgBox.exec();
    if(ret == QMessageBox::Yes)
        emit requestAdcClkDivSync();
}

void ReadoutAppMainWindow::addMenuFileActions()
{
    createMenuIfAbsent(&menuFile);

    menuItemDiscoverDialog = menuFile->addAction(
                tr("Select devices..."),
                this,
                &ReadoutAppMainWindow::actionDiscoverDialog,
                QKeySequence::Find);

    PnpBaseMainWindow::addMenuFileActions();
}

void ReadoutAppMainWindow::addMenuOptionsActions()
{
    PnpBaseMainWindow::addMenuOptionsActions();
    createMenuIfAbsent(&menuOptions);

    auto addOptAction = [this](const QString &name,
                               auto pt2ConstMember,
                               AFI_SDB_ID sdb)
    {
        auto a = new QAction(name, menuOptions);
        a->setCheckable(true);
        a->setChecked(true);
        connect(a, &QAction::toggled, this, pt2ConstMember);
        a->setProperty(k_sdb_list, QVariantList{sdb});
        a->setVisible(false);
        menuOptions->insertAction(menuOptions->actions().value(1, nullptr), a);
        return a;
    };

    menuItemShowRunLogic = addOptAction(tr("Show 'Run Logic' section"),
                                        &ReadoutAppMainWindow::sendShowRunLogicState,
                                        AFI_SDB_RUN_LOGIC);
    connect(menuItemShowRunLogic, &QAction::changed,
            this, &ReadoutAppMainWindow::sendShowRunLogicState); // change of visiblity
    menuItemShowRunLogic->setToolTip("Run logic widget should be always visible if any limit is set");

    menuItemWarnOnFeLinkErr = addOptAction(tr("Send run warn on FeLink err"),
                                           &ReadoutAppMainWindow::sendErrOnFeLink,
                                           AFI_SDB_FE_LINK_CSR);
    menuItemErrorOnFeLinkErr = addOptAction(tr("Send run error on FeLink err"),
                                            &ReadoutAppMainWindow::sendErrOnFeLink,
                                            AFI_SDB_FE_LINK_CSR);
    menuItemErrorOnXOff = addOptAction(tr("Send run error on trig XOff"),
                                       &ReadoutAppMainWindow::errOnTrigOnXOffChanged,
                                       AFI_SDB_TDC_RUN_STAT_CNTS);
    menuItemErrorOnSerDesUnlock = addOptAction(tr("Send run error on SerDes unlock"),
                                               &ReadoutAppMainWindow::errOnSerDesUnlockChanged,
                                               AFI_SDB_Deserializers_Contr);
}

void ReadoutAppMainWindow::addMenuToolsActions()
{
    createMenuIfAbsent(&menuTools);

    addMenuToolsConfigureActions();

    menuTools->addSection("Show status");

    menuTools->addAction(tr("Device status..."),
                     this,
                     &ReadoutAppMainWindow::actionDeviceStatus);
    addToolAction(tr("Fe-Link status..."),
                  &ReadoutAppMainWindow::actionFelinkStatus,
                  {AFI_SDB_FE_LINK_CSR});

    if (DaqModuleTraits::hasHptdc(Globals::getDaqModuleType())) {
        addToolAction(tr("HPTDC status..."),
                      &ReadoutAppMainWindow::actionHptdcStatus,
                      {AFI_SDB_HPTDC_RegIO_JTAG});
    }
    if (DaqModuleTraits::hasMstream(Globals::getDaqModuleType())) {
        addToolAction(tr("MStream statistics..."),
                      &ReadoutAppMainWindow::actionMstreamStatistics);
    }
    menuTools->addSection("");

    PnpBaseMainWindow::addMenuToolsActions();
}

void ReadoutAppMainWindow::addMenuToolsConfigureActions()
{
    menuTools->addSection("Configure");

    addToolAction(tr("Sync ADC Clock divider..."),
                  &ReadoutAppMainWindow::confirmAdcClockDividerSync,
                  {AFI_SDB_Clock_Divider_Contr});
    addToolAction(tr("ADC Ser-Des calibration..."),
                  &ReadoutAppMainWindow::actionAdcSerDesCalibration,
                  {AFI_SDB_Deserializers_Contr});
    menuItemPeriodicalDesCheck = addToolAction(tr("Periodical deserializer check"),
                                               nullptr,
                                               {AFI_SDB_Deserializers_Contr});
    menuItemPeriodicalDesCheck->setCheckable(true);
    menuItemPeriodicalDesCheck->setChecked(config.defaultSetup.periodicalAdcDelayCheck);
    connect(menuItemPeriodicalDesCheck, &QAction::toggled,
            this, &ReadoutAppMainWindow::setPeriodicalDeserializerCheck);


    addToolAction(tr("Network port config..."),
                  &ReadoutAppMainWindow::actionNetworkPortConfig,
                  {AFI_SDB_ETH_PORT_V1});
    addToolAction(tr("Sparse config..."), &ReadoutAppMainWindow::actionSparseDialog,
                  {AFI_SDB_WAVE_REC, AFI_SDB_ADC_ERC});
    addToolAction(tr("ADC MPD trig config..."), &ReadoutAppMainWindow::actionAdcMpdTrigDialog,
                  {AFI_SDB_ADC_MPD_TRIG});
    addToolAction(tr("Clock control config..."),
                  &ReadoutAppMainWindow::actionClockControlConfig,
                  {AFI_SDB_Clock_Control});
    addToolAction(tr("Crosspoint switch config..."),
                  &ReadoutAppMainWindow::actionCpsConfig,
                  {AFI_SDB_CROSSPOINT_SWITCH});
    addToolAction(tr("Device status ZMQ publisher..."),
                  &ReadoutAppMainWindow::actionDeviceStatusZmqConfig);
}

void ReadoutAppMainWindow::saveAppGuiSettings(QSettings &s) const
{
    PnpBaseMainWindow::saveAppGuiSettings(s);
    s.beginGroup(SETTING_GROUP_MAINWINDOW);
    s.setValue(k_show_run_logic, menuItemShowRunLogic->isChecked());
    s.endGroup();
}

void ReadoutAppMainWindow::loadAppGuiSettings(const QSettings &s)
{
    PnpBaseMainWindow::loadAppGuiSettings(s);
    auto showRL = s.value(SETTING_GROUP_MAINWINDOW + "/" + k_show_run_logic, false).toBool();
    if(config.defaultSetup.runLogic.isAnyLimit())
        showRL = true;
    menuItemShowRunLogic->setChecked(showRL);
    sendShowRunLogicState();
}

void ReadoutAppMainWindow::onFsmStateChange(const FsmState &state)
{
    PnpBaseMainWindow::onFsmStateChange(state);
    bool en = state == FsmState::Error || state == FsmState::Ready;
    auto items = {menuItemDiscoverDialog};
    for (auto a : items)
        if (a)
            a->setEnabled(en);
    updateControlState();
}

void ReadoutAppMainWindow::setErrOnTrigOnXOff(bool err)
{
    QSignalBlocker(this->menuItemErrorOnXOff);
    menuItemErrorOnXOff->setChecked(err);
}

void ReadoutAppMainWindow::setErrOnSerDesUnlock(bool err)
{
    QSignalBlocker(this->menuItemErrorOnSerDesUnlock);
    menuItemErrorOnSerDesUnlock->setChecked(err);
}

void ReadoutAppMainWindow::setErrOnFeLinkErr(bool isErr, bool isWarn)
{
    QSignalBlocker(this->menuItemErrorOnFeLinkErr);
    QSignalBlocker(this->menuItemWarnOnFeLinkErr);
    menuItemErrorOnFeLinkErr->setChecked(isErr);
    menuItemWarnOnFeLinkErr->setChecked(isWarn);
}

void ReadoutAppMainWindow::setDeviceSdbList(const DeviceIndex &, const QVariantList &devSdbList) const
{
    for(QAction *a : menuTools->actions()+menuOptions->actions()) {
        if(a->isVisible())
            continue;
        auto sdbList = a->property(k_sdb_list);
        if(!sdbList.isValid())
            continue;
        for(auto v : sdbList.toList()) {
            if(v.isValid() && devSdbList.contains(v)) {
                a->setVisible(true);
                break;
            }
        }
    }
}

void ReadoutAppMainWindow::adcDeserConfigChecked(const DeviceIndex &index, const AdcDeserCheckStatus &st)
{
    if(!config.knownConfigs.contains(index))
        return;

    if(!adcDesCheckLabel) {
        adcDesCheckLabel = new QLabel(this);
        statusBar()->addWidget(adcDesCheckLabel);
        adcDesCheckLabel->hide();
        connect(adcDesCheckLabel, &QLabel::destroyed,
                [this](){ adcDesCheckLabel = nullptr; });
    }

    auto indexStr = index.getSerialStr();
    auto map = adcDesCheckLabel->property(k_des_check).toMap();

    auto it = map.find(indexStr);
    if(it != map.end()) {
        if(it->value<AdcDeserCheckStatus>() == st)
            return;
        *it = QVariant::fromValue(st);
    } else {
        map[indexStr] = QVariant::fromValue(st);
        if(st.getStatus() == AdcDeserCheckStatus::Status::Ok)
            return;
    }
    adcDesCheckLabel->setProperty(k_des_check, map);

    updateAdcDesCheckLabel(map);
}

void ReadoutAppMainWindow::clockDividerStatusUpdated(const DeviceIndex &index, const ClockDividerStatus &st)
{
    if(!config.knownConfigs.contains(index))
        return;

    if(!clockDividerLabel) {
        clockDividerLabel = new QLabel("Clk divider sync req.", this);
        statusBar()->addWidget(clockDividerLabel);
        clockDividerLabel->hide();
        connect(clockDividerLabel, &QLabel::destroyed,
                [this](){ clockDividerLabel = nullptr; });
        auto p = clockDividerLabel->palette();
        p.setColor(clockDividerLabel->foregroundRole(), Qt::darkYellow);
        clockDividerLabel->setPalette(p);
    }

    auto list = clockDividerLabel->property(k_des_check).toList();

    auto indexVar = QVariant::fromValue(index);
    bool listIsEmpty = list.isEmpty();
    if(st.needSync()) {
        if(!list.contains(indexVar)) {
            list << indexVar;
        } else {
            return;
        }
    } else {
        if(list.removeAll(indexVar) == 0)
            return;
    }

    if(listIsEmpty != list.isEmpty()) {
        clockDividerLabel->setVisible(!list.isEmpty());
    }

    QStringList strList{"Clock divider synchronization requiered for:"};
    for(auto i : list) {
        strList << i.value<DeviceIndex>().getIdent();
    }
    clockDividerLabel->setToolTip(strList.join('\n'));
    clockDividerLabel->setProperty(k_des_check, list);
}

void ReadoutAppMainWindow::setConfig(const ConfigKey &, const BaseDeviceAppConfig &_config)
{
    if(config == _config)
        return;
    auto runLogicOn = _config.defaultSetup.runLogic.isAnyLimit();
    bool runLogicUpdated = runLogicOn != config.defaultSetup.runLogic.isAnyLimit();

    config = _config;

    if(menuItemPeriodicalDesCheck &&
        menuItemPeriodicalDesCheck->isChecked() != config.defaultSetup.periodicalAdcDelayCheck) {
        QSignalBlocker(this->menuItemPeriodicalDesCheck);
        menuItemPeriodicalDesCheck->setChecked(config.defaultSetup.periodicalAdcDelayCheck);
    }

    if(runLogicUpdated && menuItemShowRunLogic) {
        if(!menuItemShowRunLogic->isChecked() && runLogicOn) {
            menuItemShowRunLogic->setChecked(true);
        }
        menuItemShowRunLogic->setEnabled(!runLogicOn);
        sendShowRunLogicState();
    }
}

void ReadoutAppMainWindow::onRemoteStatusUpdated(const RemoteControlServerStatus &status)
{
    if(underRemoteCtrl == status.portInUse)
        return;

    underRemoteCtrl = status.portInUse;
    updateControlState();
}
