//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#include "RcClientManager.h"
#include "Client.h"
#include "ClientManagerFsm.h"
#include "pnp-server/PNPMessage.h"
#include "remote-control-server/RunControlCommands.h"
#include "base/ThreadController.h"

#include <QElapsedTimer>
#include <QHostAddress>
#include <QJsonObject>
#include <QJsonDocument>
#include <QTimer>

RcClientManager::RcClientManager(QObject *parent)
    : QObject(parent)
    , fsm(new ClientManagerFsm(this))
    , warnCheckTimer(new QTimer(this))
    , curSpillNum(0)
{
    qRegisterMetaType<RC_CLIENT_PROTO_ID>();
    qRegisterMetaType<FsmState>();
    qRegisterMetaType<ClientInfo>();
    qRegisterMetaType<RcProgramState>();

    connect(warnCheckTimer, &QTimer::timeout,
            this, &RcClientManager::checkBackupStatus);
    warnCheckTimer->setInterval(500);
    warnCheckTimer->setSingleShot(true);

    connect(fsm, &BaseFsm::stateChanged,
            this, &RcClientManager::onFsmStateChange);
    connect(fsm, &BaseFsm::stateChanged,
            this, &RcClientManager::runStateChanged);
    connect(this, &RcClientManager::printFsmInput,
            fsm, &BaseFsm::printFsmInput);
}

RcClientManager::~RcClientManager()
{
    emit deleteClient(ClientIndex());
}

void RcClientManager::setRcClientsConfig(const RcClientConfigMap &newConfig)
{
    if(config == newConfig)
        return;

    for(auto index : clientsInfo.keys()) {
        if(!newConfig.contains(index)) {
            removeClient(index);
        }
    }

    ++emit_client_config_guard;

    const auto oldSet = makeClientIndexSet(config.keys());
    const auto newSet = makeClientIndexSet(newConfig.keys());
    for(auto i : oldSet - newSet)
        removeClient(i);

    for(auto i : newSet & oldSet) {
        if(config[i].evNumCheck != newConfig[i].evNumCheck) {
            emit setEvNumCheck(i, newConfig[i].evNumCheck);
        }

        if(config[i].en != newConfig[i].en) {
            if(newConfig[i].en)
                emit startListen(i);
            else
                emit stopListen(i);
        }
    }

    config = newConfig;

    auto toAdd = newSet - oldSet;
    for(auto i : toAdd) {
        if(!addClient(i, config.value(i))) {
            qWarning()<<QString("Failed to add client from saved configuration: %1")
                        .arg(i.toString());
        }
    }

    updateFsmClientSet();
    emit rcProgramStateUpdated(getRcProgramState());
    --emit_client_config_guard;
}

void RcClientManager::onRestartAll(const RcRunId &runId, const QVariant &startEvNum)
{
    runInfo.setId(runId);
    runInfo.setStartEvNum(startEvNum);

    fsm->setRequestState(FsmState::Starting);
}

void RcClientManager::onStopAll(const QVariant &stopEvNum)
{
    runInfo.setStopEvNum(stopEvNum);
    fsm->setRequestState(FsmState::Stopping);
}

void RcClientManager::clientRunError(const ClientIndex &clientId, const RcRunId &runId, const QString &errStr)
{
    if(clientId.isEmpty())
        return;
    //todo: somehow add uuid into incoming runId
    if(get_run_id().toString() != runId.toString())
        return;

    QString clErrStr = QString("Error in client %1 (runNumber=%2)")
            .arg(clientId.toString()).arg(get_run_id().runNumber);
    if(!errStr.isEmpty()){
        clErrStr.append(":").append(errStr);
    }
    qCritical() << clErrStr;
    emit runError(get_run_id(), clErrStr);
}

void RcClientManager::clientRunWarn(const ClientIndex &clientId, const RcRunId &runId, const QString &warnStr)
{
    //todo: somehow add uuid into incoming runId
    if (runId.toString() != get_run_id().toString() || warnStr.trimmed().isEmpty())
        return;

    auto clInfoIt = clientsInfo.find(clientId);
    if(clInfoIt == clientsInfo.end())
        return;

    auto &warn = clInfoIt->warn;
    warn.active = true;
    warn.msg = warnStr;
    warn.spillNum = curSpillNum;
    warn.timer.start();

    if(!warnCheckTimer->isActive())
        warnCheckTimer->start(/*ClientWarn::TIMEOUT*/);

    updateWarnString();
    emit clientInfoUpdated(*clInfoIt);
}

void RcClientManager::clientBackPressure(const ClientIndex &clientId, const RcRunId &runId, bool bp)
{
    if(clientId.isEmpty())
        return;

    bool oldBP = isBackPressureOn();
    clientsInfo[clientId].backPressure = bp;
    bool newBP = isBackPressureOn();

    if(oldBP != newBP) {
        emit backPressure(runId, newBP);
        emit clientInfoUpdated(clientsInfo[clientId]);
    }
}

void RcClientManager::clientStatusUpdated(const ClientIndex &id, const QString &str)
{
    auto clInfoIt = clientsInfo.find(id);
    if(clInfoIt == clientsInfo.end())
        return;

    if(clInfoIt->status == str)
        return;

    clInfoIt->status = str;
    emit clientInfoUpdated(*clInfoIt);
}

void RcClientManager::clientFsmStateUpdated(const ClientIndex &id, FsmState state)
{
    qDebug() << __PRETTY_FUNCTION__
             << QString("id:%1 state:%2").arg(id.toString()).arg(fsmStateToString(state));
    if(state == FsmState::Error && fsm->getState() != FsmState::Error)
        qWarning() << "Got error state from client" << id.toString();
    auto clInfoIt = clientsInfo.find(id);
    if(clInfoIt == clientsInfo.end())
        return;

    if(clInfoIt->fsmState == state)
        return;

    clInfoIt->fsmState = state;
    fsm->setClientState(id, state);
    emit clientInfoUpdated(*clInfoIt);

    // clear after disconnect
    if(state == FsmState::Unknown)
        rcProgramState.clients.remove(id);

    if(!rcProgramState.clients.contains(id))
        emit rcProgramStateUpdated(getRcProgramState());
}

void RcClientManager::clientRcProgramStateUpdated(const ClientIndex &id, const RcProgramState &state)
{
    if(!config.contains(id))
        return;

    if(id != state.index)
        return;

    auto it = rcProgramState.clients.find(id);
    if(it == rcProgramState.clients.end()) {
        rcProgramState.clients.insert(id, state);
    } else {
        if(*it == state)
            return;
        *it = state;
    }

    emit rcProgramStateUpdated(getRcProgramState());
}

void RcClientManager::clientVmeFlag(const ClientIndex &id, bool hasVme)
{
    auto clInfoIt = clientsInfo.find(id);
    if(clInfoIt == clientsInfo.end())
        return;

    if(clInfoIt->has_vme == hasVme)
        return;

    clInfoIt->has_vme = hasVme;
    emit clientInfoUpdated(*clInfoIt);
}

void RcClientManager::clientProtocolChanged(const ClientIndex &id, RC_CLIENT_PROTO_ID protoId)
{
    auto clInfoIt = clientsInfo.find(id);
    if(clInfoIt == clientsInfo.end())
        return;

    if(clInfoIt->protoId == protoId)
        return;

    clInfoIt->protoId = protoId;
    emit clientInfoUpdated(*clInfoIt);
}

void RcClientManager::setClientStatus(const ClientIndex &id, const RcClientPeerInfo &st)
{
    auto clInfoIt = clientsInfo.find(id);
    if(clInfoIt == clientsInfo.end())
        return;

    if(clInfoIt->clientStatus == st)
        return;

    clInfoIt->clientStatus = st;
    emit clientInfoUpdated(*clInfoIt);
}

void RcClientManager::onFsmStateChange(const FsmState &state, const FsmState &oldState)
{
    switch (state) {
    case FsmState::Starting :
    {
        sendClientMap();
        if(hasAnyEnabledClient()) {
            emit sendRunState(FsmState::Starting, runInfo.get_run_id(), runInfo.get_start_ev_num());
        }
        fsm->setRequestState(FsmState::Run);
        break;
    }
    case FsmState::Stopping :
    {
        //todo: transmit to clients stopEvNum
        //runInfo.get_stop_ev_num();
        if(hasAnyEnabledClient()) {
            emit sendRunState(FsmState::Stopping, runInfo.get_run_id(), runInfo.get_stop_ev_num());
        }
        fsm->setRequestState(FsmState::Ready);
        break;
    }
    case FsmState::Error :
    {
        if(oldState != FsmState::Init)
            qCritical("Error while executing %s",
                     fsmStateToString(oldState).toStdString().c_str());
        if(oldState == FsmState::Starting
                || oldState == FsmState::Stopping)
            printFailedClients(getExpectedClientsState(oldState));
        emit sendRunState(FsmState::Stopping, runInfo.get_run_id(), {});
        break;
    }
    default:;
    }
}

void RcClientManager::pnpCacheUpdated(const PnpInfoCache &cache)
{
    ClientIndexSet clientToSearch;
    for(auto clIt=clientsInfo.begin(); clIt!=clientsInfo.end(); ++clIt) {
        if(clIt->prog_uuid.isNull()) {
            clientToSearch << clIt.key();
        } else {
            auto pdIt = cache.constFind(clIt->prog_uuid);
            if(pdIt == cache.cend()) {
                clIt->prog_uuid = {};
                emit setClientEmptyAddress(clIt.key());
            } else {
                updateClientAddress(*pdIt);
            }
        }
    }

    if(!clientToSearch.isEmpty()) {
        for(auto pdIt=cache.cbegin(); pdIt!=cache.cend(); ++pdIt) {
            if(clientToSearch.remove(pdIt->getClientIndex())) {
                updateClientAddress(*pdIt);

                if(clientToSearch.isEmpty())
                    break;
            }
        }
    }
}

void RcClientManager::checkBackupStatus()
{
    QElapsedTimer earliesTimer;
    bool changed = false;
    for(auto it=clientsInfo.begin(); it!=clientsInfo.end(); ++it) {
        auto &warn = it->warn;
        if(warn.isActive(curSpillNum)) {
            if(!earliesTimer.isValid() || earliesTimer.msecsTo(warn.timer)<0)
                earliesTimer = warn.timer;
            continue;
        }
        if(warn.active) {
            warn.active = false;
            changed = true;
            emit clientInfoUpdated(*it);
        }
    }
    if(earliesTimer.isValid())
        warnCheckTimer->start(/*ClientWarn::TIMEOUT - earliesTimer.elapsed()*/);
    if(changed)
        updateWarnString();
}

bool RcClientManager::addClient(const ClientIndex &index, const RcClientConfig &cfg)
{
    if(index.getProgramType().isEmpty() || index.getProgramIndex().isEmpty())
        return false;
    auto clIdent = index.toString();
    if(clientsInfo.contains(index)) {
        qWarning() << "Can't add client with dublicated " << clIdent;
        return false;
    }

    clientsInfo.insert(index, ClientInfo(index));

    Client *c = new Client(index);
    c->setObjectName(QString("Client for %1 thread").arg(clIdent));
    new ThreadController(c, this);

    connect(this, &RcClientManager::deleteClient, c, &Client::deleteClient);
    connect(c, &Client::runError, this, &RcClientManager::clientRunError);
    connect(c, &Client::runWarning, this, &RcClientManager::clientRunWarn);
    connect(c, &Client::backPressure, this, &RcClientManager::clientBackPressure);
    connect(c, &Client::statusUpdated, this, &RcClientManager::clientStatusUpdated);
    connect(c, &Client::fsmStateUpdated, this, &RcClientManager::clientFsmStateUpdated);
    connect(c, &Client::rcProgramStateUpdated, this, &RcClientManager::clientRcProgramStateUpdated);
    connect(c, &Client::vmeFlag, this, &RcClientManager::clientVmeFlag);
    connect(c, &Client::protocolChanged, this, &RcClientManager::clientProtocolChanged);
    connect(c, &Client::statusChanged, this, &RcClientManager::setClientStatus);
    connect(this, &RcClientManager::setClientAddress, c, &Client::setAddress);
    connect(this, &RcClientManager::setClientEmptyAddress, c, &Client::setEmptyAddress);
    connect(this, &RcClientManager::startListen, c, &Client::startReadOut);
    connect(this, &RcClientManager::stopListen, c, &Client::stopReadOut);
    connect(this, &RcClientManager::setEvNumCheck, c, &Client::setEvNumCheck);
    connect(this, &RcClientManager::sendRunState, c, &Client::sendRunState);
    connect(this, &RcClientManager::sendRunDocumentToClients, c, &Client::sendRunDocumentToClients);
    connect(this, &RcClientManager::allRcProgramStateChanged, c, &Client::sendRcProgramState);
//    connect(this, &RcClientManager::updateClientReadout,
//            c, &Client::updateClientReadout);

    connect(c, &Client::config_transfer, this, &RcClientManager::config_transfer);

    connect(this, &RcClientManager::evNumRequest,
            c, &Client::evNumRequest);
    connect(c, &Client::evNumResponse,
            this, &RcClientManager::evNumResponse);
    clientProtocolChanged(index, RC_CLIENT_PROTO_UNKNOWN);
    if(cfg.en)
        emit startListen(index);
    emit setEvNumCheck(index, cfg.evNumCheck);

    return true;
}

void RcClientManager::updateClientAddress(const ProgramDescription &pd)
{
    auto clIt = clientsInfo.find(pd.getClientIndex());
    if(clIt == clientsInfo.end())
        return;

    bool ok;
    const ProgramInterface &i = pd.getProgramInterface(ProgramInterfaceRemoteControl, &ok);
    if(ok && !i.host.isNull() && i.port) {
        RcClientPeerInfo peer{.host = i.host, .port = i.port, .hostName = pd.hostName};
        peer.host = i.host;
        peer.port = i.port;
        peer.hostName = pd.hostName;
        emit setClientAddress(clIt->getIndex(), peer);
        clIt->prog_uuid = pd.uuid;
    }
}

// Send client map to EvNumChecker of this app
void RcClientManager::sendClientMap() const
{
    ClientIndexSet set;
    for(auto it=config.cbegin(); it!=config.cend(); ++it) {
        if(!it->en || !it->evNumCheck)
            continue;

        set << it.key();
    }
    emit  clientMapUpdated(set);
}

void RcClientManager::updateWarnString() const
{
    QStringList clWarnStr;
    for(auto clInfo: clientsInfo) {
        if(clInfo.warn.active)
            clWarnStr << QString("%1:%2") .arg(clInfo.getIndex().toString()).arg(clInfo.warn.msg);
    }
    emit runWarn(runInfo.get_run_id(), clWarnStr.join("; "));
}

void RcClientManager::removeClient(const ClientIndex &clientId)
{
    qDebug()<<"Going to delete client #"<<clientId;
    emit deleteClient(clientId);

    clientsInfo.remove(clientId);
    rcProgramState.clients.remove(clientId);
}

void RcClientManager::onClientConfigChanged() const
{
    sendUpdatedClientConfig();
    updateFsmClientSet();
}

void RcClientManager::sendUpdatedClientConfig() const
{
    if(emit_client_config_guard)
        return;
    emit rcClientsConfigChanged(config);
}


void RcClientManager::updateFsmClientSet() const
{
    ClientIndexSet enabledClients;
    QMap<ClientIndex, FsmState> clientStatus;
    for(auto it=config.cbegin(); it!=config.cend(); ++it) {
        if(!it->en)
            continue;
        enabledClients << it.key();
        auto infoIt = clientsInfo.constFind(it.key());
        if(infoIt!=clientsInfo.cend()) {
            clientStatus[it.key()] = infoIt->fsmState;
        }
    }
    fsm->setClientList(enabledClients, clientStatus);
}

bool RcClientManager::isBackPressureOn() const
{
    for(auto it=config.cbegin(); it!=config.cend(); ++it) {
        if(!it->en)
            continue;
        auto infoIt = clientsInfo.constFind(it.key());
        if(infoIt!=clientsInfo.cend() && infoIt->backPressure) {
            return true;
        }
    }
    return false;

}

bool RcClientManager::hasAnyEnabledClient() const {
    for(auto c : config) {
        if(c.en) {
            return true;
        }
    }
    return false;
}

void RcClientManager::printFailedClients(const FsmState &expectedSt) const
{
    for(auto it=config.cbegin(); it!=config.cend(); ++it) {
        if(!it->en)
            continue;
        auto infoIt = clientsInfo.constFind(it.key());
        if(infoIt!=clientsInfo.cend() && infoIt->fsmState != expectedSt) {
            qCritical() << QString("Client %1 not in %2: remote state=%3")
                               .arg(it.key().toString())
                               .arg(fsmStateToString(expectedSt))
                               .arg(fsmStateToString(infoIt->fsmState));
        }
    }
}

FsmState RcClientManager::getExpectedClientsState(const FsmState &state) const
{
    switch (state) {
    case FsmState::Calibrate:
    case FsmState::Starting:
    case FsmState::Run:
        return FsmState::Run;
    default:
        return FsmState::Ready;
    }
}

bool RcClientManager::isCfgChangeAllowed() const
{
    return fsm->getState() == FsmState::Error || fsm->getState() == FsmState::Ready;
}

bool RcClientManager::clientsHasVme() const
{
    for(auto it=config.cbegin(); it!=config.cend(); ++it) {
        if(!it->en)
            continue;
        auto infoIt = clientsInfo.constFind(it.key());
        if(infoIt!=clientsInfo.cend() && infoIt->hasVme())
            return true;
    }
    return false;
}

RcProgramState RcClientManager::getRcProgramState() const
{
    auto ret = rcProgramState;
    for(auto it=config.cbegin(); it!=config.cend(); ++it) {
        if(!it->en)
            continue;

        if(ret.clients.contains(it.key()))
            continue;

        auto &st = ret.clients[it.key()];
        st.index = it.key();
        auto infoIt = clientsInfo.constFind(it.key());
        if(infoIt != clientsInfo.cend())
            st.fsm = infoIt->fsmState;
    }
    return ret;
}
