//
//    Copyright 2014-2015 Ivan Filippov
//    Copyright 2014-2015 Ilja Slepnev
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef RCCLIENTMANAGER_H
#define RCCLIENTMANAGER_H

#include "RunStates.h"
#include "config/RcClientConfig.h"
#include "config/RcClientProtoVersions.h"
#include "remote-control-server/types.h"
#include "remote-control-server/RcProgramState.h"
#include "RcClientData.h"
#include "RcRunInfo.h"

#include "base/ClientIndexEnabled.h"

#include <QObject>
#include <QVector>
#include <QMap>
#include <QDateTime>
#include <QUuid>

class Client;
class ClientManagerFsm;
class Config;
class ProgramDescription;
class PnpInfoCache;
struct RunDocument;

class QElapsedTimer;
class QHostAddress;
class QTimer;

class RcClientManager : public QObject
{
    Q_OBJECT
public:
    explicit RcClientManager(QObject *parent = nullptr);
    virtual ~RcClientManager() override;

    // temporary proxies
    RcRunId get_run_id() const { return runInfo.get_run_id(); }

signals:
    void deleteClient(const ClientIndex &index); // 0 - for all
    void startListen(const ClientIndex &index); // 0 - for all
    void stopListen(const ClientIndex &index); // 0 - for all
    void setEvNumCheck(const ClientIndex &i, bool evCheck) const;
    void sendRunState(FsmState state, const RcRunId &id, const QVariant &evNum);
    void runError(const RcRunId &id, QString errStr);
    void runWarn(const RcRunId &id, QString warnStr) const;
    void setClientAddress(const ClientIndex &index, const RcClientPeerInfo &peer);
    void setClientEmptyAddress(const ClientIndex &index);
    void clientMapUpdated(const ClientIndexSet &set) const; // for evNumCheck
    void evNumRequest(int checkIndex) const;
    void evNumResponse(int checkIndex, const ClientIndex &clientIndex, EvNumSet evNum) const;
    void backPressure(const RcRunId &id, bool backPressure);
    void config_transfer(const ClientIndex &index, const QJsonObject &json); // proxy signal from client to CoreApp
    void vmeFlagUpdated(bool hasVme) const;
    void runStateChanged(const FsmState &state);
    void printFsmInput();
    void sendRunDocumentToClients(const RunDocument &data) const;
    void clientInfoUpdated(const ClientInfo &info) const;
    void rcProgramStateUpdated(const RcProgramState &) const; // ClientManager's state (to Core)
    void allRcProgramStateChanged(const RcProgramState &) const; // proxy signal: Whole state (Core to Clients)

signals: void rcClientsConfigChanged(const RcClientConfigMap &) const;
public slots: void setRcClientsConfig(const RcClientConfigMap &newConfig);

public slots:
    void onRestartAll(const RcRunId &runId, const QVariant &startEvNum={});
    void onStopAll(const QVariant &stopEvNum={});
    void onFsmStateChange(const FsmState &state, const FsmState &oldState);
    void pnpCacheUpdated(const PnpInfoCache &cache);

private slots:
    void clientRunError(const ClientIndex &id, const RcRunId &runId, const QString &errStr);
    void clientRunWarn(const ClientIndex &id, const RcRunId &runId, const QString &warnStr);
    void clientBackPressure(const ClientIndex &id, const RcRunId &runId, bool bp);
    void clientStatusUpdated(const ClientIndex &id, const QString &str);
    void clientFsmStateUpdated(const ClientIndex &id, FsmState state);
    void clientRcProgramStateUpdated(const ClientIndex &id, const RcProgramState &state);
    void clientVmeFlag(const ClientIndex &id, bool hasVme);
    void clientProtocolChanged(const ClientIndex &id, RC_CLIENT_PROTO_ID protoId);
    void setClientStatus(const ClientIndex &index, const RcClientPeerInfo &st);
    void checkBackupStatus();

private:
    bool addClient(const ClientIndex &index, const RcClientConfig &cfg);
    void updateClientAddress(const ProgramDescription &pd);
    void sendClientMap() const;
    void updateWarnString() const;
    void removeClient(const ClientIndex &clientId);
    void sendUpdatedClientConfig() const;
    void onClientConfigChanged() const;
    void updateFsmClientSet() const;
    bool isBackPressureOn() const;
    bool hasAnyEnabledClient() const;
    void printFailedClients(const FsmState &expectedSt) const;
    FsmState getExpectedClientsState(const FsmState &state) const;
    bool isCfgChangeAllowed() const;
    bool clientsHasVme() const;
    RcProgramState getRcProgramState() const;

    RcRunInfo runInfo; // tmp: instead of 'RcClientManager * const m;'
    ClientManagerFsm * const fsm;
    RcProgramState rcProgramState;

    QMap<ClientIndex, ClientInfo> clientsInfo;
    RcClientConfigMap config;
    QTimer *const warnCheckTimer;
    quint32 curSpillNum;
    int emit_client_config_guard = 0;
};

#endif // RCCLIENTMANAGER_H
