/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "DeviceStatusDialog.h"

#include <algorithm>

#include <QFont>
#include <QTextEdit>
#include <QGridLayout>
#include <QScrollBar>

InfoViewHelper::InfoViewHelper(QString title, QWidget *parent)
    : title(title), textEdit(new QTextEdit(parent))
{}

void InfoViewHelper::clear()
{
    text.clear();
    if (textEdit)
        textEdit->clear();
}

void InfoViewHelper::updateText(const QString &s)
{
    if (!textEdit || s == text)
        return;
    text = s;

    auto vScrollBar = textEdit->verticalScrollBar();
    auto hScrollBar = textEdit->horizontalScrollBar();
    int vScrollBarValue = vScrollBar ? vScrollBar->value() : 0;
    int hScrollBarValue = hScrollBar ? hScrollBar->value() : 0;

    textEdit->setHtml(text);

    if (vScrollBar)
        vScrollBar->setValue(vScrollBarValue);
    if (hScrollBar)
        hScrollBar->setValue(hScrollBarValue);
}


DeviceStatusDialog::DeviceStatusDialog(QWidget *parent, Qt::WindowFlags f)
    : QDialog(parent, f),
    tabWidget(this)
{
    setModal(false);
    setInitialSize();
    auto layout = new QGridLayout(this);
    layout->addWidget(&tabWidget);
    infoViewHelper.insert(InfoType::Info, InfoViewHelper{"Info", this});
    infoViewHelper.insert(InfoType::Sdb, InfoViewHelper{"Sdb", this});
    infoViewHelper.insert(InfoType::Ports, InfoViewHelper{"Ports", this});
    infoViewHelper.insert(InfoType::Wr, InfoViewHelper{"WR", this});
    infoViewHelper.insert(InfoType::FeLink, InfoViewHelper{"FeLink", this});
    for (auto &r : infoViewHelper) {
        r.textEdit->setReadOnly(true);
        tabWidget.addTab(r.textEdit, r.title);
    }
}

void DeviceStatusDialog::resizeEvent(QResizeEvent *event)
{
    QDialog::resizeEvent(event);
    resized = true;
}

void DeviceStatusDialog::setMultiStatus(const QMap<DeviceIndex, ModularDeviceStatus> &r)
{
    statusMap = r;
    updateView();
}

void DeviceStatusDialog::setStatus(const ModularDeviceStatus &r)
{
    const auto &deviceIndex = r.getDeviceIndex();
    statusMap[deviceIndex] = r;
    if (currentDeviceIndex.isNull() && !deviceIndex.isNull())
        currentDeviceIndex = deviceIndex;
    if (deviceIndex == currentDeviceIndex)
        updateView();
}

void DeviceStatusDialog::setIndex(const DeviceIndex &r)
{
    currentDeviceIndex = r;
    updateView();
}

void DeviceStatusDialog::selectDevice(const DeviceIndex &r)
{
    setIndex(r);
}

void DeviceStatusDialog::showDevice(const DeviceIndex &r)
{
    setIndex(r);
    show();
}

void DeviceStatusDialog::updateView()
{
    if (! statusMap.contains(currentDeviceIndex)) {
        for (auto &r : infoViewHelper) {
            r.clear();
        }
        setWindowTitle("Status " + currentDeviceIndex.getIdent());
        return;
    }
    const auto &status = statusMap[currentDeviceIndex];
    setWindowTitle("Status " + status.getDeviceIdent());
    infoViewHelper[InfoType::Info].updateText(status.getDeviceDescrText());
    infoViewHelper[InfoType::Sdb].updateText(status.getSdbText());
    infoViewHelper[InfoType::Ports].updateText(status.getNetworkPortsText());
    infoViewHelper[InfoType::Wr].updateText(status.getWrText());
    infoViewHelper[InfoType::FeLink].updateText(status.getFeLinkText());
    adjustSize();
}

void DeviceStatusDialog::setInitialSize()
{
    QFont font; font.setStyleHint(QFont::Monospace);
    QFontMetrics fm(font);
    int testWidth = fm.horizontalAdvance("0123456789abcdef");
    int testHeight = fm.height();
    int width = std::clamp(testWidth * 7, 400, 2000);
    int height = std::clamp(testHeight * 40, 200, 1000);
    resize(width, height);
}

void DeviceStatusDialog::adjustSize()
{
    if (resized)
        return;
    if (infoViewHelper.empty())
        return;
    QSize docSize;
    QSize pSize = size();
    if (parentWidget())
        pSize = parentWidget()->size();

    for (auto &r : infoViewHelper) {
        auto textEdit = r.textEdit;
        QSize s(textEdit->document()->size().toSize());
        docSize = docSize.expandedTo(s);
    }
    QSize mySize = size();
    bool expandWidth = docSize.width() > mySize.width();
    bool expandHeight = docSize.height() > mySize.height();
    if (expandWidth || expandHeight) {
        auto newSize = docSize; // .grownBy({widthAdj, heightAdj, 0, 0});
        int newWidth = std::clamp(width(), newSize.width(), pSize.width());
        int newHeight = std::clamp(height(), newSize.height(), pSize.height());
        resize(newWidth, newHeight);
    }
}
