/*
**    Copyright 2021 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#include "MqttPublisher.h"

#include <QCoreApplication>
#include <QDebug>
#include <QHostInfo>
#include <QLoggingCategory>
#include <QSettings>
#include <QTimer>

#include "qmqtt.h"
#include "base/Globals.h"
#include "base/ServiceConfigUrl.h"

namespace {
Q_DECLARE_LOGGING_CATEGORY(lc_mqtt)
Q_LOGGING_CATEGORY(lc_mqtt, "afi.mqtt")
static const char *DEFAULT_URL = "mqtt://localhost";
static const bool DEFAULT_ENABLE = true;
static const int DEFERRED_INIT_TIMEOUT = 0;
}

MqttPublisher::MqttPublisher(const QString &serviceName, QObject *parent)
    : AbstractPublisher(parent)
{
    setObjectName("MqttPublisher");
    ServiceConfigUrl u(serviceName, QUrl::fromUserInput(DEFAULT_URL), DEFAULT_ENABLE);
    enabled = u.isEnabled();
    url = u.getUrl();
    if (url.isEmpty())
        enabled = false;
    if (!enabled)
        return;

    // thread may change after ctor, use QueuedConnection
    reconnectTimer = new QTimer(this);
    reconnectTimer->setSingleShot(true);
    reconnectTimer->start(DEFERRED_INIT_TIMEOUT);
    connect(reconnectTimer, &QTimer::timeout, this, &MqttPublisher::deferredInit, Qt::QueuedConnection);
}

MqttPublisher::~MqttPublisher()
{
    deleteConnection();
}


void MqttPublisher::deferredInit()
{
    createConnection();
}

void MqttPublisher::createConnection()
{
    deleteConnection();
    if (!enabled)
        return;
    // resolve to IP address
    const QHostInfo &hostInfo = QHostInfo::fromName(url.host());
    const QList<QHostAddress> &addresses = hostInfo.addresses();
    if (addresses.isEmpty()) {
        qWarning(lc_mqtt) << QString("no IP address for %1")
                          .arg(url.toDisplayString());
        return;
    }
    if (url.port() >= 0)
        mqtt = new QMQTT::Client(addresses.first(), url.port());
    else
        mqtt = new QMQTT::Client(addresses.first());
    connect(mqtt, &QMQTT::Client::connected, this, &MqttPublisher::connected);
    connect(mqtt, &QMQTT::Client::disconnected, this, &MqttPublisher::disconnected);
    connect(mqtt, &QMQTT::Client::error, this, &MqttPublisher::error);
    mqtt->setClientId(QCoreApplication::applicationName());
    qInfo(lc_mqtt).noquote().nospace() << "connecting to " + url.toString();
    mqtt->connectToHost();
}

void MqttPublisher::deleteConnection()
{
    delete mqtt;
    mqtt = nullptr;
}

bool MqttPublisher::event(QEvent* ev)
{
    if (ev->type() == QEvent::ThreadChange) {
        if (mqtt) {
            deleteConnection();
            QTimer::singleShot(0, this, &MqttPublisher::deferredInit);
        }
        return true;
    }
    return QObject::event(ev);
}

void MqttPublisher::pub(const QString &topic, const QByteArray &text)
{
    if (!enabled)
        return;
    if (!mqtt || !mqtt->isConnectedToHost())
        return;

    QMQTT::Message message(0, topic, text);
    quint16 ret = mqtt->publish(message);
    Q_UNUSED(ret)
}

void MqttPublisher::connected()
{
    qInfo(lc_mqtt).noquote().nospace() << "connected to " + url.toString();
}

void MqttPublisher::disconnected()
{
    qInfo(lc_mqtt).noquote().nospace() << "disconnected from " + url.toString();
}

void MqttPublisher::error(const int error)
{
    const auto &err = static_cast<QMQTT::ClientError>(error);
    qCritical(lc_mqtt).noquote().nospace() << err;
}
