﻿#include "MongoDB.h"

#include <QDateTime>
#include <QDebug>
#include <QVariant>
#include <QJsonDocument>

#ifndef Q_MOC_RUN
#include <mongocxx/options/client.hpp>
#endif

#include "base/ConfigMeta.h"
#include "base/ConfigMetaConverter.h"
#include "base/ConfigSelector.h"
#include "base/ConfigSelectorConverter.h"
#include "base/ConfigDocumentConverter.h"

#include "base/Globals.h"
#include "base/RcRunId.h"

#include "MongoConsts.h"
#include "mongoqt/BsonQtConverter.h"
#include "mongoqt/MongoHost.h"
#include "mongoqt/MongoCredentialsFile.h"
#include "RootConfigUtil.h"

using bsoncxx::builder::basic::kvp;
using bsoncxx::builder::basic::make_document;

namespace {
    template<typename T> bsoncxx::document::view_or_value mk_doc(std::string key, T val)
    {
        return make_document(kvp(key, val));
    }
}

void MongoDB::connectToDb()
{
    configureConnection();
    const auto &host = getConfiguredHost();
    if(host.has_value() && host->enable)
        alive = try_connection();
    else
        alive = false;
}

void MongoDB::mongo_init()
{
    QStringList primary_keys_saved;
    primary_keys_saved << program_type << program_index << configuration_name;
    if (remove_broken_docs(primary_keys_saved) != 0){
        qWarning() << "Remove brocken docs from Mongo Saved collection";
    }

//    if (!remove_duplicates()){
//        qWarning() << "Remove duplicates from Mongo Saved failed!";
    //    }
}

ConfigDocument MongoDB::readConfigDocument(const ConfigSelector &cs)
{
    if(cs.getIndex().isDefault())
        return {};

    if(cs.getType().isEmpty()){
        qWarning() << "Read config failed. Set program type";
        return {};
    }
    set_collection(COLLECTION_SAVED);
    // const QJsonObject projection({ {_id, 0} }); // 0 - exclude '_id' field from results
    const QJsonObject metaNotPhantom {{k_phantom, false}};
    const QJsonObject filter({ {k_config_selector, QVariant::fromValue(cs).value<QJsonValue>()},
                             {k_meta, metaNotPhantom} });

    QList<QJsonObject> result = MongoQt::find(filter/*, projection*/);

    if(result.size() >= 2){
        qWarning().noquote().nospace() << QString(
            "MongoDB warning: read config warning "
            "More then one configuration with the same "
            "ConfigSelector in database: %1").arg(cs.toString());
    }
    if(result.isEmpty()){
        qWarning() << QString(
            "MongoDB warning: read config failed. "
            "Can't find configuration: %1").arg(cs.toString());
        return {};
    }
    ConfigDocument doc = QVariant::fromValue(result.first()).value<ConfigDocument>();
    return doc;
}

bool MongoDB::writeConfigDocument(const ConfigDocument &cd)
{
    if(cd.getConfigSelector().getIndex().isDefault())
        return true;

    if(cd.isEmpty()){
        qWarning() << "write config failed. Set program type";
        return false;
    }

    set_collection(COLLECTION_SAVED);
    QJsonObject filter({
        {k_config_selector,
        QVariant::fromValue(cd.getConfigSelector()).value<QJsonObject>()}
    });
    auto cd_obj = QVariant::fromValue(cd).value<QJsonObject>();
    cd_obj.remove(_id);
    bool result = replace_one(
        filter,
        cd_obj,
        true
    );
    return result;
}

bool MongoDB::writeConfigRun(const RcRunId &rc, const QJsonObject &jsonCfg)
{
    Q_UNUSED(rc)
    set_collection(COLLECTION_HISTORY);
    return insert_one(jsonCfg);
}

QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > MongoDB::get_config_map(const DaqModule &daq_module)
{
    set_collection(COLLECTION_SAVED);
    QMap<ProgramIndex, QMap<ConfigurationName, QDateTime> > config_map;
    if(daq_module.isEmpty())
        return config_map;


    auto daqModuleJson = QVariant::fromValue(daq_module).value<QJsonValue>();
    QJsonObject metaNotPhantom {{k_phantom, false}};
    QJsonObject query {{k_config_selector, getJsonCSFromDaqModule(daq_module)},
                       {k_meta, metaNotPhantom} };
    QJsonObject projection;
    projection[k_config_selector] = 1;
    projection[k_meta] = 1;
    QList<QJsonObject> result = find(query, projection);

    for(auto &cfg_record : result) {
        auto cs = QVariant::fromValue(cfg_record[k_config_selector].toObject()).value<ConfigSelector>();
        QString program_index = cs.getConfigIndexName().getIndex().toString();
        QString configuration_name = cs.getConfigIndexName().getConfigurationName();
        auto mod_dt = RootConfigUtil::getModificationTimestamp(cfg_record);
        config_map[cs.getIndex()][cs.getConfigurationName()] = mod_dt;
    }
    return config_map;
}

QMap<ProgramIndex, QMap<ConfigurationName, QJsonObject> > MongoDB::get_json_map(const DaqModule &daq_module)
{
    set_collection(COLLECTION_SAVED);
    QMap<ProgramIndex, QMap<ConfigurationName, QJsonObject> > config_map;
    if(daq_module.isEmpty())
        return config_map;


    auto daqModuleJson = QVariant::fromValue(daq_module).value<QJsonValue>();
    QJsonObject metaNotPhantom {{k_phantom, false}};
    QJsonObject query {{k_config_selector, getJsonCSFromDaqModule(daq_module)},
                       {k_meta, metaNotPhantom}};
    QJsonObject projection;
    projection[k_config_selector] = 1;
    projection[k_config] = 1;
    QList<QJsonObject> result = find(query, projection);

    for(auto &cfg_record : result) {
        auto cs = QVariant::fromValue(cfg_record[k_config_selector].toObject()).value<ConfigSelector>();
        QString program_index = cs.getConfigIndexName().getIndex().toString();
        QString configuration_name = cs.getConfigIndexName().getConfigurationName();
        config_map[cs.getIndex()][cs.getConfigurationName()] = cfg_record[k_config].toObject();
    }
    return config_map;
}

int MongoDB::delete_config(const ConfigSelector &cs)
{
    set_collection(COLLECTION_SAVED);
    QJsonObject filter({ {k_config_selector, QVariant::fromValue(cs).value<QJsonObject>()} });
    return delete_many(filter);
}

RcRunId MongoDB::check_run_number(const RcRunId &rc)
{
    RcRunId to_return;
    set_collection(COLLECTION_HISTORY);
    QVector<int> run_numbers;

    QJsonObject match;
    match[run_index] = rc.runIndex;
    QJsonObject gte;
    gte["$gte"] = rc.runNumber;
    match[run_number] = gte;

    QJsonObject project;
    project[run_number] = 1;

    mongocxx::pipeline p{};
    p.match(BsonQtConverter::from_json(match));
    p.sort(mk_doc(QString(run_number).toStdString(), 1));
    p.project(BsonQtConverter::from_json(project));
    try {
        auto collection = get_collection();
        auto cursor = collection.aggregate(p);
        for (const auto &i : cursor) {
            run_numbers.append(BsonQtConverter::to_json(i)[run_number].toInt());
        }
    } catch (const std::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo check run number warning: %1").arg(str);
    }

    if(run_numbers.size() ==0)
        return rc;
    if(run_numbers.size() != 0)
        if(rc.runNumber != run_numbers[0]){
            return rc;
        }

    int k = rc.runNumber;
    for(int i=0; i< run_numbers.size(); ++i){
        if(run_numbers[i]-k > 1){
            break;
        }
        k = run_numbers[i];
    }
    to_return = rc;
    to_return.runNumber = k+1;
    return to_return;
}

void MongoDB::configureConnection()
{
    const auto &host = MongoCredentialsFile::get_host(Globals::getEnvironment());
    configuredHost = host;
    if (!host.enable)
        return;
    auto str = host.get_uri_string();
    set_uri(str);
    set_db(host.db);
}

//---remove docs from mongo
int MongoDB::remove_broken_docs(const QStringList &primary_keys)
{
    //TODO : candidate for deletion without fixes it can ruin whole DB
    return 0;
    set_collection(COLLECTION_SAVED);
    try {
        auto collection = get_collection();

        bsoncxx::builder::basic::array pk_array;
        for(int i=0; i<primary_keys.size(); ++i){
            pk_array.append(
                mk_doc(
                    primary_keys[i].toStdString(),
                    mk_doc("$exists", false)
                )
            );
        }
        bsoncxx::stdx::optional<mongocxx::result::delete_result> result  =
            collection.delete_many(
                make_document(
                    kvp("$or", pk_array)
                )
            );
        if(result){
            return result->deleted_count();
        }

        return 0;
    } catch (const std::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo remove broken docs warning: %1").arg(str);
        return 0;
    }
}

// BUG !
int MongoDB::remove_duplicates()
{
    set_collection(COLLECTION_SAVED);
    try {
        auto collection = get_collection();
        mongocxx::pipeline p{};

        p.group(make_document(
            kvp("_id",
                make_document(
                    kvp("configuration_name", "$configuration_name"),
                    kvp("program_type", "$program_type"),
                    kvp("program_index", "$program_index")
                )
            ),
            kvp("id", make_document(kvp("$first", "$id"))), // todo - BUG !!!!
            kvp("program_type", make_document(kvp("$first", "$program_type"))),
            kvp("program_index", make_document(kvp("$first", "$program_index"))),
            kvp("configuration_name", make_document(kvp("$first", "$configuration_name"))),
            kvp("date_of_modification", make_document(kvp("$max", "$date_of_modification")))
        ));

        QVector<QString> not_dublicated_docs;
        mongocxx::cursor cursor_aggregate = collection.aggregate(p, mongocxx::options::aggregate{});
        for (const bsoncxx::document::view& doc : cursor_aggregate){
            QString id;
            QString program_type;
            QString program_index;
            QString date_of_modification;

            bsoncxx::document::element id_element = doc["id"];
            std::string oid= id_element.get_oid().value.to_string();
            id = QString::fromStdString(oid);
            not_dublicated_docs.append(id);
        }

        mongocxx::options::find opt{};
        opt.projection(
            make_document(
                kvp("_id", 1)
            )
        );

        int delete_count = 0;
        mongocxx::cursor cursor_find = collection.find({}, opt);
        for (const bsoncxx::document::view& doc : cursor_find){
            bsoncxx::document::element id_element = doc["_id"];
            std::string oid= id_element.get_oid().value.to_string();
            QString id = QString::fromStdString(oid);
            if(!not_dublicated_docs.contains(id)){
                bsoncxx::stdx::optional<mongocxx::result::delete_result> result  =
                    collection.delete_one(
                        make_document(kvp("_id", doc["_id"].get_value())));
                delete_count += result->deleted_count();
                qInfo() << QString("Find duplicates in Mongo %1 success ").arg(MongoQt::get_collection_name());
            }
        }
        return delete_count;
    } catch (const std::exception& xcp) {
        QString str = QString::fromStdString(xcp.what());
        qWarning() << QString("Mongo remove duplicates warning: %1").arg(str);
        return 0;
    }
}
