/*
**    Copyright 2022 Ilia Slepnev
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**    This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <https://www.gnu.org/licenses/>.
*/

#ifndef MSTREAMV2CLIENT_H
#define MSTREAMV2CLIENT_H

#include <arpa/inet.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <sys/select.h>

#include <set>

#include <QHostAddress>

#include "base/HwipSocketAddress.h"
#include "types.h"

struct AckItem {
    AckItem() = default;
    AckItem(const AckItem &) = default;
    AckItem(const MStreamFragmentHeader &h, bool ver2 = true) :
        fid(h.getFragId(ver2)), foff(h.getFragOffset(ver2))
    {}

    friend bool operator==(const AckItem &, const AckItem &) = default;
    friend bool operator<(const AckItem &l, const AckItem &r) {
        return (l.fid < r.fid) ? true : (l.foff < r.foff);
    }
    friend uint qHash(const AckItem &r, uint seed = 0) {
        return qHash(r.fid) ^ qHash(r.foff) ^ seed;
    }
    int fid = -1;
    int foff = -1;
};

using AckList = std::set<AckItem>;

class MStreamV2Client
{
public:
    enum ProtoFlags {Ack = 1, Rst = 2, Syn = 4, Fin = 8};

    MStreamV2Client(const QString &name, int _timeout) : name(name), timeout(_timeout) {}
    ~MStreamV2Client();

    HwipSocketAddress getLocalAddress() const;
    HwipSocketAddress getPeerAddress() const;
    void setPeerAddress(const QHostAddress &address);
    bool connectToHardware();
    bool disconnectFromHardware();
    bool isBound() const;
    bool isConnected() const;
    void receivedFlags(quint32 seq, int flags);
    bool send_ack(const AckList &ackList);
    int recvmmsg (struct mmsghdr *vmessages, unsigned int vlen);
    void setAckSizeLimit(int limit) { ackSizeLimit = limit; }

private:
    enum ProtoState {Init, Closed, SynSent, SynRcvd, Open, CloseWait};
    void runProtoFsm();
    void init_si_other(const QHostAddress &ipaddr);
    bool close_mstream();
    bool bind_mstream();
    void logErrno(const QString &message) const;
    bool send_mlink_packet(const QByteArray &payload);
    bool send_ack_fragment(const QVector<AckItem> &ackList);
    bool sendProtoFlags(int flags);
    uint16_t getLocalPort() const;

    int ackSizeLimit = 0; // no limit
    QHostAddress peerAddress;
    QString name;
    int timeout = 0;
    int s_mstream = -1;
    struct sockaddr_in si_me;
    struct sockaddr_in si_other;
    int mlinkSeq = 0;
    ProtoState protoState = Init;
};

#endif // MSTREAMV2CLIENT_H
