//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//


#include <QDebug>
#include "qwt_plot_panner.h"
#include "qwt_plot_grid.h"
#include "qwt_plot_layout.h"
#include "qwt_plot_marker.h"
#include "qwt_legend.h"
#include "qwt_legend_item.h"
#include "qxw/QxwPlot.h"
#include "qxw/QxwZoomer.h"

QxwPlot::QxwPlot(QWidget *parent) :
    QwtPlot(parent), zoomer(0)
{
    // grid
    grid = new QwtPlotGrid;
    grid->setMajPen(QPen(QColor(200,200,200), 0, Qt::DotLine));
    //    grid->setMinPen(QPen(Qt::gray, 0 , Qt::DotLine));
    grid->attach(this);
    //  insertLegend(new QwtLegend(), QwtPlot::RightLegend);
    //    legend = new QwtLegend();
    setupZoomer();
    enableLegend(true);
    connect(this, SIGNAL(legendChecked(QwtPlotItem*,bool)), SLOT(onLegendChecked(QwtPlotItem*,bool)));
    setAutoReplot(true);
}

QxwPlot::~QxwPlot()
{
    delete grid;
    clear();
    delete zoomer;
    delete panner;
    if (legend())
        delete legend();
}

/// LeftButton for the zooming
/// MidButton for the panning
/// RightButton: zoom out by 1
/// Ctrl+RighButton: zoom out to full size
void QxwPlot::setupZoomer()
{
    plotLayout()->setAlignCanvasToScales(true);
    //    replot();

    updateZoomer(new QxwZoomer(canvas()));

    panner = new QwtPlotPanner(canvas());
    panner->setAxisEnabled(QwtPlot::yRight, false);
    panner->setMouseButton(Qt::MidButton);
}

void QxwPlot::updateZoomer(QxwZoomer * _zoomer)
{
    if(zoomer)
        delete zoomer;
    zoomer = _zoomer;

#if QT_VERSION < 0x040000
    zoomer->setMousePattern(QwtEventPattern::MouseSelect2,
                            Qt::RightButton, Qt::ControlButton);
#else
    zoomer->setMousePattern(QwtEventPattern::MouseSelect2, Qt::RightButton, Qt::ControlModifier);
#endif
    zoomer->setMousePattern(QwtEventPattern::MouseSelect3, Qt::RightButton, Qt::NoModifier);
}

//QwtPlotZoomer * QxwPlot::getZoomer() const
//{
//    return zoomer;
//}

QRectF QxwPlot::getDataBoundRect() const
{
    QRectF r;
    //  double xmin=MAXDOUBLE, xmax=-MAXDOUBLE, ymin=MAXDOUBLE, ymax=-MAXDOUBLE;
    foreach(int id, curves.keys()) {
        const QxwCurvePtr curve = curves.value(id);
        if (!curve || !curve->plot()) continue;
        const QRectF &curRect = curve->boundingRect();
        if (!curRect.isValid()) continue;
#if QT_VERSION < 0x050000
        r = r.unite(curRect);
#else
        r = r.united(curRect);
#endif
    }
    return r;
}

void QxwPlot::clear()
{
    clearCurves();
    clearXMarkers();
    QwtPlot::clear();
}

QwtPlotCurve &QxwPlot::getCurve(int id)
{
    QxwCurveSet::iterator it = curves.find(id);
    Q_ASSERT(it != curves.end());
    QxwCurvePtr p = *it;
    Q_ASSERT(p);
    return *p;
}

void QxwPlot::clearCurves()
{
    foreach(int id, curves.keys()) {
        if (curveAttached(id)) {
            getCurve(id).detach();
        }
        delete &getCurve(id);
    }
    curves.clear();
}

void QxwPlot::clearXMarkers()
{
    foreach(int id, xmarkers.keys()) {
        QxwMarkerPtr marker = xmarkers[id];
        if (marker){
            if(marker->plot()) {
                marker->detach();
            }
            delete marker;
        }
    }
    xmarkers.clear();
}

bool QxwPlot::curveAttached(int id) const
{
    QxwCurveSet::const_iterator it = curves.find(id);
    return it != curves.end() && (*it)->plot();
}

bool QxwPlot::curveExists(int id) const
{
    return curves.find(id) != curves.end();
}

bool QxwPlot::xmarkerExists(int id) const
{
    return xmarkers.find(id) != xmarkers.end();
}

bool QxwPlot::markrefExists(int id) const
{
    return markref.find(id) != markref.end();
}

void QxwPlot::setCurveBrush(int id, const QBrush &brush)
{
    if (curveExists(id)) {
        getCurve(id).setBrush(brush);
    }
}

void QxwPlot::setCurvePen(int id, const QPen &pen)
{
    if (curveExists(id)) {
        getCurve(id).setPen(pen);
    }
}

void QxwPlot::setCurveStyle(int id, QwtPlotCurve::CurveStyle style)
{
    if (curveExists(id)) {
        getCurve(id).setStyle(style);
    }
}

void QxwPlot::setCurveSymbol(int id, const QwtSymbol &symbol)
{
    if (curveExists(id)) {
        getCurve(id).setSymbol(symbol);
    }
}

void QxwPlot::setBrush(const QBrush &brush)
{
    foreach(int id, curves.keys())
        setCurveBrush(id, brush);
}

void QxwPlot::setPen(const QPen &pen)
{
    foreach(int id, curves.keys())
        setCurvePen(id, pen);
}

void QxwPlot::setStyle(QwtPlotCurve::CurveStyle style)
{
    foreach(int id, curves.keys())
        setCurveStyle(id, style);
}

void QxwPlot::setSymbol(const QwtSymbol &symbol)
{
    foreach(int id, curves.keys())
        setCurveSymbol(id, symbol);
}

void QxwPlot::setXMarkers(const std::vector<double> &v)
{

    foreach(int id, xmarkers.keys()) {
        xmarkers[id]->detach();
        delete xmarkers[id];
    }
    xmarkers.clear();
    int id = 0;
    foreach(double value, v) {
        QxwMarkerPtr mark(new QwtPlotMarker());
        mark->setLabelAlignment(Qt::AlignRight | Qt::AlignBottom);
        mark->setLineStyle(QwtPlotMarker::VLine);
        mark->setLinePen(QPen(QColor(Qt::black), 1, Qt::DotLine));
        mark->setLabel(tr("%1").arg(value));
        mark->setXValue(value);
        mark->attach(this);
        xmarkers[id] = mark;
        ++id;
    }
}

void QxwPlot::onLegendChecked(QwtPlotItem *plotItem, bool on)
{
    plotItem->setVisible(!on);
    qDebug() << "onLegendChecked(" << plotItem->title().text() <<  ", " << on << ")";
    //    qDebug() << "autoReplot " << (autoReplot()?"on":"off");
    if (autoReplot()) replot();
}

void QxwPlot::updateLegendItems()
{
    /*
    if (!legend()) return;
    foreach(int id, curves.keys()) {
        QxwCurvePtr curve = curves.value(id);
        if (!curve || !curve->plot()) continue;
        qDebug() << "curve " << curve->title().text() << (curve->isVisible()?" shown":" hidden");
        if (curve->isVisible()) {
            curve->setItemAttribute(QwtPlotItem::Legend, true);
            curve->updateLegend(legend());
            QwtLegendItem *legendItem = static_cast<QwtLegendItem*>(curve->legendItem());
            qDebug() << "legendItem: " << legendItem->text().text();
            legendItem->setChecked(true);

        }
    }
*/
}

void QxwPlot::enableLegend(bool enable)
{
    if (enable) {
        if (! legend()) {
            QwtLegend *legend = new QwtLegend(this);
            legend->setItemMode(QwtLegend::CheckableItem);
            insertLegend(legend, QwtPlot::RightLegend);
        }
    } else {
        if (legend())
            delete legend();
    }
}
